(* Mathematica Package *)

BeginPackage["BIFParser`", { "Combinatorica`", "DDUtils`", "DependencyDiagrams`"}]
(* Exported symbols added here with SymbolName::usage *)  
BIF2DD::usage = "";

Begin["`Private`"] (* Begin Private Context *) 
BIF2DD[fname_String] :=
	Module[ {input, variables = {}, functions = {}, diagram = EmptyGraph[0, Type -> Directed], i,
		     string, spaced, split, upper, vars, table},
		string = Import[fname, "String"];
		spaced = StringReplace[string,{";" -> "   ",
									   "(" -> " ( ",
									   ")" -> " ) "}];
		split = StringSplit[spaced, Whitespace | ","];
		upper = Map[ToUpperCase, split];
    	input = Cases[upper, x_ /; x != ""];
		For[i = 1, i <= Length[input], i++,
			Which[
				(* CASE 1 *)
				input[[i]] == "VARIABLE",
				Module[ {vname, typestr, vtype, vvals, vsize},
					vname = input[[++i]];
					i += 3;
					typestr = input[[i]];
					Which[
						typestr == "DISCRETE",
						vtype = Discrete;
		                i += 2;
		                vsize = FromDigits[input[[i]]];
		                i += 3;
		                vvals = Map[Symbol, Part[input, i ;; i + vsize - 1]];
		                i += vsize + 1,
		                True,
		                Print["badly formatted input at token number ", i];
		                Break[];
		           ];
		           AppendTo[variables, {vname, vtype, vvals}];
		           diagram = AddLabeledVertex[diagram, Symbol[vname], VariableN];
	            ];,
	           
				(* CASE 2 *)
				input[[i]] == "PROBABILITY",
      			Module[ {pname, pnodenum, args = {}, gvns = {}, addingargs,  vname, v, m, n, vnum, edges},
      				pname = SymbolNotInGraph[diagram, "prob"];
		          	diagram = AddLabeledVertex[diagram, pname, FunctionN];
			        pnodenum = VertexNum[diagram, pname];
			        i += 2;
			        AppendTo[args, input[[i++]]];
			        addingargs = True;
			        While[input[[i]] != ")",
			        	Which[
			        		(* case 1 *)
				            input[[i]] == "|",
				            addingargs = False,
				            (* case 2 *)
				            addingargs,
				            AppendTo[args, input[[i]]];,
				            (* default (case 3) *)
				          	True,
				          	AppendTo[gvns, input[[i]]];
			         	];
           				i++;
         			];
			        n = Length[gvns];
			        vname = args[[1]];
			        vnum = VertexNum[diagram, Symbol[vname]];
			        v = Cases[variables, l_List /; First[l] == vname][[1]];
			        m = Length[v[[3]]];
			        (*Print[gvns];
			        Print[args];			        
			        Print[variables];*)
					vars = Map[
						Module[{l},
							l = Cases[variables, var_ /; MatchQ[var[[1]], #]][[1]];
							{l[[1]], l[[3]]}
						]&, Join[gvns, args]];
			        If[ n == 0,
		            	Module[ {values, strings, whichBody},
		                	values = input[[i + 3 ;; i + 2 + m]];
		                    table = Map[ToExpression, values];
		                  	i += m + 2;
              			];,
              			(* else *)
              			Module[ {j, k, givens, values, case, valz, sizes, findIndex, findListAndIndex, place},
                  			givens = Map[Cases[variables, l_List /; First[l] == #][[1]] &, gvns];
			                k = Fold[Times, 1, Map[Length[#[[3]]] &, givens]];
			                i = i + 3;
			                valz = {};
			                For[j = 1, j <= k, j++,
			                	case = input[[i ;; i + n - 1]];
			                   	i += n + 1;
  			                   	(*Print[case];
  			                   	Print[ocase];
  			                   	Print[gvnIndices];*)
(*			                   	bools = MapIndexed["MatchQ[Slot[" <> ToString[#2[[1]] + 1] <> "]," <> #1 <> "]" &, ocase];
                   				bool = Fold[#1 <> " && " <> #2 &, bools[[1]], bools[[2 ;;]]];*)
                   				values = Map[ToExpression, input[[i ;; i + m - 1]]];
                   				AppendTo[valz, {MapThread[(#1->#2)&,{gvns, case}], values}];
(*                   				strings = MapThread["MatchQ[Slot[1]," <> ToString[#1] <> "]," <> ToString[#2] &, {v[[3]], values}];
                   				whichBody = ListToCommaSeparatedString[strings];*)
                   				i += m + 1;
                   			];
                  			i--;
(*                  			Print[valz];*)
							sizes = Map[{Length[#[[3]]]}&, givens];
							table = Table[{}, Evaluate[ReplacePart[sizes, 0->Sequence]]];
							findIndex[list_, item_] := Position[Map[ToString,list], item][[1,1]];
							findListAndIndex[var_ -> val_] := 
								findIndex[Cases[givens, gv_ /; MatchQ[Symbol[gv[[1]]], Symbol[var]]][[1,3]], val];
							For[j =1, j <= Length[valz], j++,
								place = Map[findListAndIndex[#]&, valz[[j,1]]];
								table[[ ReplacePart[place, 0->Sequence] ]] = valz[[j,2]];
							];
              			];
          			];
          			functions = Join[functions, functionBuilder[pname, vname, vars, table]];
          			diagram = AddEdge[diagram, {{pnodenum, vnum}, EdgeLabel -> CondDep, Type -> Directed}];
          			edges = Map[{{#, pnodenum}, EdgeLabel -> CondDep, Type -> Directed} &, Map[VertexNum[diagram, Symbol[#]] &, gvns]];
          			Map[Set[diagram , AddEdge[diagram, #1]] &, edges];
      			];
      		];
    	];
        {diagram, Map[ReplacePart[#, 1 -> Symbol[#[[1]]]] &, variables], functions}
    ];

functionBuilder[cpdName_, depVar_, vars_List, table_] :=
	Module[{mapMaker, map, testsMaker, tests, dtest, ns, slots, list, partMaker, part, 
		    pdfString, partMaker2, samplerPart, choices, samplerString},
		mapMaker[{name_, opts_List}] := 
			(ToString[name] -> MapIndexed[(#1->#2[[1]])&, opts]);
		testsMaker[{name_, opts_List}] :=
			"MemberQ[" <> ToString[FullForm[opts]] <> "," <> ToString[name] <> "]";
		partMaker[{name_, opts_List}] := 
			ToString[name] <> " /. (" <> ToString[FullForm[ToString[name]]] <> " /. map)";
		partMaker2[{name_, opts_List}] := 
			If[MatchQ[name, depVar],
				"All",
				ToString[name] <> " /. (" <> ToString[FullForm[ToString[name]]] <> " /. map)"
			];
		map = Map[mapMaker, vars];			
		tests = Fold[(#1 <> " && " <> #2)&, "True", Map[testsMaker, Cases[vars, x_ /; !MatchQ[x[[1]], depVar]]]];
		dtest = Fold[(#1 <> " && " <> #2)&, "True", Map[testsMaker, Cases[vars, x_ /;  MatchQ[x[[1]], depVar]]]];
		ns = ToString[cpdName];
		slots = ListToCommaSeparatedString[Map[(ToString[#[[1]]] <> "_")&, Sort[Cases[vars, x_ /; !MatchQ[x[[1]], depVar]]]]];
		list  = ListToCommaSeparatedString[Map[ ToString[#[[1]]]        &, Sort[Cases[vars, x_ /; !MatchQ[x[[1]], depVar]]]]];
		part  = ListToCommaSeparatedString[Map[partMaker, vars]];
		samplerPart = ListToCommaSeparatedString[Map[partMaker2, vars]];		
		choices = Cases[vars, x_ /; MatchQ[x[[1]], depVar]][[1,2]];
		pdfString = 
			ns <> " /:
			  PDF[" <> ns <> "[" <> slots <> "], "<> ToString[depVar] <> "_] := 
			  PDF[" <> ns <> "[" <> list <> "]][" <> ToString[depVar] <> "] /; " <> dtest <> "; " <> 
			ns <> " /:
			  PDF[" <> ns <> "[" <> slots <> "]] :=
				  Function[{" <> ToString[depVar] <> "},
				    Module[{table, map},
				     table = " <> ToString[FullForm[table]] <> ";
				     map = "<> ToString[FullForm[map]] <> ";
				     table[[" <> part <> "]]
				     ]
				    ] /; " <> tests <> ";";
		samplerString = 
			ns <>" /:
			  RandomChoice[" <> ns <> "[" <> slots <> "]] :=
				  Module[{table, choices, row, map},
					table = " <> ToString[FullForm[table]] <> ";
					choices = " <> ToString[FullForm[choices]] <> ";
					map = " <> ToString[FullForm[map]] <> ";
			   		row = table[[" <> samplerPart <> "]];
			   		RandomChoice[row -> choices]
			     ];";
		{pdfString, samplerString}
	];  


End[] (* End Private Context *)

EndPackage[]