(* Mathematica Package *)
BeginPackage["DDUtils`", {"Combinatorica`", "MultivariateStatistics`"}]

ListLength::usage = "";
DDTotal::usage = "";
DD2Norm::usage = "";
DDManhattan::usage = "";
DDFactorian::uage = "";
DDMvNrml::usage = "";
PlusOne::usage = "PlusOne[a,b] is one if b = a+1, 0 otherwise.";
MinusOne::usage = "";
NotKD::usage = "";
DeltaCondDist::usage = "";
EqualsIndic::usage = "";
IndicOther::usage = "";
Heaviside::usage = "";
Dirac::usage = "";
ListToSequence::usage = "";
PlusToList::usage = "";
TimesToList::usage = "";
ListToCommaSeparatedString::usage = "";
SymbolNotInExpression::usage = "";
IteratorFor::usage = "";
SymbolNotInGraph::usage = "";
DefaultContinuousProposal::usage = "";
DefaultIntegerProposal::usage = "";
DefaultDiscreteProposal::usage = "";

Begin["`Private`"]; (* Begin Private Context *) 
(* math used by generated code *) 

SetAttributes[ListLength, HoldFirst];
ListLength[x_List] :=
    Length[x];
    
DDTotal[x_,{minlev_,maxlev_}] :=
    If[ minlev > 1,
        Map[DDTotal[Part[x,#],{minlev-1,maxlev}]&,Range[ListLength[x]]],
        Module[ {v,y,z,w},
            v = IteratorFor[x];
            If[ maxlev >1,
                z = DDTotal[v,{minlev,maxlev-1}];
                y = IteratorFor[z];
                w = ReplaceAll[z,v->Part[x,y]];
                Sum[w,Evaluate[{y,1,ListLength[x]}]],
                Sum[Part[x,v],Evaluate[{v,1,ListLength[x]}]]
            ]
        ]
    ];
    
DD2Norm[x_,y_] :=
    Module[ {it},
        it = IteratorFor[x];
        Power[Sum[Power[Part[x,it],y],Evaluate[{it,1,ListLength[x]}]], 1/y]
    ];
DDDot[x_,y_] :=
    Module[ {it},
        it = IteratorFor[{x,y}];
        Sum[Times[Part[x,it],Part[y,it]],Evaluate[{it,1,ListLength[x]}]]
    ];
DDManhattan[x_] :=
    Module[ {it},
        it = IteratorFor[x];
        Sum[Abs[Part[x,it]],Evaluate[{it,1,ListLength[x]}]]
    ];
DDFactorial[x_] :=
    Module[ {it},
        it = IteratorFor[x];
        Product[it, Evaluate[{it,1,x}]]
    ];
DDMvNrml /:
	PDF[DDMvNrml[mu_, sigma_], x_] :=
		PDF[DDMvNrml[mu, sigma]][x];
DDMvNrml /:
	PDF[DDMvNrml[mu_, sigma_]] :=
		Function[{x},
			Module[{it1, it2, isigma},
				it1 = IteratorFor[x];
				it2 = IteratorFor[x];
				isigma = Inverse[sigma];
				Times[
					(*Heaviside[1-Dirac[Det[sigma]]], *)
					1/(Times[Power[2*Pi, (ListLength[x]/2)], Power[Det[sigma], .5]]),
					Power[E, 
						Sum[
							Times[ 
								-.5,
								Part[isigma, it1, it2],
								Plus[
									Times[Part[x,  it1], Part[x,  it2]    ],
									Times[Part[x,  it1], Part[mu, it2], -2],
									Times[Part[mu, it1], Part[mu, it2]    ] ]],
							Evaluate[{it1, 1, ListLength[x]}, {it2, 1, ListLength[x]}]]]]]];
DDMvNrml /:  
  Random`DistributionVector[DDMvNrml[m_, s_], n_Integer, prec_?Positive] :=
  	Table[RandomReal[MultinormalDistribution[m,s]], {n}]
    
PlusOne[x_, y_] :=
    KroneckerDelta[x+1,y];
    
MinusOne[x_, y_] :=
    KroneckerDelta[x,y+1];

NotKD[x_,y_] := (1-KroneckerDelta[x,y]);


DeltaCondDist /:
	PDF[DeltaCondDist[x_], y_] :=
		Piecewise[{{1, x == y}}];
DeltaCondDist /:
	PDF[DeltaCondDist[x_]] :=
		Piecewise[{{1, # == x}}] &;
DeltaCondDist /:  
  Random`DistributionVector[DeltaCondDist[x_], n_Integer, prec_?Positive] :=
  	Table[x, {n}];
DeltaCondDist /:
	RandomChoice[DeltaCondDist[x_]] :=
		x;


(* x is 1 with prob. 1 if y==z, and 0 with prob. 1 otherwise *)
EqualsIndic /:
	PDF[EqualsIndic[x_, y_], z_] :=
		Piecewise[{{1, x == y && z == 1}, {1, x != y && z == 0}}];
EqualsIndic /:
	PDF[EqualsIndic[x_, y_]] :=
		Piecewise[{{1, x == y && # == 1}, {1, x != y && # == 0}}] &;
EqualsIndic /:
	Random`DistributionVector[EqualsIndic[x_, y_], n_Integer, prec_?Positive] :=
		If[x == y, Table[1, {n}], Table[0, {n}]];
	
(* x is 1 with prob. 1 if y is 0, 0 with prob. 1 if y is 1. *)
IndicOther /:
	PDF[IndicOther[x_], y_] :=
		Piecewise[{{1, x == 0 && y == 1}, {1, y == 0 && x == 1}}];
IndicOther /:
	PDF[IndicOther[x_]] :=
		Piecewise[{{1, x == 0 && # == 1}, {1, # == 0 && x == 1}}] &;
IndicOther /:
	Random`DistributionVector[IndicOther[x_], n_Integer, prec_?Positive] :=
		Table[1 - x, {n}];
(*	(Dirac[y]*KroneckerDelta[x,1])+((1-Dirac[y])*Dirac[x]);*)
    
Heaviside[x_] :=
    If[ x > 0,
        1,
        0
    ];
    
Dirac[x_] := KroneckerDelta[x,0];
(*    If[ x == 0,
        1,
        0
    ];*)
(* end code needed by generated programs *)    

(* begin misc utilities *)
ListToSequence[{s___}] := s;
PlusToList[e1_Plus] := 	
		ReplacePart[e1, 0->List];
TimesToList[e1_Times] := 	
		ReplacePart[e1, 0->List];
		
ListToCommaSeparatedString[l_List] :=
    If[ Length[l] >0,
        Fold[ (#1 <> "," <> ToString[#2])&, ToString[First[l]], Rest[l]],
        "",
        ""
    ];
    
SymbolNotInExpression[expression_, base_String:"x"] :=
Module[{n = 0,string = base <> "0"},
	If[!ValueQ[DDUtils`generatedSymbols],
		DDUtils`generatedSymbols = {};
	];
	While[
		NameQ[string] || 
		!FreeQ[{expression, DDUtils`generatedSymbols},Symbol[string]] ,
		string = base <> ToString[++n];
	];
	AppendTo[DDUtils`generatedSymbols,Symbol[string]];
	Symbol[string]
];

IteratorFor[x_] := SymbolNotInExpression[x, "iterator"];
    
SymbolNotInGraph[graph_Graph, base_String:"x"] :=
    Module[ {string = base <> "0", c = 0},
        If[!ValueQ[DDUtils`generatedSymbols],
			DDUtils`generatedSymbols = {};
		];
		While[
        	NameQ[string] || 
        	MemberQ[DDGetVertexLabels[graph], h_[x__]/; {x}[[1]]==Symbol[string]] || 
        	MemberQ[DDUtils`generatedSymbols, Symbol[string]],
        	string = base <>ToString[++c];
        ];
       	AppendTo[DDUtils`generatedSymbols,Symbol[string]];
        Symbol[string]
    ];

SCALE[] := .1;
DefaultContinuousProposal /:
	PDF[DefaultContinuousProposal[old_, listOfRanges_], new_] :=
		PDF[DefaultContinuousProposal[old, listOfRanges]][new];
DefaultContinuousProposal /:
	PDF[DefaultContinuousProposal[old_, listOfRanges_]] :=
		Module[{y, ml, sizes,  var},
   			If[! MatchQ[old, _List],
    			y = {old};
			    ml = {listOfRanges},
			    y = old;
			    ml = listOfRanges;
    		];
		   	sizes = Map[#[[2]] - #[[1]] &, Map[Sort, ml]];
			var = Map[#*SCALE[] &, sizes];
			If[Length[y] == 1,
			    PDF[MultinormalDistribution[y, DiagonalMatrix[var]], {#}] &,
			    PDF[MultinormalDistribution[y, DiagonalMatrix[var]], #] &
			]
		];
DefaultContinuousProposal /:
	Random`DistributionVector[DefaultContinuousProposal[old_, listOfRanges_], n_Integer, prec_?Positive] :=
		Module[{y, ml, sizes,  var, r},
			If[! MatchQ[old, _List],
				y = {old};
				ml = {listOfRanges},
			    y = old;
			    ml = listOfRanges;
			];
		   	sizes = Map[#[[2]] - #[[1]] &, Map[Sort, ml]];
		   	var = Map[#*SCALE[] &, sizes];
		   	r = RandomReal[MultinormalDistribution[y, DiagonalMatrix[var]],n];
		   	If[Length[y] == 1,
		   		Map[#[[1]] &, r],
		    	r
		   	]
   		];

DefaultIntegerProposal /:
	PDF[DefaultIntegerProposal[old_, listOfRanges_], new_] := 
		PDF[DefaultIntegerProposal[old, listOfRanges]][new];
DefaultIntegerProposal /:
	PDF[DefaultIntegerProposal[old_, listOfRanges_]] :=
		Module[{ml},
			If[! MatchQ[old, _List],
			    ml = Sort[listOfRanges];
			    PDF[UniformDistribution[ml], #] &,
			    ml = Map[Sort, listOfRanges];
			    Fold[Times, 1, 
			      MapThread[PDF[UniformDistribution[#1], #2] &, {ml, #}]] &
		    ]
   		];
DefaultIntegerProposal /: 
	Random`DistributionVector[DefaultIntegerProposal[old_, listOfRanges_], n_Integer,  prec_?Positive] :=
		Module[{y, ml, mk1},
		   If[! MatchQ[old, _List],
			   y = {old};
			   ml = {listOfRanges},
			   y = old;
			   ml = listOfRanges;
		   ];
		   mk1[] := 
		   	Module[{r},
				r = Map[RandomInteger[#] &, Sort[ml]];
			    If[Length[y] == 1,
			    	r[[1]],
			    	r
			    ]
			];
			Table[mk1[], {n}]
	   ];

DefaultDiscreteProposal /:
	PDF[DefaultDiscreteProposal[old_, choice_], new_] :=
		PDF[DefaultDiscreteProposal[old, choice]][new];
DefaultDiscreteProposal /:
	PDF[DefaultDiscreteProposal[old_, choice_]] :=
		Piecewise[{
			{
				PDF[UniformDistribution[{0, Length[choice]}], 
					Position[choice, #][[1, 1]]], 
			 	MemberQ[choice, #]
			 }}] &;
DefaultDiscreteProposal /:
	RandomChoice[DefaultDiscreteProposal[old_, choice_]] :=
		RandomChoice[choice];

(* end misc utilities *)

End[]; (* End Private Context *)

EndPackage[];