Unprotect[Label];

(* ::Package:: *)
BeginPackage["DependencyDiagrams`", {"Combinatorica`", "DDUtils`", "FunctionMaker`"}];

Func::usage =  "Func[i, f, a, g, gates, fnode, dnode, indices, argIndices] represents all of the useful information about a function.";
Var::usage = "Var[i, v, exists, indices] represents all of the useful information about a variable";
DDIndex::usage = "DDIndex[i, ind, min, max] represents all of the useful information about an index";
IndexConstraint::usage = "IndexConstraint[i, c, relatedIndices] represents all of the useful information about an index constraint";
Parse::usage = "Parse[] is the result of parsing a graph, and contains a good deal of useful information about said graph.";
MCMCResults::usage = "MCMCResuts[g, t, t', pairs] represent the results of 'unrolling' a Dependency Diagram, along with some useful information generated along the way";
ShowGr::usage = "";
PlotGr::usage = "";
PrettyPrint::usage = "";
PrintGraphToFile::usage = "";
PreprocessGraph::usage = "";
AddLabeledVertex::usage = "";
FactorizeGraph::usage = "";
ExplicitPriors::usage = "";
FullySpecifyGraph::usage = "";
ParseGraph::usage = "";
PsiFunction::usage = "";
DDCollect::usage = "";
IndListToPSpec::usage = "";
IndListToTableSize::usage = "";
IndListToTable::usage = "";
GenerateMCMCDependencyDiagram::usage = "";
GenerateMCMCCode::usage = "";
GenerateForwardSamplingCode::usage = "";
MCMCDiagram::usage = "";
Gate::usage = "";
CondDep::usage = "";
IndexL::usage = "";
IndexCons::usage = "";
ArgIndex::usage = "";  
UnCondDep::usage = "";
Existence::usage = "";
FactorEdge::usage = ""; 
VariableN::usage = "";
ConstantN::usage = "";
FunctionN::usage = "";
ConstraintN::usage = "";
IndexN::usage = "";
SetIndexType::usage = "";
VertexNum::usage = "";
SortByOrder::usage = "";
Discrete::usage = "";
IntegerRange::usage = "";
Continuous::usage = "";

Begin["`Private`"]
INDEXTYPE = Part; 
(* Begin Private Context *) 
(* begin code relating to intermediate data structures *) 

Node[x_Func] :=
    x[[1]];
Label[x_Func] :=
    x[[2]];
Args[x_Func] :=
    x[[3]];
Givens[x_Func] :=
    x[[4]];
Gates[x_Func] :=
    x[[5]]; 
Vars[x_Func] :=
    Join[Args[x],Givens[x]];
IsFNode[x_Func] :=
    x[[6]];
IsDNode[x_Func] :=
    x[[7]];
Indices[x_Func] :=
    x[[8]];
ArgIndices[x_Func] :=
    x[[9]];
    
ImpingingVars[x_Func, y_List] :=
    Module[ {vz, vdz, ez},
        vz = Vars[x];
        vdz = Cases[y, z_Var /;MemberQ[vz,Label[y]]];
        ez = Map[Exist, vdz];
        Union[vz, ez, Gates[x]]
    ];

Node[x_Var] :=
    x[[1]];
Label[x_Var] :=
    x[[2]];
Exist[x_Var] :=
    x[[3]];
Indices[x_Var] :=
	x[[4]];

Node[x_DDIndex] :=
    x[[1]];
Label[x_DDIndex] :=
    x[[2]];
IndexMinName[x_DDIndex] :=
    x[[3]];
IndexMaxName[x_DDIndex] :=
    x[[4]];

Node[x_IndexConstraint] :=
    x[[1]];         
Label[x_IndexConstraint] :=
    x[[2]];
Indexed[x_IndexConstraint] :=
    x[[3]];
Constrained[x_IndexConstraint] :=
	Indexed[x]; (* "Indexed," the old name, is dumb. but i don't feel like searching for uses of it.  *)

AlgebraicExpression[p_Parse] :=
    p[[1]];
Funcs[p_Parse] :=
    p[[2]];
Vars[p_Parse] :=
    p[[3]];
IndexConstraints[p_Parse] :=
    p[[4]];
Indices[p_Parse] :=
    p[[5]];

(* end code relating to intermediate data structures *) 

(* begin orphaned code *)
EdgeNameMap = {
	Gate -> ToString[\[Gamma]], 
	CondDep -> Style["d", Italic], 
	IndexL -> ToString[\[Iota]], 
	IndexCons -> ToString[Subscript[\[Delta], \[Iota]]], 
	ArgIndex -> Overscript["\[Iota]", "~"], 
	UnCondDep -> Style["u", Italic], 
	Existence -> ToString[\[Epsilon]],
	FactorEdge -> Style["f", Italic]
};

EdgeDirectednessMap =  {Gate -> True,	CondDep -> True, IndexL -> True, 
        				IndexCons -> True, ArgIndex -> True, UnCondDep -> False,
        				Existence -> True,	FactorEdge -> False};
(* end orphaned code *)

(* begin graph functions *)
(* Returns the index within graph of the edge {p,c}, or 0 if no such edge exists *)
EdgeIndex[graph_Graph,p_Integer ,c_Integer] :=
    Module[ {q},
        q = Position[Edges[graph], {p,c}];
        If[ Length[q]==0,
            0,
            q[[1,1]]
        ]
    ];
    
VertexNum[g_Graph, v_Symbol] :=
	Module[{q}, 
		q = Position[DDGetVertexLabels[g], v]; 
		If[Length[q] == 0, 0, q[[1, 1]]]
	];    


RelativesOf[graph_Graph, vertexNum_Integer, dir_Integer, indices_:False] :=
    Module[ {edges,ret,i, other},
        ret = {};
        edges = Edges[graph];
        If[ dir ==1 || dir == 2,
            If[ dir == 1,
                other = 2,
                other = 1
            ];
            For[i = 1,i<=M[graph],i++,
            	If[ edges[[i,dir]]==vertexNum,
                	If[ indices,
                    	AppendTo[ret, {edges[[i,other]], i} ];,
                    	AppendTo[ret,  edges[[i,other]]     ];
                	];
	            ];
			];
        ];
        ret
    ];


ParentsOf[graph_Graph,vertexNum_Integer] :=
    RelativesOf[graph, vertexNum, 2, False];
ParentsOfIndices[graph_Graph,vertexNum_Integer] :=
    RelativesOf[graph, vertexNum, 2, True];
ChildrenOf[graph_Graph,vertexNum_Integer] :=
    RelativesOf[graph, vertexNum, 1, False];
ChildrenOfIndices[graph_Graph,vertexNum_Integer] :=
    RelativesOf[graph, vertexNum, 1, True];


Related[graph_Graph,node_Integer] :=
    Append[Union[ParentsOf[graph,node],ChildrenOf[graph,node]],node];
    
ConvertGraph[cg_Graph, opts___?OptionQ] :=
	Module[{v, l},
   		v = DDGetVertexLabels[cg, FilterRules[{opts}, Options[DDGetVertexLabels]]];
		l = DDGetEdgeLabels[cg, FilterRules[{opts}, Options[DDGetEdgeLabels]]];
		MapThread[{v[[#1[[1]]]] -> v[[#1[[2]]]], #2} &, {Edges[cg], l}]
	];
    
AllMaximalCliques[graph_Graph,removeSingletons_:True] :=
    Module[ {connected,pi,all, size, cliques, c, compsub},
        connected = ToAdjacencyMatrix[graph];
        size = Length[Vertices[graph]];
        connected = connected+IdentityMatrix[size];
        compsub = {};
        all = {};
        For[pi = 1, pi<=size, pi++,
	        AppendTo[all,pi];
   	    	AppendTo[compsub,0];
        ];
        cliques = {};
        c = 0;
        ExtendVersion2[connected,all,0,size, size, Unevaluated[cliques], Unevaluated[c], Unevaluated[compsub]];
        If[ removeSingletons,
            For[pi = 1,pi<=Length[cliques],pi++,
            	If[ Length[cliques[[pi]]]==1,
                	cliques = Drop[cliques,{pi}];
                	pi--;
	            ];
			];
        ];
        cliques
    ];


ExtendVersion2[connected_,oldPre_,neP_,ce_, size_, cliques_, c_, compsub_] :=
    Module[ {new,nod,fixp,newne,newce,i,j,count,pos,p,s,sel,minnod,old,ne, loc, clique},
        ne = neP;
        old = {};
        new = Table[0, {ce}];
        old = oldPre;
        fixp = 0;
        minnod = ce;
        nod = 0;
        For[i = 1,i<=ce&&minnod!=0,i++,
	        p = old[[i]];
   	    	count = 0;
        	For[j = ne+1, j<=ce && count<minnod, j++,
	        	If[ connected[[p,old[[j]] ]]==0,
    	        	count++;
	        	    pos = j;
    	    	];
			];
	        If[ count<minnod,
    	        fixp = p;
        	    minnod = count;
           		If[ i<=ne,
                	s = pos,
	                s = i;
	                nod = 1;
	            ];
	        ];
        ];
        For[nod=minnod+nod, nod>=1 ,nod--,
	        p = old[[s]];
	        old[[s]] = old[[ne+1]];
	        old[[ne+1]] = p;
    	    sel = old[[ne+1]];
        	newne = 0;
	        For[i = 1,i<=ne,i++,
    		    If[ connected[[sel,old[[i]] ]]==1,
            		newne++;
		            new[[newne]] = old[[i]];
       			];
			];
       		newce = newne;
        	For[i = ne+2, i<=ce, i++,
				If[ connected[[sel,old[[i]] ]]==1,
			    	newce++;
				    new[[newce]] = old[[i]];
				];
			];
        	c++;
        	compsub[[c]] = sel;
        	If[ newce==0,
            	clique = {};
            	For[loc = 1,loc<=c,loc++,
            		AppendTo[clique,compsub[[loc]] ];
				];
	            AppendTo[cliques,clique];,
    	        If[ newne<newce,
        	        ExtendVersion2[connected,new,newne,newce, size,Unevaluated[cliques], Unevaluated[c], Unevaluated[compsub]];
            	];
        	];
        	c--;
        	ne = ne+1;
        	If[ nod>1,
            	s = ne+1;
            	While[connected[[fixp,old[[s]] ]]==1,
            		s++;
            	];
	        ];
        ];
    ];
    
(* this is a modified form of the version in Combinatorica *)
Options[DDGetEdgeLabels] = {DefaultEdgeLabelFunction -> (#[[1]]&), EdgeLabelMap->{}};
DDGetEdgeLabels[g_Graph, opts : OptionsPattern[]] :=
	Module[{l},
		l = MapIndexed[
			If[Position[#1, _[EdgeLabel, _]] != {}, 
				EdgeLabel /. #1, 
				OptionValue[DefaultEdgeLabelFunction][#2]
			] &,
			Map[Rest, Edges[g, All]]
		];
		l /. OptionValue[EdgeLabelMap]
	]

(* this is a modified form of the version in Combinatorica *)
Options[DDGetVertexLabels] = {DefaultVertexLabelFunction -> (#[[1]]&)};
DDGetVertexLabels[g_Graph, opts : OptionsPattern[]] := 
	MapIndexed[
		If[Position[#1, _[VertexLabel, _]] != {}, 
			VertexLabel /. #1, 
    		OptionValue[DefaultVertexLabelFunction][#2]
    	] &,
  		Map[Rest, Vertices[g, All]]
  ]

(* end graph functions *)

(* begin drawing functions *)
Hexagon[{x_,y_}] :=
    Polygon[{{x-1/16, y+1/8},
             {x+1/16, y+1/8},
             {x+3/16, y},
             {x+1/16, y-1/8},
             {x-1/16, y-1/8},
             {x-3/16, y}}];


SquarePts[{x_,y_}] :=
    Polygon[{{x-1/8, y+1/8},
             {x+1/8, y+1/8},
             {x+1/8, y-1/8},
             {x-1/9, y-1/8}}];

ShowGr[graph_Graph] :=
    ShowGraph[graph, PlotRange->All];

Options[PlotGr] = {ArrowheadSize->Automatic,
				   Thickness -> Thick
				   };
Options[PlotGr] = Union[Options[PlotGr], Options[GraphPlot]];
PlotGr[graph_List, d_List: {}, opts:OptionsPattern[]] :=  
	Module[{plotopts, vrf, erf, optsweallowchanged, dirs, ec, arsize, thick},
		arsize = OptionValue[ArrowheadSize];
		thick = OptionValue[Thickness];
   (* list should have one entry for each edge, 
   "True" if the edges is directed, "False" otherwise. 
   We probably should do more sanity checking here, but we don't, 
   for now *)
   		dirs = d;
   (* EdgeCount vriable. This is used by the edge drawing function, 
   to index into the dirs list. *)
   		ec = 1;
   (* If we don't get a list of directions, 
   we assume all links are directed *)   
		If[Length[dirs] == 0, 
			dirs = Map[True &, graph]
		];
   		optsweallowchanged = 
   			Cases[
    			Options[GraphPlot], 
    			(h_[l_,r_]) /; 
     				(!MatchQ[l, VertexLabeling] && 
        			 !MatchQ[l, VertexRenderingFunction])
   			];
		plotopts = FilterRules[{opts}, optsweallowchanged];
   		vrf = 
    		Which[
      			MatchQ[#2, _VariableN],
      			{White, EdgeForm[{thick, Black}], Disk[#, 1/8], Black, Text[Style[extractLabel[#2], 20], #]},
    	  		MatchQ[#2, _ConstantN],
      			{White, EdgeForm[{thick, Black}], Disk[#, 1/8], Disk[#, 1/9], Black, Text[Style[extractLabel[#2], 20], #]},
	      		MatchQ[#2, _FunctionN] || MatchQ[#2, _ConstraintN],
    	  		{White, EdgeForm[{thick, Red}], Hexagon[#], Red, Text[Style[extractLabel[#2],20], #]},
      			MatchQ[#2, _IndexN],
      			{White, EdgeForm[{thick, Blue}], SquarePts[#], Blue, Text[Style[extractLabel[#2], 20], #]},
    	  		True,
      			{White, EdgeForm[{thick, Black}], Disk[#, 1/16], Pink, Text[Style[#2, 20], #]}
      		] &;
   		erf = 
   			(If[dirs[[Mod[(ec++) - 1, Length[dirs]] + 1]], 
   				{Black, Arrowheads[arsize], thick, Arrow[#1, .15], Text[Style[#3,20,Bold],Mean[#]+{.05,0}]}, 
   				{Black,thick, Line[#1],        Text[Style[#3,20,Bold],Mean[#]+{.05,0}]},
   				{Red,  thick,  Line[#1],        Text[Style[#3,20,Bold],Mean[#]+{.05,0}]}
   			 ]) &;
   		GraphPlot[
    		graph,
    		Union[
    			plotopts, 
    			{VertexLabeling -> True,
    			 EdgeLabeling -> True, 
      			 VertexRenderingFunction -> vrf, 
      			 EdgeRenderingFunction -> erf}
      		]
    	]
   ];

Options[PrettyPrint] = Options[PlotGr];
PrettyPrint[graph_Graph, opts:OptionsPattern[]] :=
   Module[{gr} ,
   	gr = PreprocessGraph[graph];
	PlotGr[
  		ConvertGraph[gr, DefaultEdgeLabelFunction -> ("" &), EdgeLabelMap -> EdgeNameMap],
   		Map[(EdgeDirection /. Rest[#]) &, Edges[gr, All]],
   		opts
    ]];
PrettyPrint[graph_List, opts:OptionsPattern[]] := PlotGr[graph, opts]; 
PrintGraphToFile[gr_, name_String, fileType_String,dir_:"/home/todd/Desktop/"] := Export[dir <> name <> "." <> fileType, PrettyPrint[gr,ImageSize -> 1000]]
(* end drawing functions *)

(* begin graph-dd-specific functions *)
PreprocessGraph[graph_Graph] :=
    Module[ {localgraph, labels, newlabels, same, e, k,dirs},
        localgraph = graph;
        labels = Sort[MapIndexed[{#1,#2[[1]]}&, DDGetVertexLabels[localgraph]]];
        labels = Map[{Head[#1[[1]]][#1[[1,1]]], #1[[2]]}&, labels];
        newlabels = {};
        While[Length[labels]!= 0,
        	same  = Cases[labels, {labels[[1,1]], _}];
        	labels = Complement[labels, same];
        	If[ Length[same] > 1,
            	newlabels = Union[newlabels, MapIndexed[{#1[[2]], Append[#1[[1]],#2[[1]] ]}&, same] ],
          		newlabels = Union[newlabels, {{same[[1,2]], same[[1,1]]}}]
        	];
        ];
        newlabels = Map[#1[[2]]&, newlabels];
        localgraph = SetGraphOptions[localgraph, 
        	ListToSequence[
        		Union[Options[localgraph],
        		      {DepDiagPreProc->True,Type->Directed,EdgeDirection->True}]]];
        dirs = Map[
        	Module[{d,l},
        		d = (EdgeDirection /. Rest[#]);
        		If[MatchQ[d,EdgeDirection],
        			l = (EdgeLabel /. Rest[#]);	
        			d = l /. EdgeDirectednessMap;
        			d = If[d,d,d,True]
        		];
        		d
        	] &, 
        	Edges[localgraph, All]];
        localgraph = SetVertexLabels[localgraph, newlabels];
        localgraph = Graph[ 
        		MapThread[{#1[[1]],ListToSequence[Union[Rest[#1],{EdgeDirection->#2}]]}&,{Edges[localgraph,All],dirs}], 
        		Vertices[localgraph,All],
        		ListToSequence[Union[Options[graph],Options[localgraph]]]
        	]
(* why have I done this? *)
(* couldn't come up with a reason, so i'm disabling it. 
        If[ !MemberQ[GraphOptions[localgraph], EdgeDirection-> False],
            For[k = 1, k<= M[localgraph], k++, 
           		e = Edges[localgraph][[k]];
            	If[ DDGetEdgeLabels[localgraph][[k]] == UnCondDep && !MemberQ[Edges[localgraph], Reverse[e]],
                	localgraph = AddEdge[localgraph, {Reverse[e], EdgeLabel->UnCondDep}];
            	];

            ];
        ];
*)
        
    ];


AddLabeledVertex[graph_Graph, label_, type_] :=
    Module[ {vlabels, matches,rgraph},
        rgraph = PreprocessGraph[graph];
        vlabels = DDGetVertexLabels[rgraph];
        matches = Cases[vlabels, x_type/; x[[1]]==label];
        rgraph = SetGraphOptions[AddVertex[rgraph], ListToSequence[Options[graph]]];
        Which[
        	Length[matches] > 1,
        	rgraph = SetVertexLabels[rgraph, Append[vlabels, type[label, Last[Last[matches]]+1]]],
        	Length[matches] > 0,
        	rgraph = SetVertexLabels[rgraph, Append[vlabels, type[label]]];
        	rgraph = PreprocessGraph[rgraph],
        	Length[matches] == 0,
        	rgraph = SetVertexLabels[rgraph, Append[vlabels, type[label]]]
        ];
        rgraph
    ];


FactorizeGraph[graph_Graph] :=
    Module[ {ed,el,vl,i,j,k,edge, rgraph, usgraph, uscliques, 
    	     par1, par,parIndex, ei, end1, end2, indexForward, indexBackward},
        rgraph = PreprocessGraph[graph]; 
        ed = Edges[rgraph];
        el = DDGetEdgeLabels[rgraph];
        vl = DDGetVertexLabels[rgraph];
(* what does this code do? no one knows! *)        
        For[i = 1,i<=Length[ed],i++,
	        edge = ed[[i]];
    	    If[ MatchQ[el[[i]],CondDep] && 
        	   (MatchQ[vl[[edge[[1]]]],_VariableN] || MatchQ[vl[[edge[[1]]]],_ConstantN]) &&
               (MatchQ[vl[[edge[[2]]]],_VariableN] || MatchQ[vl[[edge[[2]]]],_ConstantN]),
	            par1 = ParentsOf[rgraph,edge[[2]]];
            	par = {};
	            For[k = 1,k<=Length[par1],k++,
    		        ei = EdgeIndex[rgraph,par1[[k]],edge[[2]]];
            		If[ MatchQ[el[[ei]],CondDep] &&
              		   (MatchQ[vl[[par1[[k]]]],_VariableN] || MatchQ[vl[[par1[[k]]]],_ConstantN]),
		                AppendTo[par,par1[[k]]];
        		    ];
            	];
            	If[ Length[par]>0,
                	rgraph = AddLabeledVertex[rgraph, SymbolNotInGraph[graph,"Factor"], FunctionN];                	
            	    parIndex = V[rgraph];
                	For[j = 1,j<=Length[par],j++,
                		rgraph = DeleteEdge[rgraph,{par[[j]],edge[[2]]}];
	                	rgraph = AddEdge[rgraph,{{par[[j]],parIndex},EdgeDirection->True,EdgeLabel->CondDep}];
    	            ];
        	        rgraph = AddEdge[rgraph,{{parIndex,edge[[2]]},EdgeDirection->True,EdgeLabel->CondDep}];
            	    i = 0;
                	ed = Edges[rgraph];
	                el = DDGetEdgeLabels[rgraph];
    	            vl = DDGetVertexLabels[rgraph];
        	    ];
        	];
        ];
(* begin processing non-conditional dependence ('u', or undirected)  links *)
        ed = Edges[rgraph];
        el = DDGetEdgeLabels[rgraph];
        vl = DDGetVertexLabels[rgraph];
        usgraph = EmptyGraph[Length[vl]];
        For[i = 1,i<=Length[el],i++,
	        edge = ed[[i]];
    	    If[ MatchQ[el[[i]],UnCondDep] && 
        	   (MatchQ[vl[[edge[[1]]]],_VariableN] || MatchQ[vl[[edge[[1]]]],_ConstantN]) &&
          	   (MatchQ[vl[[edge[[2]]]],_VariableN] || MatchQ[vl[[edge[[2]]]],_ConstantN]) &&
        	   !MemberQ[Edges[usgraph], Reverse[edge]],
            	usgraph = AddEdge[usgraph, {{edge[[1]],edge[[2]]}, EdgeDirection->False}];
        	];
        ];
        uscliques = AllMaximalCliques[usgraph];
        rgraph = SetGraphOptions[rgraph, ListToSequence[Union[Options[rgraph],{EdgeDirection->True}]]];
        For[i = 1,i<=Length[uscliques],i++,
        	rgraph = AddLabeledVertex[rgraph, SymbolNotInGraph[rgraph, "Factor"], FunctionN];
	        parIndex = V[rgraph];
    	    For[j = 1,j<=Length[uscliques[[i]]],j++,
    	    	rgraph = AddEdge[rgraph, {{parIndex,uscliques[[i]][[j]]},EdgeDirection->True,EdgeLabel->FactorEdge}];
                For[k = j+1,k<=Length[uscliques[[i]]],k++,
	                end1 = uscliques[[i, k]];
                    end2 = uscliques[[i, j]];
                    indexForward  = EdgeIndex[rgraph, end1,end2];
                    If[ indexForward  >0 && MatchQ[el[[indexForward]],UnCondDep],
    	                rgraph = DeleteEdge[rgraph,{end1,end2}];
                    ];
                    indexBackward = EdgeIndex[rgraph, end2, end1];
                    If[ indexBackward > 0 && MatchQ[el[[indexBackward]],UnCondDep],
                        rgraph = DeleteEdge[rgraph,{end2,end1}];
                    ];
                ];
            ];
        ];
        rgraph
    ];    
    
ExplicitPriors[graph_Graph] :=
	Module[{checkparents, setprior, needPriors, labels, fgr},
		fgr = FactorizeGraph[graph];
		labels = DDGetVertexLabels[fgr];
		checkparents = 
			Function[{vertexid},
				Module[{parents, label},
					label = labels[[vertexid]];
					If[MatchQ[label, _VariableN] ||
						MatchQ[label, _ConstantN],
						parents = ParentsOf[fgr, vertexid];
						MatchQ[Cases[Map[labels[[#]]&,parents],_FunctionN],{}],						
						False
					]
				]				
			];
		setprior = 
			Function[{tf,id},
				If[tf,
				  	fgr = AddLabeledVertex[fgr, SymbolNotInGraph[fgr, "Prior"], FunctionN];
					fgr = AddEdge[fgr, {{V[fgr],id[[1]]}, EdgeDirection->True,EdgeLabel->CondDep}];			
				]
			];
		needPriors = Map[checkparents, Range[V[fgr]]];
		MapIndexed[setprior,needPriors];
		fgr
	];
	
FullySpecifyGraph[graph_Graph] := ExplicitPriors[FactorizeGraph[PreprocessGraph[graph]]];
(* end graph-dd-specific functions *)




MakeSymbolString[sym_, variables_] :=
    MakeString[sym, GetIndices[variables, sym]];


GetIndices[list_, target_] :=
    Flatten[Map[Indices,Select[list, Label[#] == target&]]];


MakeString[sym_, inds_] :=
    Module[ {string,kinds, start, end,j,l},
        l = extractLabel[sym];
        string = ToString[FullForm[l]];
        kinds = Map[extractLabel, inds];
        If[ Length[kinds] > 0,
     		start = ToString[INDEXTYPE]<>"[";
        	end = "]";
            string = start <> string <>  "," <> ListToCommaSeparatedString[kinds] <> end;
        ];
        string
    ];

MakeIndexList[sym_, variables_, argindices_, indexDefs_] :=
    Module[ {symInds,int,symIndsEL, list,id,i},
    	(*** HAXXX gtg 3/36/8 -- can i hack in index arguments? ***)
    	If[MatchQ[Head[sym],IndexN],
    		symInds = {};,
	        symInds = Indices[Cases[variables, x_ /;Label[x]==sym][[1]]];
    	];
        symIndsEL = Map[extractLabel, symInds];
        int = Cases[argindices, x_/;MemberQ[symIndsEL, extractLabel[x]]];
        list = {};
        For[i = 1,i<=Length[symInds],i++,
	        If[ MemberQ[argindices, x_/;extractLabel[x]==extractLabel[symInds[[i]]]],
    	        id = Cases[indexDefs, x_DDIndex /;Label[x]==symInds[[i]]][[1]];
        	    list = Append[list, {IndexMinName[id], IndexMaxName[id]}],
            	list = Append[list, extractLabel[symInds[[i]]]]
        	];
		];
        list
    ];
MakeArgString[sym_,variables_, argindices_, indexDefs_] :=
    Module[ {symInds,symIndsEL,int,list,i,id},
    	(*** HAXXX gtg 3/36/8 -- can i hack in index arguments? ***)
    	If[MatchQ[Head[sym],IndexN],
    		symInds = {};,
	        symInds = Indices[Cases[variables, x_ /;Label[x]==sym][[1]]];
    	];
        symIndsEL = Map[extractLabel, symInds];
        int = Cases[argindices, x_/;MemberQ[symIndsEL, extractLabel[x]]];
        If[ int=={},
            MakeSymbolString[sym, variables],
            list = MakeIndexList[sym,variables,argindices,indexDefs];
            "DDList[" <> ToString[FullForm[extractLabel[sym]]] <> "," <> ToString[FullForm[list]]<> "]"
        ]
    ];


ParseGraph[graph_Graph,addPriors_:False, argumentOrdersArg_:{},indexOrdersArg_:{}] :=
    Module[ {factoredGraph, vlabels, elabels, ret, check, functions, 
    	variables, parents, children, arguments, givens, gates, edges,
    	 fcount, dcount, edgeNumber, fnode, dnode, string, power, vars,
    	  exists, exp, pow, vindices, findices, indices, bar, i, j, k, 
    	  indexcons, relindices, return, list, fargindices, indicesNH, 
    	  ind, allIt,it,elabel,plabel,label,label2ind,getArgOrder, getIndOrder,
    	  indexOrders, argumentOrders,argo,indo,m,funcExps,varlab2indices,indlab2DDI,
    	  subset,indz2indconz,fvinds,fits,indconz2pow},

        factoredGraph = If[addPriors,FullySpecifyGraph[graph],FactorizeGraph[PreprocessGraph[graph]]];
                
        {indexOrders, argumentOrders} = CompleteOrderings[indexOrdersArg,argumentOrdersArg,factoredGraph];
        
        getArgOrder[num_] := Module[{cases},
        		cases = Cases[argumentOrders, ao_/;Length[ao]>0 && ao[[1]]==num];
        		If[cases!={},
        			cases[[1,2]],
        			{},{}
        		]
        	];
        	
        getIndOrder[num_] := Module[{cases},
        		cases = Cases[indexOrders, io_/;Length[io]>0 && io[[1]]==num];
        		If[cases!={},
        			cases[[1,2]],
        			{},{}
        		]
        	];
        
        vlabels = DDGetVertexLabels[factoredGraph];
        elabels = DDGetEdgeLabels[factoredGraph];
		label2ind[l_] := Module[{p},
				p = Position[vlabels, l]; 
				If[Length[p] > 0,
					p[[1,1]],
					Print["um, what? <", l, ">", p, vlabels];-1,	
					Print["hm, what? <", l, ">", p, vlabels];-1
				]
			];
        edges = Edges[factoredGraph];
        functions = {};
        variables  = {};
        indexcons = {};
        For[i = 1, i <= V[factoredGraph], i++,
	        label = vlabels[[i]];
	        Which[
		        MatchQ[label, _FunctionN],
		        parents = ParentsOfIndices[factoredGraph, i];
		        children = ChildrenOfIndices[factoredGraph, i];
		        arguments = {};
		        givens = {};
		        gates = {};
		        findices = {};
		        fargindices = {};
		        fcount = 0;
		        dcount = 0;
		        For[j = 1, j <= Length[children], j++,
			        edgeNumber = children[[j,2]];
			        If[ MatchQ[elabels[[edgeNumber]], FactorEdge],
			            arguments = Append[arguments, vlabels[[children[[j,1]]]]];
			            fcount++,
			            If[ MatchQ[elabels[[edgeNumber]], CondDep],
			                arguments = Append[arguments, vlabels[[children[[j,1]]]]];
			                dcount++
			            ];
			        ];
		        ];
		        For[j = 1, j <= Length[parents], j++,
		        	edgeNumber = parents[[j,2]];
		        	elabel = elabels[[edgeNumber]];
		        	plabel = vlabels[[parents[[j,1]]]];
		        	Which[
			       		MatchQ[elabel, Gate],
		    	    	gates = Append[gates, plabel],
		       			MatchQ[elabel, CondDep],
			       	 	givens = Append[givens, plabel];
		    	   	 	dcount++,
		        		MatchQ[elabel, IndexL],
		        		findices = Append[findices, plabel],
		        		MatchQ[elabel, ArgIndex],
		        		fargindices = Append[fargindices, plabel]
		        	];
		        ];
		        fnode = (fcount > 0 && dcount== 0);
		        dnode = (fcount==0 && dcount== (Length[arguments] + Length[givens]) && (Length[givens]==0 || Length[arguments] > 0));
		        
		        (* getting the orderings right. *)
		        argo = getArgOrder[i];
		        indo = getIndOrder[i];

		        arguments = Sort[arguments, CompareByList[label2ind[#1],label2ind[#2],argo]&];
				givens = Sort[givens, CompareByList[label2ind[#1],label2ind[#2],argo]&];	
				findices = Sort[findices, CompareByList[label2ind[#1],label2ind[#2],indo]&];	
(*		        Print[arguments, givens, findices, fargindices, argo];		        *)
		        AppendTo[functions , Func[i, label, arguments, givens, gates, fnode, dnode, findices, fargindices]],
		
		        MatchQ[label, _VariableN] || MatchQ[label, _ConstantN],
		        exists = {};
		        vindices = {};
		        parents = ParentsOfIndices[factoredGraph, i];
		        For[j = 1, j<= Length[parents], j++,
			        If[ MatchQ[elabels[[ parents[[j,2]] ]], Existence],
			            AppendTo[exists, vlabels[[parents[[j,1]]] ] ];,
			            If[ MatchQ[elabels[[ parents[[j,2]] ]], IndexL],
			                AppendTo[vindices, vlabels[[parents[[j,1]] ]] ];
			            ];
			        ];
				];
				indo = getIndOrder[i];
				vindices = Sort[vindices, CompareByList[label2ind[#1],label2ind[#2],indo]&];					
		        AppendTo[variables, Var[i, label, exists, vindices]],
		
		        MatchQ[label, _ConstraintN] ,
		        relindices = {};
		        children = ChildrenOfIndices[factoredGraph, i];
		        For[j = 1, j<= Length[children], j++,
			        If[ MatchQ[vlabels[[children[[j,1]] ]], _IndexN] &&
			        	MatchQ[elabels[[children[[j,2]] ]], IndexCons],
			            AppendTo[relindices, vlabels[[children[[j,1]]] ] ];
			        ];
				];
		        argo = getArgOrder[i];				
		        relindices = Sort[relindices, CompareByList[label2ind[#1],label2ind[#2],argo]&];
		        AppendTo[indexcons, IndexConstraint[i, label, relindices]];
		    ];
        ];
        indices = 
        	Fold[Union,{},
        		Union[Map[Indices, variables],
    	    		 Map[Indices, functions]]];

        bar = Map[{SymbolNotInGraph[factoredGraph, ToString[#]<>"min"],SymbolNotInGraph[factoredGraph, ToString[#]<>"max"]}&, Map[extractLabel, indices]];
        ind = MapThread[ReplacePart[Flatten[{Position[vlabels, #1][[1,1]] ,#1, #2}],0->DDIndex]&, {indices, bar}];

        m = Table[0,{Length[ind]},{Length[ind]}];

        funcExps = MapIndexed[MakeFullFunc[#2[[1]], functions, variables, ind]&, functions,1];
        varlab2indices = 
        	Function[{vlab},
		        (*** HAXXX gtg 3/36/8 -- can i hack in index arguments? ***)
		    	If[MatchQ[Head[vlab],IndexN],
		    		{},
			        Indices[Cases[variables, x_Var /;Label[x]==vlab ][[1]]]
		    	]
        	];
        indlab2DDI = Function[{ilab},Cases[ind, x_DDIndex/;Label[x]==ilab][[1]]];
        subset = Function[{smaller, larger}, Union[Map[MemberQ[larger, #]&,smaller]]=={True}];
        indz2indconz = Function[{indz},Cases[indexcons, x_IndexConstraint/;subset[Indexed[x],indz]]];
        indconz2pow = Function[{conz}, Times[Evaluate[ListToSequence[Map[MakeIndCond[#]&,conz]]]]];
        fvinds = Map[Union[Complement[Union[ListToSequence[Map[varlab2indices, Vars[#]]]],ArgIndices[#]], Indices[#]]&,functions];
        funcExps = MapThread[#1^indconz2pow[#2]&, {funcExps,Map[indz2indconz,fvinds]}];
        fits = Map[DDIndicesToIterators,Map[indlab2DDI,fvinds, {2}]];
        return = Times[ListToSequence[MapThread[Product[#1, Evaluate[ReplacePart[#2, 0->Sequence]]]&,{funcExps,fits}]]];
        Quiet[Parse[Flatten[return], functions, variables, indexcons, ind],{Part::pspec,Rule::argr}]
    ];

DDIndicesToIterators[ind_List] :=
    Union[Map[{extractLabel[Label[#]], IndexMinName[#], IndexMaxName[#]}&, ind]];
    

MakeIndCond[arg_IndexConstraint] :=
    ToExpression[
    ToString[FullForm[extractLabel[Label[arg]]]]<>"["  <> 
    Fold[ 
    (#1 <> "," <> ToString[FullForm[extractLabel[#2]]])&, 
    ToString[FullForm[extractLabel[First[Indexed[arg]]]]], Rest[Indexed[arg]]]
    <> "]"];


MakeFullFunc[i_, functions_, variables_, indices_] :=
    Module[ {func, arguments, givens, gates, string, j, power, vars, k, exp, pow, argsString, argStrings, givStrings, givsString, arginds},
        func = functions[[i]];
        arguments = Args[func];
        givens = Givens[func];
        gates = Gates[func];
        arginds = ArgIndices[func];
        string = MakeSymbolString[Label[func], functions];
(*        string = string <>"[";*)
        If[ IsDNode[func] || IsFNode[func],
            argStrings = Map[MakeArgString[#,variables, arginds, indices]&,arguments];
            argsString = ListToCommaSeparatedString[argStrings];
(*            string = string <> argsString;*)
            If[ IsDNode[func]&&Length[givens] > 0,
            (*the "," here would be a "|" for conditional probability, if Mathematica allowed for that *)
                (*string = string <> ",";*)
                givStrings = Map[MakeArgString[#,variables, arginds, indices]&, givens];
                givsString = ListToCommaSeparatedString[givStrings];,
                givsString = "";
(*                string = string <> givsString;*)
            ];
        ];
(*        string = string <> "]";*)
		If[IsDNode[func],
			If[Length[arguments] > 1,
				string = "PDF[" <> string <> "[" <> givsString <> "]," <> "{" <> argsString <> "}" <> "]";,
				string = "PDF[" <> string <> "[" <> givsString <> "]," <> argsString <> "]";
			];,
			If[IsFNode[func],
				string = string <> "[" <> argsString <> "]";
			];
		];
        power = 1;
        For[j = 1, j <= Length[gates], j++,
        	power = power * Heaviside[ToExpression[MakeSymbolString[gates[[j]], variables]]];
		];
        For[ j = 1, j<= Length[arguments], j++, 
			vars = Flatten[Map[Part[#,3]&,Select[variables, Part[#,2]== arguments[[j]]&]]];
			For[k = 1, k<= Length[vars], k++,
				power = power * Heaviside[ToExpression[MakeSymbolString[ vars[[k]], variables]] ];
			];
		];
        exp = Quiet[ToExpression[string],{Part::"pspec",Rule::"argr"}];
        pow = exp ^ power
    ];


PsiFunction[graph_Graph] :=
    AlgebraicExpression[ParseGraph[graph]];


IndCl[graph_Graph] :=
    Module[ {ngraph,b,i,j,ni,nj},
        ngraph = EmptyGraph[V[graph]];
        Map[Set[ngraph , AddEdge[ngraph, {#1, EdgeDirection->False}]]&,
        Fold[Union, {}, Map[Subsets[#, {2}]&,
        Map[ChildrenOf[graph, #]&, 
        Flatten[Position[DDGetVertexLabels[graph], _ConstraintN]]]]]];
        b = AllMaximalCliques[ngraph, False];(*used to be cliques*)
        b = Cases[b, x_/; Length[x]>1||MatchQ[DDGetVertexLabels[graph][[x[[1]]]],_IndexN]];
        For[i = 1,i<=Length[b],i++,
	        ni = Map[extractLabel, Map[DDGetVertexLabels[graph][[#]]&,b[[i]]]];
	        For[j = i+1,j<=Length[b],j++,
		        nj = Map[extractLabel, Map[DDGetVertexLabels[graph][[#]]&,b[[j]]]];
		        If[ Length[Intersection[ni,nj]]>0,
		            b = ReplacePart[b, i->Union[b[[i]],b[[j]]]];
		            b = Delete[b, j];
		            j--;
		        ];
	        ];
        ];
        b
    ];


OrderIndices[graph_Graph,clusterDimensions_:False] :=
    Module[ {i,j,k,p, dimensions, orderedIndices, seenDim, orderedVariables},
        dimensions = Map[Sort, IndCl[graph]];
        orderedIndices = {};
        seenDim = Table[{}, {Length[dimensions]}];
        orderedVariables = Cases[Map[{#, DDGetVertexLabels[graph][[#]]}&, TopologicalSort[graph]], {_,_VariableN}];
        For[i = 1, i <= Length[orderedVariables], i++,
	        p = ParentsOf[graph, orderedVariables[[i,1]]];
    	    For[j = 1, j<= Length[dimensions], j++,
        		For[k = 1, k<= Length[p], k++,
			        If[ MemberQ[dimensions[[j]], p[[k]]],
           			 	seenDim[[j]] = Union[seenDim[[j]], {p[[k]]}];
            			dimensions[[j]] = DeleteCases[dimensions[[j]], p[[k]]];
			            If[ dimensions[[j]] == {},
			            	If[clusterDimensions,
			            		orderedIndices = Append[orderedIndices, Reverse[seenDim[[j]]]];,
            			    	orderedIndices = Join[orderedIndices, Reverse[seenDim[[j]]]];
			            	];
            			];
        			];
				];
			];
        ];
        If[clusterDimensions,
            Map[Map[{#, DDGetVertexLabels[graph][[#]]}&,#]&,orderedIndices],
	        Map[{#, DDGetVertexLabels[graph][[#]]}&,orderedIndices]
        ]
    ];

(* begin code needed by generated programs *)    
(* helper functions used by generated code *)
SetAttributes[DDAdvance, HoldFirst];
DDCollect[n_, indices_] :=
    If[ MatchQ[INDEXTYPE,Part],
        Part[n,ListToSequence[IndListToPSpec[indices]]],
        DDSubCollect[n, indices, {}]
    ];
DDSubCollect[n_, indices_, fixed_] :=
    If[ Length[indices]==0,
        INDEXTYPE[n,fixed/.List->Sequence],
        Module[ {h,r,hold,i,count},
            h = First[indices];
            If[ MatchQ[h, _List],
                count = h[[2]]-h[[1]]+1;
                r = Table[0,{count}];
                For[i = 1,i<=count,i++,
                r[[i]] = DDSubCollect[n, Rest[indices], Append[fixed,i]];
],
                r = DDSubCollect[n, Rest[indices], Append[fixed,h]];
            ];
            r
        ]
    ];

IndListToPSpec[l_List] :=
    Map[If[ MatchQ[Head[#],List],
            Span[#[[1]],#[[2]]],
            #
        ]&,l];
IndListToTableSize[l_List] :=
    Module[ {r},
        r = Cases[Map[If[ MatchQ[Head[#],List],
                          {#[[2]]-#[[1]]+1}
                      ]&,l],_List];
        r
    ];
IndListToTable[l_List] :=
    Table[0,Evaluate[ListToSequence[IndListToTableSize[l]]]];

    
extractLabel[_[x_,y_]] := x;
extractLabel[_[x_]] := x;
		
FirstInListWithName[llist_,nname_]:=
			First[Cases[llist,v_/; MatchQ[extractLabel[Label[v]],nname]]];

CompareByList[a_,b_,ordering_]:=
	Module[{ai,bi},
		ai = Position[ordering, a];
		bi = Position[ordering, b];
		If[ai=={},ai={{Length[ordering]+1}}];
		If[bi=={},bi={{Length[ordering]+1}}];
		ai[[1, 1]] < bi[[1, 1]]
	];	
	
SortByOrder[l_List, o_List] :=
	Map[#[[2]] &, Sort[MapThread[{#1, #2} &, {o, l}]]];	

OrderVariables[vars_List, names_List, fullgr_Graph] := 
	Module[{unvars,order},
			unvars = Map[FirstInListWithName[vars,#]&, names];
			order = If[!UndirectedQ[fullgr] && AcyclicQ[fullgr],
						TopologicalSort[fullgr], 
						Range[V[fullgr]],
						Range[V[fullgr]]
					];
			Sort[unvars, CompareByList[Node[#1],Node[#2],order]&]
	];

GenerateMCMCDependencyDiagram[graph_Graph] := 
	Module[{fullgr, parse,mcmcgraph,uniquevariablenames,uniqueconstantnames,processonevariable,
		    processonepasstwo,vlabels,passoneres,getfirstpassres,processoneconstant,order,constO,
		    propDistNr,propVarNr,acceptVarNr,propDistName,propVarName,acceptVarName,e2nr,e3nr,
		    passoneVARres,passoneCONSTres,indexinorderV,indexinorderO,describemcmc},
		fullgr = ExplicitPriors[FactorizeGraph[PreprocessGraph[graph]]];
		parse = ParseGraph[fullgr];
		describemcmc = {};
(*
order variables
add two time indices
add +1 link between them
For each variable:
	add two copies
	link one time index to each	
	create equality link E1 between them (use 'delta dist. @' conditional dist. for equality links)

	add proposal distribution
	add proposal variable
	add 'future' time index to proposal var.
    add an update choice variable
    link proposal dist. to proposal var
    link past-time version of var as input to proposal dist.
    
	create an update prob. function
		input: 'past'-time version of variable, some past and some current times of markov blanket
		       proposed variable value	
		output:  update choice var
		       
    add equality link E2 between propsal variable, 'future' time of variable
    add equality link E3 between 'past' and 'future' variables (link analagous to E1, only we need a second copy)


   	existence E2 with update choice var
    gate E3 with Dirac of update choice var.		       

	pick out & add a set of canonical indices
	link these to two copies of variable		       
	create a variable for each index
	link 'future' time index to index variables	
   	gate E1 with NotKD taking index, index variable as arugments 	
   	gate E2 with KD taking index, index variable as arugments 	
    add distribution for index variable
    
For each constant: 
    add one copy
  	pick out & add a set of canonical indices
  	link indices to variable
  	insure hooked into appropriate update functions for variables
		*)
		mcmcgraph = EmptyGraph[0,Type->Directed];

		(* add two time indices *)
		mcmcgraph = AddLabeledVertex[mcmcgraph, SymbolNotInGraph[mcmcgraph,"t"], IndexN];
		mcmcgraph = AddLabeledVertex[mcmcgraph, SymbolNotInGraph[mcmcgraph,"t"], IndexN];
		
		(* add +1 link between them *) 
		mcmcgraph = AddLabeledVertex[mcmcgraph, PlusOne, ConstraintN];
		mcmcgraph = AddEdge[mcmcgraph, {{V[mcmcgraph],1}, EdgeLabel->IndexCons,EdgeDirection->True}];
		mcmcgraph = AddEdge[mcmcgraph, {{V[mcmcgraph],2}, EdgeLabel->IndexCons,EdgeDirection->True}];

		vlabels = DDGetVertexLabels[fullgr];
		uniquevariablenames = 
			Union[Flatten[Map[extractLabel, 
				Cases[vlabels, _VariableN]]]];
  		uniqueconstantnames = 
			Union[Flatten[Map[extractLabel, 
				Cases[vlabels, _ConstantN]]]];
		
		order = OrderVariables[Vars[parse], uniquevariablenames, fullgr];		
		constO = Map[FirstInListWithName[Vars[parse],#]&,uniqueconstantnames];
		order = Join[order,constO];
		

		(* For each variable: *)
		processonevariable = Function[{name},
			Module[{var,secondvarn,indices,pickvars,ration,fname,vard,
				nAcceptVarName,nAcceptVarNr,nAcceptVarDist,nAcceptDistNr},
				
				var = FirstInListWithName[Vars[parse],name];
				
				vard = MCMCVariableDescriptor[];
				vard = SetVariableName[vard,name];
				vard = SetVariableDesc[vard,var];				
				
				(* add two copies *)
				mcmcgraph = AddLabeledVertex[mcmcgraph, name, VariableN];
				mcmcgraph = AddLabeledVertex[mcmcgraph, name, VariableN];				
				secondvarn = V[mcmcgraph];
				
				vard = SetVariableNr[vard,secondvarn];
				vard = SetVariableOldNode[vard,name,secondvarn-1];
				vard = SetVariableNewNode[vard,name,secondvarn ];
				
				(* add one time index to each *)
				mcmcgraph = AddEdge[mcmcgraph, {{1,secondvarn-1},EdgeLabel->IndexL,EdgeDirection->True}];
				mcmcgraph = AddEdge[mcmcgraph, {{2,secondvarn},  EdgeLabel->IndexL,EdgeDirection->True}];
				(* add equality link E1 between them *)
(*				If[Length[Indices[var]]>0,
					
			    ];				*)
				
				(* add proposal distribution
                   add proposal variable
    			   add an update choice variable  *)
				propDistName = SymbolNotInGraph[mcmcgraph, "Proposal"<>ToString[name]];
				propVarName = SymbolNotInGraph[mcmcgraph, "proposed"<>ToString[name]];
				acceptVarName = SymbolNotInGraph[mcmcgraph, "accept"<>ToString[name]];
				nAcceptVarName = SymbolNotInGraph[mcmcgraph, "nAccept"<>ToString[name]];
				
				mcmcgraph = AddLabeledVertex[mcmcgraph, propDistName, FunctionN];
				propDistNr = V[mcmcgraph]; 
    			mcmcgraph = AddLabeledVertex[mcmcgraph, propVarName, VariableN];
				propVarNr = V[mcmcgraph];
				vard = SetVariablePropose[vard,propVarName,propVarNr,propDistName,propDistNr];				

    			mcmcgraph = AddLabeledVertex[mcmcgraph, acceptVarName, VariableN];
				acceptVarNr = V[mcmcgraph];				
				
				mcmcgraph = AddLabeledVertex[mcmcgraph, nAcceptVarName, VariableN];
				nAcceptVarNr = V[mcmcgraph];				
				
				mcmcgraph = AddLabeledVertex[mcmcgraph, IndicOther,FunctionN];
				nAcceptDistNr = V[mcmcgraph];
                mcmcgraph = AddEdge[mcmcgraph, {{acceptVarNr,nAcceptDistNr},EdgeLabel->CondDep}];
                mcmcgraph = AddEdge[mcmcgraph, {{nAcceptDistNr,nAcceptVarNr},EdgeLabel->CondDep}];				

				(* link proposal dist. to proposal var.  *)
				mcmcgraph = AddEdge[mcmcgraph, {{propDistNr, propVarNr},EdgeLabel->CondDep}];
				mcmcgraph = AddEdge[mcmcgraph, {{2, propVarNr},EdgeLabel->IndexL}];
				mcmcgraph = AddEdge[mcmcgraph, {{2, acceptVarNr},EdgeLabel->IndexL}];				
				mcmcgraph = AddEdge[mcmcgraph, {{2, nAcceptVarNr},EdgeLabel->IndexL}];								
				(* link past-time version of var as input to proposal dist.*)
				mcmcgraph = AddEdge[mcmcgraph, {{secondvarn-1,propDistNr},EdgeLabel->CondDep}];

				(* create an update prob. function *)
				fname = SymbolNotInGraph[mcmcgraph, "F"<>ToString[name]];
				mcmcgraph = AddLabeledVertex[mcmcgraph, fname, FunctionN];
				ration = V[mcmcgraph];
				
				vard = SetVariableAccept[vard,acceptVarName,acceptVarNr,fname,ration];
				vard = SetVariableNotAccept[vard, nAcceptVarName, nAcceptVarNr,IndicOther,nAcceptDistNr];
				
				mcmcgraph = AddEdge[mcmcgraph, {{secondvarn-1,ration},EdgeLabel->CondDep,EdgeDirection->True}];
				mcmcgraph = AddEdge[mcmcgraph, {{propVarNr,ration},EdgeLabel->CondDep,EdgeDirection->True}];
				mcmcgraph = AddEdge[mcmcgraph, {{ration,acceptVarNr},EdgeLabel->CondDep,EdgeDirection->True}];
				
   			    (* add equality link E2 between propsal variable, 'future' time of variable *)
				mcmcgraph = AddLabeledVertex[mcmcgraph, DeltaCondDist,FunctionN];
				e2nr = V[mcmcgraph];
				mcmcgraph = AddEdge[mcmcgraph, {{propVarNr,e2nr}, EdgeLabel->CondDep,EdgeDirection->True}];
				mcmcgraph = AddEdge[mcmcgraph, {{e2nr,secondvarn},EdgeLabel->CondDep,EdgeDirection->True}];

                (* add equality link E3 between 'past' and 'future' variables (link analagous to E1, only we need a second copy) *)
				mcmcgraph = AddLabeledVertex[mcmcgraph, DeltaCondDist, FunctionN];
				e3nr = V[mcmcgraph];
				mcmcgraph = AddEdge[mcmcgraph, {{secondvarn-1,e3nr}, EdgeLabel->CondDep,EdgeDirection->True}];
				mcmcgraph = AddEdge[mcmcgraph, {{e3nr,secondvarn},   EdgeLabel->CondDep,EdgeDirection->True}];
				
            	(* gate E2 with Heaviside choice var *)
(*                mcmcgraph = AddLabeledVertex[mcmcgraph, Heaviside, FunctionN];
				mcmcgraph = AddEdge[mcmcgraph, {{},  EdgeLabel->FactorEdge,EdgeDirection->True}];*)
				mcmcgraph = AddEdge[mcmcgraph, {{acceptVarNr,e2nr},EdgeLabel->Gate,EdgeDirection->True}];
(*            	mcmcgraph = AddEdge[mcmcgraph, {{acceptVarNr, e2nr}, EdgeLabel->Existence, EdgeDirection->True}];*)
            	
                (* gate E3 with Dirac of update choice var.*) 
(*   	            mcmcgraph = AddLabeledVertex[mcmcgraph, Dirac, FunctionN];
				mcmcgraph = AddEdge[mcmcgraph, {{V[mcmcgraph],acceptVarNr},  EdgeLabel->FactorEdge,EdgeDirection->True}];*)
				mcmcgraph = AddEdge[mcmcgraph, {{nAcceptVarNr,e3nr},EdgeLabel->Gate,EdgeDirection->True}];
				
				(* pick out canonical set of indices *)
				indices = Map[extractLabel, Indices[var]];

				(* for each canonical index ... *)
				pickvars = Map[
					Module[{constraintn,v,indName,oldTIndexVarNr,vdist,vnr,indicName,nIndicName,indicNr,nIndicNr},
						mcmcgraph = AddLabeledVertex[mcmcgraph, DeltaCondDist, FunctionN];
			
						constraintn = V[mcmcgraph];
						mcmcgraph = AddEdge[mcmcgraph, {{V[mcmcgraph],secondvarn},  EdgeLabel->CondDep,EdgeDirection->True}];
						mcmcgraph = AddEdge[mcmcgraph, {{secondvarn-1,V[mcmcgraph]},EdgeLabel->CondDep,EdgeDirection->True}];
						
						(* add old-t version of the index, link it to the old-t version of the variable and argindex it 
						   to the ratio node and the proposal function *)
						indName = SymbolNotInGraph[mcmcgraph,ToString[#]];
						mcmcgraph = AddLabeledVertex[mcmcgraph,indName,IndexN];
						oldTIndexVarNr = V[mcmcgraph];												

						Set[mcmcgraph, AddEdge[mcmcgraph, {{oldTIndexVarNr,secondvarn-1},EdgeLabel->IndexL,EdgeDirection->True}]];
						Set[mcmcgraph, AddEdge[mcmcgraph, {{oldTIndexVarNr,ration},EdgeLabel->ArgIndex,EdgeDirection->True}]];
						Set[mcmcgraph, AddEdge[mcmcgraph, {{oldTIndexVarNr,propDistNr},EdgeLabel->ArgIndex,EdgeDirection->True}]];
						(* add new-t version of the index, link it to new-t version of the variable 
						Set[mcmcgraph, AddLabeledVertex[mcmcgraph,SymbolNotInGraph[mcmcgraph,ToString[#]],IndexN]];
						Set[mcmcgraph, AddEdge[mcmcgraph, {{V[mcmcgraph],secondvarn},EdgeLabel->IndexL,EdgeDirection->True}]]; *)

						(* link old-t index to new-t version of variable as well *)
						Set[mcmcgraph, AddEdge[mcmcgraph, {{oldTIndexVarNr,secondvarn},EdgeLabel->IndexL,EdgeDirection->True}]];
						
						vdist = SymbolNotInGraph[mcmcgraph, "Dist"<>ToString[v]];
						
						(* add variable for random value of index, link it to new-t index *)
						v = SymbolNotInGraph[mcmcgraph,"p"<>ToString[#]];
						vdist = SymbolNotInGraph[mcmcgraph, "Dist"<>ToString[v]];
						Set[mcmcgraph, AddLabeledVertex[mcmcgraph,v,VariableN]];
						vnr = V[mcmcgraph];
						Set[mcmcgraph, AddEdge[mcmcgraph, {{2,V[mcmcgraph]},EdgeLabel->IndexL,EdgeDirection->True}]];
						(* gate E1 with NotKD taking index, index variable as arugments 	 *)
						(* add gate function which says the equality between new-t and old-t version of variable
						   only counts if the variable for the index is not equal to the index
						   add links forward to the equality constraint, and backward to the index and the index-variable	*)
(*						Set[mcmcgraph, AddLabeledVertex[mcmcgraph,NotKD,FunctionN]];*)

(*						Set[mcmcgraph, AddEdge[mcmcgraph, {{V[mcmcgraph],V[mcmcgraph]-1},EdgeLabel->FactorEdge,EdgeDirection->True}]];
						Set[mcmcgraph, AddEdge[mcmcgraph, {{V[mcmcgraph],oldTIndexVarNr},EdgeLabel->FactorEdge,EdgeDirection->True}]];*)
						(* gate E2 with KD taking index, index variable as arugments 	 *)
(*						Set[mcmcgraph, AddLabeledVertex[mcmcgraph,KroneckerDelta,FunctionN]];*)
(*						Set[mcmcgraph, AddEdge[mcmcgraph, {{V[mcmcgraph],V[mcmcgraph]-2},EdgeLabel->FactorEdge,EdgeDirection->True}]];
						Set[mcmcgraph, AddEdge[mcmcgraph, {{V[mcmcgraph],oldTIndexVarNr},EdgeLabel->FactorEdge,EdgeDirection->True}]];*)
						(* add distribution for index variable *)						
						Set[mcmcgraph, AddLabeledVertex[mcmcgraph,vdist,FunctionN]];
						mcmcgraph = AddEdge[mcmcgraph, {{V[mcmcgraph], vnr},EdgeLabel->CondDep}];
						(* proposal dist. also needs index variable *)
						mcmcgraph = AddEdge[mcmcgraph, {{vnr, propDistNr},EdgeLabel->CondDep}];
						mcmcgraph = AddEdge[mcmcgraph, {{vnr, ration    },EdgeLabel->CondDep}];						
						
						indicName = SymbolNotInGraph[mcmcgraph, "indic"<>ToString[#]];
						nIndicName = SymbolNotInGraph[mcmcgraph, "nIndic"<>ToString[#]];
						mcmcgraph = AddLabeledVertex[mcmcgraph, indicName, VariableN];
						mcmcgraph = AddLabeledVertex[mcmcgraph, nIndicName, VariableN];						
						nIndicNr = V[mcmcgraph];
						indicNr = nIndicNr - 1;
						mcmcgraph = AddEdge[mcmcgraph, {{2, indicNr },EdgeLabel->IndexL}];																				
						mcmcgraph = AddEdge[mcmcgraph, {{2, nIndicNr},EdgeLabel->IndexL}];														
						mcmcgraph = AddEdge[mcmcgraph, {{nIndicNr,constraintn},EdgeLabel->Gate,EdgeDirection->True}];
						mcmcgraph = AddEdge[mcmcgraph, {{indicNr,e2nr},EdgeLabel->Gate,EdgeDirection->True}];						
						mcmcgraph = AddEdge[mcmcgraph, {{indicNr,e3nr},EdgeLabel->Gate,EdgeDirection->True}];						
						mcmcgraph = AddEdge[mcmcgraph, {{oldTIndexVarNr,nIndicNr},EdgeLabel->IndexL,EdgeDirection->True}];
						mcmcgraph = AddEdge[mcmcgraph, {{oldTIndexVarNr, indicNr},EdgeLabel->IndexL,EdgeDirection->True}];
						mcmcgraph = AddLabeledVertex[mcmcgraph, IndicOther,FunctionN];
						mcmcgraph = AddEdge[mcmcgraph, {{indicNr,V[mcmcgraph]},EdgeLabel->CondDep}];
						mcmcgraph = AddEdge[mcmcgraph, {{V[mcmcgraph],nIndicNr},EdgeLabel->CondDep}];
						mcmcgraph = AddLabeledVertex[mcmcgraph, EqualsIndic,FunctionN];
						mcmcgraph = AddEdge[mcmcgraph, {{oldTIndexVarNr,V[mcmcgraph]},EdgeLabel->CondDep}];
						mcmcgraph = AddEdge[mcmcgraph, {{vnr,V[mcmcgraph]},EdgeLabel->CondDep}];
						mcmcgraph = AddEdge[mcmcgraph, {{V[mcmcgraph],indicNr},EdgeLabel->CondDep}];						
						
						vard = AddVariableIndex[vard, #, indName,oldTIndexVarNr,v,vnr,vdist,V[mcmcgraph],indicName,indicNr,nIndicName,nIndicNr];
						{V[mcmcgraph]-3,v,oldTIndexVarNr}
					]&,indices];
					AppendTo[describemcmc,vard];
					{name,var,secondvarn,fname,ration,pickvars,vard}
			]
		];
		processoneconstant = 
			Function[{name},
				Module[{indices,const,pickvars,constNr,constd},
					
     			    const = FirstInListWithName[Vars[parse],name];

					(* 	add one copy *)
					mcmcgraph = AddLabeledVertex[mcmcgraph, name, ConstantN];
					constNr = V[mcmcgraph];
					(* 	pick out & add a set of canonical indices*)
	    			indices = Map[extractLabel, Indices[const]];
	    			
	    			constd = MCMCConstantDescriptor[];
	    			constd = SetConstantNr[constd, constNr];
	    			constd = SetConstantName[constd, name];

					(* for each canonical index ... *)
					pickvars = Map[
						Module[{v,indexVarNr,mcmcname},
							(* add the index *)
							mcmcname = SymbolNotInGraph[mcmcgraph,ToString[#]];
							Set[mcmcgraph, AddLabeledVertex[mcmcgraph,mcmcname,IndexN]];
						  	(*link indices to variable*)
							indexVarNr = V[mcmcgraph];
							Set[mcmcgraph, AddEdge[mcmcgraph, {{indexVarNr,constNr},EdgeLabel->IndexL,EdgeDirection->True}]];
							constd = AddConstantIndex[constd, #, mcmcname, indexVarNr];
							{-1,"foo",indexVarNr}
						]&,indices];
				  	
				  	(*insure hooked into appropriate update functions for variables  	*)
					AppendTo[describemcmc,constd];				  	
				  	{name, const, constNr+1, "uh...", -1, pickvars,constd}
				]
			];
		passoneVARres = Map[processonevariable, uniquevariablenames];
    	passoneCONSTres = Map[processoneconstant, uniqueconstantnames];
    	passoneres = Join[passoneVARres,passoneCONSTres];
(*    	Print["v:", passoneVARres, "\na:", passoneres]; *)
       	getfirstpassres = Cases[passoneres, xwexs_/; MatchQ[xwexs[[1]],#]]&;            			
		processonepasstwo = Function[{res},
			Module[{name,var,ratiofuncid,i,l,fncs,fncsids,allvars,allvarsfirst,othervars,hookup,neighb,vard,pos},
				name = res[[1]];
				var = res[[2]];
(*				Print["this var: ", var];*)
				indexinorderV = Position[order, var][[1,1]];
				ratiofuncid = res[[5]];
				
			(* we get a list of all functions to which the variable is an argument *)
				i = Node[var];
				l = Label[var];
            	fncs = Cases[Map[{DDGetVertexLabels[fullgr][[#]], #}&,  Union[ParentsOf[fullgr, i], ChildrenOf[fullgr, i] ] ],{_FunctionN, _}];
           	 	fncsids = Map[#[[2]]&, fncs];
            (* and a list of all of the variables (including this one) which are arguments to those functions *)
            	allvars = Fold[Union, {}, Map[ImpingingVars[#,Vars[parse]]&, 
            		                          Cases[Funcs[parse], x_Func/;MemberQ[fncsids,Node[x]]]]];

          		neighb = Map[FirstInListWithName[Vars[parse], #]&, Union[Map[extractLabel, allvars]]];          		
				vard = res[[7]];
          		If[!MatchQ[vard,Null],
          			pos = Position[describemcmc,vard][[1]];
          			Map[Set[describemcmc,ReplacePart[describemcmc,pos->AddVariableNeighbor[Extract[describemcmc,pos],#]]]&,neighb]
          		];
          		
            	allvarsfirst = Map[Label, neighb];
(*            	Print["avf: " , allvarsfirst];*)
            	othervars = DeleteCases[allvarsfirst, Label[var]];
            	hookup = Function[{ovar},
            		Module[{fpr},
            			fpr = getfirstpassres[extractLabel[ovar]];
            			If[fpr!={},
            				fpr = fpr[[1]];
            				indexinorderO = Position[order, fpr[[2]]][[1,1]];            			
(*            				Print["fpr: ", fpr, " iioV: ", indexinorderV, " <> iioO: ", indexinorderO];*)
            				If[indexinorderO < indexinorderV,
		            			mcmcgraph = AddEdge[mcmcgraph,{{fpr[[3]],ratiofuncid},EdgeLabel->CondDep,EdgeDirection->True}];,
		            			mcmcgraph = AddEdge[mcmcgraph,{{fpr[[3]]-1,ratiofuncid},EdgeLabel->CondDep,EdgeDirection->True}];
            				];
    	        			Map[Set[mcmcgraph,AddEdge[mcmcgraph,{{#[[3]],ratiofuncid},EdgeLabel->ArgIndex,EdgeDirection->True}]]&,fpr[[6]]];
            			]]];
            	Map[hookup, othervars];
			]];

		Map[processonepasstwo, passoneVARres];
		GeneratedMCMCResult[mcmcgraph, describemcmc]
	];
(****)


(*		vlabels = DDGetVertexLabels[graph];
		uniquevariablenames = 
			Union[Flatten[Map[extractLabel, 
				Cases[vlabels, _VariableN]]]];
		constO = Map[FirstInListWithName[Vars[parse],#]&,uniqueconstantnames];*)
		



GenerateMCMCCode[graph_, variableDefs_, indexDefs_, iterations_, argumentOrders_:{},indexOrders_:{},askInit_:False] :=
    Module[ {orderedIndices, parse, variables, mcmcg,magic,mcmcdescs,acceptanceRatioFunctions,mcmcLabels,
    	 mcmcVariableDefs,mcmcIndexDefs,acceptanceProbFuncSetup,forwardSampleFunc,
    	 origParse,funcSetupString,indexPickProbFuncSetup,generateOne,defineProposalFunc,defineMCMCVariable,
    	 proposalProbFuncSetup,defineMCMCIndex,defineMCMCICndex,exp},

		origParse = ParseGraph[graph, False, argumentOrders, indexOrders];

		magic = GenerateMCMCDependencyDiagram[graph];
		mcmcdescs = MCMCDescriptors[magic];
		mcmcg = MCMCDiagram[magic];
		
		orderedIndices = OrderIndices[mcmcg];
        parse = ParseGraph[mcmcg,False];
        variables = Vars[parse];
        
        mcmcLabels = DDGetVertexLabels[mcmcg];
		
		(* generate function definitions for MCMC code *)
		generateOne[x_MCMCVD] := GenerateUpdateProbabilityFunction[AlgebraicExpression[origParse], VariableDesc[x], VariableNeighbors[x],indexDefs,Indices[origParse]]/.ListLength->Length;
		acceptanceRatioFunctions = Map[generateOne, Cases[mcmcdescs, desc_/;MatchQ[Head[desc], MCMCVD]]];				
		(* done generating function definitions *)
        
        mcmcVariableDefs = {};
        (* by convention, the first two nodes in the graph will be the time indices. fool. *)
        mcmcIndexDefs = {{extractLabel[mcmcLabels[[1]]], 1, iterations},{extractLabel[mcmcLabels[[2]]], 1, iterations}};

        proposalProbFuncSetup = {};
        
        indexPickProbFuncSetup = {};	
        
        defineProposalFunc[mcv_MCMCVD, vd_] := 
        	Module[{func,oldTMCMCV,givs,slots,inds,part,randomfuncname,randomfuncargs},
        		func = Cases[Funcs[parse], f_Func /; Node[f]==VarPropose[mcv][[4]]][[1]];
        		oldTMCMCV = Cases[Vars[parse], v_Var /; Node[v]==VariableOldNode[mcv][[2]]][[1]];
        		givs = Givens[func];
        		slots = ListToCommaSeparatedString[Map[ToString[extractLabel[#]] <> "_"&, givs]]	;
				inds = Sort[VarIndices[mcv], CompareByList[#1[[2]], #2[[2]], Map[extractLabel,Indices[oldTMCMCV]]]&];
(*        		Print[givs, VarIndices[mcv], oldTMCMCV,inds]];*)
        		part = ListToCommaSeparatedString[Join[{VariableOldNode[mcv][[1]]}, Map[#[[4]]&, inds]]];
        		{randomfuncname,randomfuncargs } = 
        			Which[
        				MatchQ[vd[[2]],Discrete],{"DefaultDiscreteProposal",ToString[FullForm[vd[[3]]]]},
        				MatchQ[vd[[2]],IntegerRange],{"DefaultIntegerProposal",ToString[FullForm[vd[[3]]]]},
        				MatchQ[vd[[2]],Continuous],{"DefaultContinuousProposal",ToString[FullForm[vd[[3]]]]},
        				True,{ToString[FullForm[vd]], " this sucks "}
        			];
        		ToString[VarPropose[mcv][[3]]] <> "[" <> slots <> "] := " <> 
        		randomfuncname <> "[Part[" <> part <>"]," <>randomfuncargs <> "];" 
        	];
        	
		defineMCMCVariable[x_MCMCCD] := 
        	Module[{originalDef},
	        	Map[defineMCMCICndex, ConstIndices[x]];
	        ];        	
        
        defineMCMCVariable[x_MCMCVD] := 
        	Module[{originalDef},
	        	originalDef = Cases[variableDefs, deff_/;MatchQ[deff[[1]],VariableName[x]]][[1]];
	        	AppendTo[mcmcVariableDefs , originalDef];
	        	AppendTo[mcmcVariableDefs , {VarAccept[x][[1]], IntegerRange, {0,1}}]; 
	        	AppendTo[mcmcVariableDefs , {VarNotAccept[x][[1]], IntegerRange, {0,1}}]; 
	        	AppendTo[mcmcVariableDefs , ReplacePart[originalDef, 1->VarPropose[x][[1]]]]; 
	        	AppendTo[proposalProbFuncSetup,defineProposalFunc[x,originalDef]];
	        	Map[defineMCMCIndex, VarIndices[x]];
	        ];
	     defineMCMCIndex[x_List] :=
	     	Module[{originalDef},
	     		originalDef = Cases[indexDefs, deff_/;MatchQ[deff[[1]],x[[1]]]][[1]];
	     		AppendTo[mcmcVariableDefs, {x[[4]], IntegerRange,{originalDef[[2]], originalDef[[3]]}}];
	        	AppendTo[mcmcVariableDefs , {x[[8]], IntegerRange, {0,1}}]; 
	        	AppendTo[mcmcVariableDefs , {x[[10]], IntegerRange, {0,1}}]; 	     		
	     		AppendTo[mcmcIndexDefs, ReplacePart[originalDef, 1->x[[2]]]];
	     		AppendTo[indexPickProbFuncSetup,
	     			ToString[x[[6]]] <> "[] := DiscreteUniformDistribution[{ "<> ToString[originalDef[[2]]]<>", "<> ToString[originalDef[[3]]]<>"}];" 
	     		];
	       	];
		defineMCMCICndex[x_List] :=
	     	Module[{originalDef},
	     		originalDef = Cases[indexDefs, deff_/;MatchQ[deff[[1]],x[[1]]]][[1]];	     		
	     		AppendTo[mcmcIndexDefs, ReplacePart[originalDef, 1->x[[2]]]];
	     	];	     	
	    Map[defineMCMCVariable, mcmcdescs];
        mcmcIndexDefs = Sort[mcmcIndexDefs, CompareByList[#1[[1]],#2[[1]],Map[#[[2]]&,orderedIndices]]&];

        acceptanceProbFuncSetup = 
        	MapThread[GenerateAcceptProbabilityDistributionString[#1,#2,origParse,parse]&, {Cases[mcmcdescs, desc_/;MatchQ[Head[desc], MCMCVD]],acceptanceRatioFunctions}];
                	
        forwardSampleFunc = GenerateForwardSamplingCode[mcmcg,mcmcVariableDefs,mcmcIndexDefs,{},{},askInit];	
		
		funcSetupString = "Function[{}," <> StringJoin[Join[acceptanceProbFuncSetup,
															proposalProbFuncSetup,
															indexPickProbFuncSetup]] <> "]";
        exp = ToExpression[funcSetupString];
        		
        {ReplaceAll[exp,DDList->DDCollect], forwardSampleFunc}
        
    ];

GenerateForwardSamplingCode[graph_, variableDefs_, indexDefs_,indexOrdersArg_:{}, argumentOrdersArg_:{},askInit_:False ] := 
	Module[{lines,print,parse,vertexLabels,topoOrder,dimensionOrder,simpDOrder,variables,orderedVariables,
		context,pushContext,popContext,currentIndent,varIter,var,varP,varIdx,varDimensions,closed,
		dim,dimIter,idx,idxl,restOfDim,rDimIter,name,probDists,def,ind,indName,otherIdx,otherIndName,
		constraints,eqns,sol,variableInitializeString,setupStart,setupEnd,indSetup,indexNames,
		setupVeryEnd,modSetup,names,s,getIndexDef,getConstraintsBtwn,getIndex,
		getVarDef,MinIndexString,MinIndexStringSub,MaxIndexString,MaxIndexStringSub,onePD,
		variablesToSet},

		graph = FactorizeGraph[PreprocessGraph[graph]];
		parse = Quiet[ParseGraph[graph,False,indexOrdersArg,argumentOrdersArg]];			
		vertexLabels = DDGetVertexLabels[graph];
		topoOrder = TopologicalSort[graph];
		dimensionOrder = Reverse[OrderIndices[graph,True]];
		simpDOrder = Map[Map[#[[2]]&,#]&,dimensionOrder];
		variables = Sort[Vars[parse]];
		orderedVariables = Sort[variables, CompareByList[Node[#1],Node[#2],topoOrder]&];

		getVarDef[namee_] := Cases[variableDefs, vard_/; MatchQ[vard[[1]],namee]][[1]];
		getIndexDef[IndexN[label_]] := Cases[indexDefs,deff_/;MatchQ[deff[[1]],label]][[1]];
		getIndex[label_] := Cases[Indices[parse], indd_/; MatchQ[Label[indd],label]][[1]];
		getConstraintsBtwn[label1_,label2_] := 
			Cases[IndexConstraints[parse], constr_ /; 
				MatchQ[Sort[Constrained[constr]],Sort[{label1,label2}]] ];					
		MinIndexString[idxx_,indd_,deff_,restOfDimm_]:=
			Module[{},
				If[Length[restOfDimm]==0,
					ToString[IndexMinName[indd]],
					"Max[" <> ToString[IndexMinName[indd]] <>"," <> ListToCommaSeparatedString[Map[MinIndexStringSub[idxx,#]&, restOfDimm]] <> "]"
				]
			];
		MinIndexStringSub[idxx_,otherIdxx_] :=
			Module[{indNamee,otherIndNamee,otherInd,constraintss,eqnss,soll},
				indNamee = extractLabel[idxx];
				otherIndNamee = extractLabel[otherIdxx];
				otherInd = getIndex[otherIdxx];
				constraintss = getConstraintsBtwn[idxx ,otherIdxx];
				eqnss = Fold[#1 && (MakeIndCond[#2]==1)&,True, constraintss];
				soll = Reduce[eqnss/. otherIndNamee->IndexMinName[otherInd], indNamee];		
				ToString[Last[soll]]
			];
		MaxIndexString[idxx_,indd_,deff_,restOfDimm_]:=
			Module[{},
				If[Length[restOfDimm]==0,
					ToString[IndexMaxName[indd]],
					"Min[" <> ToString[IndexMaxName[indd]] <>"," <> ListToCommaSeparatedString[Map[MaxIndexStringSub[idxx,#]&, restOfDimm]] <> "]"
				]
			];
		MaxIndexStringSub[idxx_,otherIdxx_] :=
			Module[{indNamee,otherIndNamee,otherInd,constraintss,eqnss,soll},
				indNamee = extractLabel[idxx];
				otherIndNamee = extractLabel[otherIdxx];
				otherInd = getIndex[otherIdxx];
				constraintss = getConstraintsBtwn[idxx ,otherIdxx];
				eqnss = Fold[#1 && (MakeIndCond[#2]==1)&,True, constraintss];
				soll = Reduce[eqnss/. otherIndNamee->IndexMaxName[otherInd], indNamee];		
				ToString[Last[soll]]
			];					
				

		variablesToSet = {};
		Module[{i,vi,c},
			For[i = Length[orderedVariables],i>0,i--,
				vi = orderedVariables[[i]];
				c = Cases[variablesToSet, varrr_ /; MatchQ[extractLabel[Label[varrr]], extractLabel[Label[vi]]]];
				If[c == {},
					AppendTo[variablesToSet, vi];
				];
			];
		];
		variablesToSet = Reverse[variablesToSet];
(*		Print[variablesToSet];*)

(*
a = For[i = imin, i <= Max[imin, jmin + 0] - imin, i++,
    	Print[i]
    ] &;
(a /. {imin -> 1, jmin -> 1})[]
*)		
		

		(* this bit appears to be about setting up variable initialization code? *)
		variableInitializeString = 
			Module[{varinit,bag,i,vn,indicesForTable,j,n,defff},
				If[ MatchQ[INDEXTYPE,Part],
			        varinit = "If[initializevariables,";
			        bag = {};        
			        For[i = 1,i<=Length[variables],i++,
			        	vn = extractLabel[Label[variables[[i]]]];
			        	If[ !MemberQ[bag,vn] &&!MatchQ[Label[variables[[i]]],_ConstantN],
			            	indicesForTable = {};
			            	AppendTo[bag,vn];
			            	For[j = 1,j<=Length[Indices[variables[[i]]]],j++,
					            n = extractLabel[Indices[variables[[i]]][[j]]];
 								defff=Cases[Indices[parse], indd_/; MatchQ[extractLabel[Label[indd]],n]][[1]];
			            		AppendTo[indicesForTable,{IndexMinName[defff],IndexMaxName[defff]}];
			            	];
				            varinit = varinit <>ToString[FullForm[vn]] <> "=" <> "IndListToTable[" <> ToString[FullForm[indicesForTable]] <>"];"
				        ];
			        ];
	   			    varinit = varinit <>  BuildInitialValueSetter[variablesToSet, orderedVariables,MinIndexString,MaxIndexString,simpDOrder,
													   graph,vertexLabels,getIndexDef,getIndex,getConstraintsBtwn,getVarDef]  <> "];";
			        varinit,
			        ""
			    ]
			];
			
		variableInitializeString = variableInitializeString ;	
        (* end bit about variable initiation? *)
        
        If[ askInit,
            setupStart = "{initializevariables}, Module[{";,
            setupStart = "Module[{initializevariables=True, ";
        ];

        setupEnd = "},";
        setupVeryEnd = "]";
        indSetup = Map[ StringJoin, 
        				Map[Function[{nn},
        						Module[ {c},
                             		c = Cases[Indices[parse], x_DDIndex/;extractLabel[Label[x]]==nn[[1]]];
                            	 	If[ Length[c]>0,
                                 		ToString[IndexMinName[c[[1]]]] <> "=" <>ToString[nn[[2]]] <> "; " <> 
                                 		ToString[IndexMaxName[c[[1]]]] <> "=" <>ToString[nn[[3]]] <>"; "
                             		]
                         		]
                         	],
							indexDefs]];
							
        names = Map[ToString, 
        			Union[
(*        				Map[IndexMinName,            Indices[parse]],
        				Map[IndexMaxName,            Indices[parse]],*)
        				Map[extractLabel[Label[#]]&, Indices[parse]]
        			]
        		];
        modSetup = ListToCommaSeparatedString[names];

		(* BREAK HERE. w00t *)

		lines = {};

		print = AppendTo[lines,StringJoin[Map[ToString[#]&,{##}]]]&;

		If[!UndirectedQ[graph] && AcyclicQ[graph],

			context = {};
			pushContext = Function[{n},AppendTo[context,n]];
			popContext = Function[{},Module[{r},r=Last[context];context=Most[context];r]];
			currentIndent = Function[{},Fold[#1 <> #2 &, "", Table["\t", {Length[context]}]]];

			For[varIter = 1, varIter <= Length[variablesToSet], varIter++,
				var = variablesToSet[[varIter]];
				varP = ParentsOf[graph, Node[var]];
				If[MemberQ[Map[Head[vertexLabels[[#]]]&,varP],FunctionN],
					varIdx = Indices[var];
					varDimensions = Cases[simpDOrder, d_/;Intersection[d,varIdx]!={}];
					While[context != {} && Cases[varDimensions,x_/;Intersection[x,Last[context]]!={}]=={},
						closed = popContext[];
						print[currentIndent[], "];"];
					];
					For[dimIter = 1, dimIter<=Length[varDimensions],dimIter++,
						dim = varDimensions[[dimIter]];
						idxl = Intersection[dim,varIdx];
						If[Length[idxl]==1,
							idx = idxl[[1]];
							If[!MemberQ[Union[ListToSequence[context]],idx],
								restOfDim = Complement[dim,idxl];				
								def = getIndexDef[idx];
								ind = getIndex[idx];			
								indName = extractLabel[idx];							
								print[currentIndent[], "For[", indName,"=",MinIndexString[idx,ind,def,restOfDim], ",", indName, "<=", MaxIndexString[idx,ind,def,restOfDim], ",",indName,"++",","];							
								pushContext[dim];
								For[rDimIter=1,rDimIter<=Length[restOfDim],rDimIter++,
									otherIdx = restOfDim[[rDimIter]];
									otherIndName = extractLabel[otherIdx];
									constraints = getConstraintsBtwn[idx,otherIdx];
									eqns = Fold[#1 && (MakeIndCond[#2]==1)&,True, constraints];
									sol = Reduce[eqns, otherIndName];
									print[currentIndent[] , First[sol], " = ", Last[sol], ";"];									
								];
							];
							,
							Print["um, I think something weird happened?"];
						];
					];					
					name = extractLabel[Label[var]];
					probDists = Cases[Funcs[parse], f_Func/; Args[f]=={Label[var]}];
(*					Print[var, "><", probDists];			*)
					onePD[d_] := 
						Module[{str,cond,symStr},
							{str,cond} = MakeSampleFunc[d, Funcs[parse], Vars[parse], Indices[parse], getVarDef[name]];
							symStr = MakeSymbolString[Label[var], orderedVariables];
							If[cond == "",
								print[currentIndent[] ,  symStr, " = ", str, ";"],							
								print[currentIndent[] , "If[", cond, ","];
								print[currentIndent[] , "\t", symStr, " = ", str, ";"];
								print[currentIndent[] , "];"]
							];
						];
(*					print[currentIndent[] , "sample[", extractLabel[Label[var]], "];"];					*)
					Map[onePD, probDists];
				];
			];
			While[context != {},
				closed = popContext[];
				print[currentIndent[], "];"];
			];
			,
			Print["Wtf,mate?"];
		];
		(*Map[Print, lines];
		lines*)
		s = Fold[#1<>#2<>"\n"&,"",lines];
        ReplaceAll[ToExpression["Function[" <> ToString[setupStart] <> ToString[modSetup] <> ToString[setupEnd] <> Map[ToString, indSetup] <> Map[ToString,variableInitializeString]<> s <> ToString[setupVeryEnd] <> "]"], DDList->DDCollect]

    ];
    
BuildInitialValueSetter[variablesToSet_, orderedVariables_,mainCodeMinIndexString_,mainCodeMaxIndexString_,simpDOrder_,
						graph_,vertexLabels_,getIndexDef_,getIndex_,getConstraintsBtwn_,getVarDef_] :=
	Module[{varIter,var,varP,varIdx,varDimensions,closed,context,pushContext,popContext,currentIndent,lines,print,
			dimIter,dim,idxl,idx,restOfDim,ind,indName,def,rDimIter,otherIdx,otherIndName,constraints,eqns,sol,
			InitMinIndexString,InitMinIndexStringSub,InitMaxIndexString,InitMaxIndexStringSub,requirements},
			lines = {};
			print = AppendTo[lines,StringJoin[Map[ToString[#]&,{##}]]]&;
			requirements = {};
			context = {};
			pushContext = Function[{n, req},AppendTo[context,n];AppendTo[requirements,req]];
			popContext = Function[{},Module[{r1,r2},
					r1=Last[context];
					r2=Last[requirements];
					context=Most[context];
					requirements=Most[requirements];
					{r1,r2}]];
			currentIndent = Function[{},Fold[#1 <> #2 &, "", Table["\t", {Length[context]}]]];		
			
		InitMinIndexString[idxx_,indd_,deff_,restOfDimm_]:=
			Module[{},
				If[Length[restOfDimm]==0,
					ToString[IndexMinName[indd]],
					"Min[" <> ToString[IndexMinName[indd]] <>"," <> ListToCommaSeparatedString[Map[InitMinIndexStringSub[idxx,#]&, restOfDimm]] <> "]"
				]
			];
		InitMinIndexStringSub[idxx_,otherIdxx_] :=
			Module[{indNamee,otherIndNamee,otherInd,constraintss,eqnss,soll},
				indNamee = extractLabel[idxx];
				otherIndNamee = extractLabel[otherIdxx];
				otherInd = getIndex[otherIdxx];
				constraintss = getConstraintsBtwn[idxx ,otherIdxx];
				eqnss = Fold[#1 && (MakeIndCond[#2]==1)&,True, constraintss];
				soll = Reduce[eqnss/. otherIndNamee->IndexMinName[otherInd], indNamee];		
				ToString[Last[soll]]
			];
		InitMaxIndexString[idxx_,indd_,deff_,restOfDimm_]:=
			Module[{},
				If[Length[restOfDimm]==0,
					ToString[IndexMaxName[indd]],
					"Max[" <> ToString[IndexMaxName[indd]] <>"," <> ListToCommaSeparatedString[Map[InitMaxIndexStringSub[idxx,#]&, restOfDimm]] <> "]"
				]
			];
		InitMaxIndexStringSub[idxx_,otherIdxx_] :=
			Module[{indNamee,otherIndNamee,otherInd,constraintss,eqnss,soll},
				indNamee = extractLabel[idxx];
				otherIndNamee = extractLabel[otherIdxx];
				otherInd = getIndex[otherIdxx];
				constraintss = getConstraintsBtwn[idxx ,otherIdxx];
				eqnss = Fold[#1 && (MakeIndCond[#2]==1)&,True, constraintss];
				soll = Reduce[eqnss/. otherIndNamee->IndexMaxName[otherInd], indNamee];		
				ToString[Last[soll]]
			];								
			
			
			
			For[varIter = 1, varIter <= Length[variablesToSet], varIter++,
				var = variablesToSet[[varIter]];
				varP = ParentsOf[graph, Node[var]];
				If[MemberQ[Map[Head[vertexLabels[[#]]]&,varP],FunctionN],
					varIdx = Indices[var];
					varDimensions = Cases[simpDOrder, d_/;Intersection[d,varIdx]!={}];
					While[context != {} && Cases[varDimensions,x_/;Intersection[x,Last[context]]!={}]=={},
						closed = popContext[];
						print[currentIndent[], "];"];
					];
					For[dimIter = 1, dimIter<=Length[varDimensions],dimIter++,
						dim = varDimensions[[dimIter]];
						idxl = Intersection[dim,varIdx];
						If[Length[idxl]==1,
							idx = idxl[[1]];
							If[!MemberQ[Union[ListToSequence[context]],idx],
								restOfDim = Complement[dim,idxl];				
								def = getIndexDef[idx];
								ind = getIndex[idx];			
								indName = extractLabel[idx];							
								print[currentIndent[], "For[", indName,"=",InitMinIndexString[idx,ind,def,restOfDim], ",", indName, "<=", InitMaxIndexString[idx,ind,def,restOfDim], ",",indName,"++",","];							
								Module[{indNameS},
									indNameS = ToString[indName];
									pushContext[
										dim, 
										{indNameS <> "<" <> mainCodeMinIndexString[idx,ind,def,restOfDim] <> " || " <>
										 indNameS <> ">" <> mainCodeMaxIndexString[idx,ind,def,restOfDim],
										 indNameS <> ">=" <> ToString[IndexMinName[ind]] <> " && " <> indNameS <> "<=" <> ToString[IndexMaxName[ind]]
										}
									];
								];
							];
							,
							Print["um, I think something weird happened?"];
						];
					];
					
					Module[{orrs,ands,or,and,name,str,symStr,vd},
						name = extractLabel[Label[var]];
						symStr = MakeSymbolString[Label[var], orderedVariables];						
						vd = getVarDef[name];
	        			str = Which[
	        				MatchQ[vd[[2]],Discrete],
							"RandomChoice[" <> ToString[FullForm[vd[[3]]]] <> "];",
	        				MatchQ[vd[[2]],IntegerRange],
	        				If[MatchQ[vd[[3,1]],_List],
	        					"Map[RandomInteger[DiscreteUniformDistribution[#]]&," <> ToString[FullForm[vd[[3]]]] <> "];",
	        					"RandomInteger[DiscreteUniformDistribution[" <> ToString[FullForm[vd[[3]]]] <> "]];"
	        				],
	        				MatchQ[vd[[2]],Continuous],
	        				If[MatchQ[vd[[3,1]],_List],
	        					"Map[RandomReal[UniformDistribution[#]]&," <> ToString[FullForm[vd[[3]]]] <> "];",
	        					"RandomReal[UniformDistribution[" <> ToString[FullForm[vd[[3]]]] <> "]];"
	        				],
	        				True,{ToString[FullForm[vd]], " this sucks "}
	        			];

						orrs = Map[#[[1]]&, requirements];
						ands = Map[#[[2]]&, requirements];
						or  = Fold[ToString[#1] <> " || " <> ToString[#2]&, First[orrs], Rest[orrs]];
						and = Fold[ToString[#1] <> " && " <> ToString[#2]&, First[ands], Rest[ands]];

						print[currentIndent[] , "If[ (" <> or <> ") && " <> and <> ", "];
						print[currentIndent[], "\t", symStr, " = ", str];
						print[currentIndent[] , "];"]						
					];
				];
			];
			While[context != {},
				closed = popContext[];
				print[currentIndent[], "];"];
			];	
			Fold[#1<>#2<>"\n"&,"",lines]
	];
	
MCMCConstantDescriptor[] := MCMCCD[0,"",{}];
SetConstantNr[x_MCMCCD,y_] := ReplacePart[x,1->y];
SetConstantName[x_MCMCCD,y_] := ReplacePart[x,2->y];
AddConstantIndex[x_MCMCCD,original_,mcmcname_,mcmcnr_] := 
	ReplacePart[x, 3->Append[x[[3]],
		{
			original,mcmcname,mcmcnr
		}]];
ConstantNr[x_MCMCCD] := x[[1]];
ConstantName[x_MCMCCD] := x[[2]];
ConstIndices[x_MCMCCD] := x[[3]];		
    
MCMCVariableDescriptor[] := MCMCVD[0,"",None,{},None,None,None,None,{},None];
VariableNr[x_MCMCVD] := x[[1]];
VariableName[x_MCMCVD] := x[[2]];
VariableDesc[x_MCMCVD] := x[[3]];
VarIndices[x_MCMCVD] := x[[4]];
VarAccept[x_MCMCVD] := x[[5]];
VarPropose[x_MCMCVD] := x[[6]];
VariableOldNode[x_MCMCVD] := x[[7]];
VariableNewNode[x_MCMCVD] := x[[8]];
VariableNeighbors[x_MCMCVD] := x[[9]];
VarNotAccept[x_MCMCVD] := x[[10]];

SetVariableNr[x_MCMCVD,y_] := ReplacePart[x,1->y];
SetVariableName[x_MCMCVD,y_] := ReplacePart[x,2->y];
SetVariableDesc[x_MCMCVD,y_Var] := ReplacePart[x,3->y];
AddVariableIndex[x_MCMCVD,original_,mcmcname_,mcmcnr_,
	pickvarname_,pickvarnr_,pickdistname_,pickdistnr_,
	indicatorname_,indicnr_,notindicname_,notindicnr_] := 
	ReplacePart[x, 4->Append[x[[4]],
		{
			original,mcmcname,mcmcnr,pickvarname,pickvarnr,pickdistname,pickdistnr,
			indicatorname,indicnr,notindicname,notindicnr
		}]];
SetVariableAccept[x_MCMCVD,varname_,varnr_Integer,distname_,distnr_Integer] :=
	ReplacePart[x,5->{varname,varnr,distname,distnr}];	
SetVariablePropose[x_MCMCVD,varname_,varnr_Integer,distname_,distnr_Integer] :=
	ReplacePart[x,6->{varname,varnr,distname,distnr}];
SetVariableOldNode[x_MCMCVD,name_,nr_Integer] :=
	ReplacePart[x,7->{name,nr}];
SetVariableNewNode[x_MCMCVD,name_,nr_Integer] :=
	ReplacePart[x,8->{name,nr}];
AddVariableNeighbor[x_MCMCVD,neighbor_] := 
		ReplacePart[x,9->Append[x[[9]],neighbor]];
SetVariableNotAccept[x_MCMCVD,varname_,varnr_Integer,distname_,distnr_Integer] :=
	ReplacePart[x,10->{varname,varnr,distname,distnr}];		
		

MCMCDiagram[x_GeneratedMCMCResult] := x[[1]];
MCMCDiagram[x_Graph] := MCMCDiagram[GenerateMCMCDependencyDiagram[x]];
MCMCDescriptors[x_GeneratedMCMCResult] := x[[2]];
			
PartMaker[num_, thing_, pts_] :=
	If[num == 1,
		thing,
		Part[thing, ReplacePart[Part[pts, Span[1, num - 1]], 0 -> Sequence]]
	];	
			
(* begin function generation code *)
GenerateUpdateProbabilityFunction[pr_, var_, vars_,indexDefs_,parseInds_] := 
Module[{finitediff, changedVar, varIndices, tempInds, name, exp1,exp2,
	allnames, nameattimet, nameattimetplusone,ourpr,replaceIndex,diffname,assign},
	ourpr = ReplaceAll[pr, Subscript->Part];
	varIndices = Indices[var];
	tempInds = Map[SymbolNotInExpression[pr, "i"]&, Range[Length[varIndices]]];
	name = extractLabel[Label[var]];
	allnames = Map[extractLabel[Label[#]]&,vars];
	finitediff = 
			FiniteDiffProbability[ourpr, Part[extractLabel[Label[var]], ListToSequence[tempInds]]];
			
 	replaceIndex[idef_]:= 
        	Module[{originals},
        		originals = Cases[parseInds, idx_DDIndex /; MatchQ[extractLabel[Label[idx]],idef[[1]]]];
        		exp1 = Map[ToString[FullForm[ IndexMinName[#]->idef[[2]] ]]&, originals];
        		exp2 = Map[ToString[FullForm[ IndexMaxName[#]->idef[[3]] ]]&, originals];
        		Join[exp1, exp2]
        	];
				
	nameattimet = SymbolNotInExpression[finitediff, ToString[name]];
	nameattimetplusone = SymbolNotInExpression[{finitediff,nameattimet}, ToString[name]];
	
	diffname = SymbolNotInExpression[{finitediff,nameattimet,nameattimetplusone}, "delta" <> ToString[name]];
	assign = ToString[diffname] <> " = " <> ToString[FullForm[(nameattimetplusone-nameattimet)]] <> ";";
	
	allnames = Join[{nameattimetplusone}, ReplaceAll[allnames,name->nameattimet], tempInds];
	finitediff = ReplaceAll[finitediff, Ch[name] -> (nameattimetplusone-nameattimet)];
	finitediff = ReplaceAll[finitediff, name -> nameattimet];	
	(*Print[finitediff];*)
	finitediff = ReplaceAll[finitediff, (nameattimetplusone-nameattimet) -> diffname];
	finitediff = ReplaceAll[finitediff, ListLength -> LLPLACEHOLD];
   	(* This convoluted next bit protects us from part errors, as best as it can. *)
	finitediff = ReplaceAll[
		finitediff, 
		Part[x_, y__] :> 
			If[(Fold[#1 && (Greater[#2, 0]) &, True, {y}]) && 
				Fold[(#1 && #2)&, True, MapIndexed[ListLength[Evaluate[PartMaker[First[#2], x, {y}]]] >= #1 &, {y}]], 
				Part[x, y], 
				0
			]
		];
	finitediff = ReplaceAll[finitediff, LLPLACEHOLD -> ListLength];		

	ToExpression["ReplaceAll[Function[" <> ToString[FullForm[allnames]] <> ", Module[{" <> ToString[diffname] <> "}," <> assign <> " Min[1,Quiet[Check[" <> ToString[FullForm[finitediff]] <> ", (Print[\"this makes me unhappy.\"];1) , {Infinity::indet,Power::infy}], {Infinity::indet,Power::infy}]]]], {"<> ListToCommaSeparatedString[Flatten[Map[replaceIndex, indexDefs]]] <>"}]"]
];

GenerateAcceptProbabilityDistributionString[varDef_MCMCVD, funcDef_, origParse_Parse, mcmcParse_Parse] :=
	Module[{templ,ratioFunctionName,ratioFunctionString,funcName,funcArgsAsSlots,funcGivenSlot,
		funcArgsVars,funcGivenArg,ratFuncArgs,accFunc,origVar,indz,a,ratioInds,repString,indNameToPick},
		funcName = ToString[VarAccept[varDef][[3]]];
		ratioFunctionName = ToString[SymbolNotInExpression[funcDef, funcName <> "Ratio"]];
		ratioFunctionString = ToString[FullForm[funcDef]];
		accFunc = Cases[Funcs[mcmcParse], x_Func /; Node[x]==VarAccept[varDef][[4]]][[1]];
		origVar = VariableDesc[varDef];
		indz = VarIndices[varDef];
		indNameToPick[name_] := Cases[indz, i_ /; MatchQ[i[[1]],name]][[1,4]];
		funcArgsAsSlots = ListToCommaSeparatedString[Map[ToString[FullForm[#]] <>"_"&,Map[extractLabel,Givens[accFunc]]]];
		funcArgsVars  = ListToCommaSeparatedString[Map[ToString[FullForm[#]] &,Map[extractLabel,Givens[accFunc]]]];
		funcGivenSlot = ListToCommaSeparatedString[Map[ToString[FullForm[#]] <>"_"&,Map[extractLabel,Args[accFunc]]]];
		funcGivenArg  = ListToCommaSeparatedString[Map[ToString[FullForm[#]] &,Map[extractLabel,Args[accFunc]]]];
		a = ToString[SymbolNotInExpression[AlgebraicExpression[origParse], VarName[varDef]]];
		ratioInds = Map[indNameToPick[extractLabel[#]]&, Indices[origVar]];
		ratFuncArgs = ListToCommaSeparatedString[
						Join[{ToString[a]}, 
							  Map[ToString[extractLabel[Label[#]]]&, VariableNeighbors[varDef]],
						   	  Map[ToString, ratioInds]
					  	]];
		repString = 
			If[Length[ratioInds] > 0,
				"ReplacePart[" <> ToString[VariableName[varDef]] <>"," <> ToString[ratioInds] <>"->" <> ToString[VarPropose[varDef][[1]]] <> "]",
				 ToString[VarPropose[varDef][[1]]]
			];
		ratFuncArgs = StringReplace[ratFuncArgs, ToString[a] -> repString];
		
templ = "
RATIOFUNTIONNAME = RATIOFUNCTIONSTRING;
FUNCNAME /: PDF[FUNCNAME[FUNCARGSSLOTS], FUNCGIVENSLOT] := PDF[FUNCNAME[FUNCARGSVARS]][FUNCGIVENARG];
FUNCNAME /: PDF[FUNCNAME[FUNCARGSSLOTS]] := (RATIOFUNTIONNAME[RATFUNCARGS]*Dirac[1 - #]) + ((1 - RATIOFUNTIONNAME[RATFUNCARGS])*Dirac[#])&;
FUNCNAME /: Random`DistributionVector[FUNCNAME[FUNCARGSSLOTS], n_Integer, prec_?Positive] := 
	Table[(1 - Heaviside[-1*(RATIOFUNTIONNAME[RATFUNCARGS] - RandomReal[])]), {n}];
"	;
		StringReplace[templ, {"RATIOFUNTIONNAME"->ratioFunctionName, 
							  "RATIOFUNCTIONSTRING"->ratioFunctionString,
							  "FUNCNAME"-> funcName,
  							  "FUNCARGSSLOTS"-> funcArgsAsSlots,
							  "FUNCGIVENSLOT"-> funcGivenSlot,
							  "FUNCARGSVARS"-> funcArgsVars,
							  "FUNCGIVENARG"->funcGivenArg,
							  "RATFUNCARGS"->ratFuncArgs
							  }]
	];

DefaultIndexOrderings[graph_] := DefaultOrderings[graph][[1]];
DefaultArgumentOrderings[graph_, indexOrder_] := DefaultOrderings[graph,indexOrder][[2]];

DefaultOrderings[graph_Graph,indexOrderArg_:{}] :=
	Module[{labels, edges, labelsToNums, numsToLabels, indices, indexNums, 
		functions, functNums, variables, varblNums, nameWithIndices, argumentOrder, 
		makeIndexOrder, makeArgumentOrder,indexOrder},
		graph = PreprocessGraph[graph];
		
		labels = DDGetVertexLabels[graph];	
		edges = DDGetEdgeLabels[graph];
			
		labelsToNums[ls_] := Map[Position[labels, #][[1,1]]&,ls];
		numsToLabels[ns_] := Map[Part[labels,#]&,ns];
		
		indices = Sort[Cases[labels, x_IndexN]]; (* here is where the indices get sorted. *)
		indexNums = labelsToNums[indices];
		
		functions = Cases[labels, x_/;MatchQ[Head[x],FunctionN]  || MatchQ[Head[x],ConstraintN]  ];
		functNums = labelsToNums[functions];
		
		variables = Cases[labels, x_/;MatchQ[Head[x],VariableN]  || MatchQ[Head[x],ConstantN]  ];
		varblNums = labelsToNums[variables];				
		
		makeIndexOrder[number_] := Module[{par,ind},
			par = ParentsOfIndices[graph, number];
			ind = Cases[par, p_ /; MemberQ[indexNums,p[[1]]] && MatchQ[edges[[p[[2]]]],IndexL]];
			ind = Sort[ind, CompareByList[#1[[1]],#2[[1]],indexNums]&];
			{number, Map[#[[1]]&,ind]}
		];
		
		If[indexOrderArg == {},
			indexOrder = Map[makeIndexOrder, Join[functNums, varblNums]],
			indexOrder = indexOrderArg			
		];
		
		nameWithIndices[number_] := Module[{par,myinds,ind,nums},
			par = ParentsOfIndices[graph, number];
			myinds = Cases[indexOrder, x_/;x[[1]]==number];
			ind = Cases[par, p_ /; MemberQ[indexNums,p[[1]]] && MatchQ[edges[[p[[2]]]],IndexL]];
			ind = Sort[ind, CompareByList[#1[[1]],#2[[1]],myinds[[1]]]&];
			nums = Join[{number},Map[#[[1]]&,ind]];
			{Map[extractLabel, numsToLabels[nums]], number}
		];
		
		argumentOrder = Map[#[[2]]&, Sort[Map[{Fold[#1 <> ToString[#2] &, "", #[[1]]], #[[2]]} &,Map[nameWithIndices, Join[indexNums, varblNums]]]]];
		
		makeArgumentOrder[number_] := Module[{par,chi,giv,arg},
			par = ParentsOfIndices[graph, number];
			chi = ChildrenOfIndices[graph, number];
			
			giv = Cases[par, p_ /; (MemberQ[argumentOrder,p[[1]]]) && (MatchQ[edges[[p[[2]]]],CondDep] || MatchQ[edges[[p[[2]]]],UnCondDep] || MatchQ[edges[[p[[2]]]],FactorEdge] || MatchQ[edges[[p[[2]]]],IndexCons])];
			giv = Sort[giv, CompareByList[#1[[1]],#2[[1]],argumentOrder]&];
			
			arg = Cases[chi, p_ /; (MemberQ[argumentOrder,p[[1]]]) && (MatchQ[edges[[p[[2]]]],CondDep] || MatchQ[edges[[p[[2]]]],UnCondDep] || MatchQ[edges[[p[[2]]]],FactorEdge] || MatchQ[edges[[p[[2]]]],IndexCons])];
			arg = Sort[arg, CompareByList[#1[[1]],#2[[1]],argumentOrder]&];
			
			{number, Map[#[[1]]&,Join[arg, giv]]}
		];
		
		{indexOrder,
			Map[makeArgumentOrder, functNums]}
		
	];
MakeSampleFunc[func_, functions_, variables_, indices_, varDef_] :=
    Module[ { givens, gates, string, j, power, vars, k, exp, pow, givStrings, givsString, arginds,randFunction,condition,arguments},        
        If[ IsDNode[func] ,        
	        givens = Givens[func];
	        arguments = Args[func];
	        gates = Gates[func];
	        arginds = ArgIndices[func];
	        string = MakeSymbolString[Label[func], functions];
        
            If[ Length[givens] > 0,
                givStrings = Map[MakeArgString[#,variables, arginds, indices]&, givens];
                givsString = ListToCommaSeparatedString[givStrings];,
                givsString = "";
            ];
	
			randFunction = 
				Which[
					MatchQ[varDef[[2]],IntegerRange],
					"RandomInteger[" <> string <> "[" <> givsString <> "]]",
					MatchQ[varDef[[2]],Discrete],
					"RandomChoice[" <> string <> "[" <> givsString <> "]]",
					MatchQ[varDef[[2]],Continuous],										
					"RandomReal[" <> string <> "[" <> givsString <> "]]"					
				];

	        power = 1;
	        For[j = 1, j <= Length[gates], j++,
	        	power = power * Heaviside[ToExpression[MakeSymbolString[gates[[j]], variables]]];
			];
	        For[j = 1, j<= Length[arguments], j++, 
				vars = Flatten[Map[Part[#,3]&,Select[variables, Part[#,2]== arguments[[j]]&]]];
				For[k = 1, k<= Length[vars], k++,
					power = power * Heaviside[ToExpression[MakeSymbolString[ vars[[k]], variables]] ];
				];
			];
(*	        exp = Quiet[ToExpression[string],{Part::"pspec",Rule::"argr"}];
	        pow = exp ^ power*)
	        If[!MatchQ[power,1],
	        	condition =  ToString[FullForm[power]] <> "==1",
	        	condition = "";
	        ];
	        {randFunction, condition}
        ]
    ];	

CompleteOrderings[indexOrdersArg_:{}, argumentOrdersArg_:{}, graph_] :=
	Module[{indexOrders, argumentOrders,indCheckAndReplace,argCheckAndReplace},

        indexOrders = DefaultIndexOrderings[graph];        
		indCheckAndReplace[ord_] := Module[{defL,def,order},
			defL = Cases[indexOrders, x_/; x[[1]] == ord[[1]]];
			If[defL != {},
				def = defL[[1]];
				If[Complement[ord[[2]], def[[2]]]=={},
					order = Sort[def[[2]], CompareByList[#1,#2,ord[[2]]]&];
					indexOrders = ReplaceAll[indexOrders, def->{def[[1]], order}];
				];
			]
		];
		Map[indCheckAndReplace, indexOrdersArg];

        argumentOrders = DefaultArgumentOrderings[graph,indexOrders];
		argCheckAndReplace[ord_] := Module[{defL,def,order},
			defL = Cases[argumentOrders, x_/; x[[1]] == ord[[1]]];
			If[defL != {},
				def = defL[[1]];
				If[Complement[ord[[2]], def[[2]]]=={},
					order = Sort[def[[2]], CompareByList[#1,#2,ord[[2]]]&];
					argumentOrders = ReplaceAll[argumentOrders, def->{def[[1]], order}];
				];
			]
		];        
		Map[argCheckAndReplace, argumentOrdersArg];
        {indexOrders, argumentOrders}
	];

(*end function generating code *)
SetIndexType[t_] := If[MatchQ[t, Part]||MatchQ[t,Subscript],INDEXTYPE=t];

End[] (* End Private Context *)
	
EndPackage[];