(* ::Package:: *)
BeginPackage["FunctionMaker`", {"DDUtils`"}];

FiniteDiffEnergy::usage = "";
FiniteDiffProbability::usage = "";
SimpDiff::usage = "";
Ch::usage = "";

Begin["`Private`"] (* Begin Private Context *) 

(* begin finite difference engine code *)
SetAttributes[MyFreeQ,   HoldAll];
SetAttributes[Simp,      HoldAll];
SetAttributes[SimpDiff,  HoldAll];

SetSystemOptions["DefinitionsReordering" -> None]; 

FiniteDiffEnergy[e_,v_] :=
    Module[ {d,s},
        d = Quiet[SimpDiff[e,v], {Part::pspec}];
        s = Simp[Evaluate[d]];
        s
    ];
FiniteDiffProbability[e_,v_] :=
    E^FiniteDiffEnergy[Log[e],v];

SimpDiff[e_,v_] :=
Module[{},
    Diff[Simp[Unevaluated[e]],v]
 ];


(* Interesting fact: Some of the rules below will cause ugly pspec errors when 
   the code is first loaded, unless we do this. we do turn it back on after all
   of the declarations. *)
Off[Part::pspec];

Diff[e1_, v_] :=
    (0)/; MyFreeQ[e1,v];
Diff[v_, v_] :=
    Ch[v];
Diff[w_?NumberQ, v_] :=
    0;
Diff[w_?AtomQ, v_?AtomQ] :=
    If[ MatchQ[w,v],
        Ch[v],
        0
    ];
Diff[Times[e1_,e2_], v_] :=
	Module[{e1d, e2d},
		e1d = SimpDiff[e1,v];
		e2d = SimpDiff[e2,v];
    	(e1*e2d) + (e2*e1d) + (e1d*e2d)
	];
Diff[e1s_Plus, v_] :=
	Module[{tmp},
		tmp = PlusToList[e1s];
	    Plus[ListToSequence[Map[SimpDiff[#,v]&, tmp]]]
    ];
Diff[Product[e1_, {e2_,min_,max_}], v_] :=
    SimpDiff[E^(Log[Product[e1,{e2,min,max}]]), v];
Diff[Sum[e_,{it_,min_,max_}], v_] :=
	Module[{d,di=0,dj=0},
		d = SimpDiff[e,v];
		Module[{minch = 0,maxch=0},
			If[!MatchQ[Head[min],ListLength] && !MyFreeQ[min,v],
				minch = SimpDiff[min,v];
			];
			If[!MatchQ[Head[max],ListLength] && !MyFreeQ[max,v],
				maxch = SimpDiff[max,v];
			];
			di =  If[maxch > 0,1,0]*Sum[d + e, Evaluate[{it,max+1,max+maxch}]] -
					If[-maxch > 0,1,0]*Sum[d+e,Evaluate[{it,max+maxch+1,max}]];
			dj = -If[minch > 0,1,0]*Sum[d + e, Evaluate[{it,min,min+minch-1}]] +
					If[-minch > 0,1,0]*Sum[d+e,Evaluate[{it,min+minch,min-1}]];					
		];
	    Simp[Evaluate[Sum[d, {it,min,max}]]] + di + dj
	];
Diff[e_, h_[e_,k__]] :=
    Ch[e] /; MatchQ[h,Part] || MatchQ[h, Subscript]; 
Diff[h_[e_,i_], h_[e_,k_]] :=
    h[Ch[e],k]*KroneckerDelta[i,k] /; MatchQ[h,Part] || MatchQ[h, Subscript];
Diff[h_[e_,i__], h_[e_,k__]] :=
    Module[ {len,ii,kk,u,uu},
        len = Min[Length[{k}],Length[{i}]];
        ii = Part[{i},1;;len];
        kk = Part[{k},1;;len];
        Which[
        	Length[{k}]==Length[{i}],
	    	u = {k},
	        Length[{k}]>Length[{i}],
	        u = kk,
        	Length[{k}]<Length[{i}],
        	u = Join[kk,Part[{i},len+1;;]]
        ];
        uu = ListToSequence[u];
        h[Ch[e],uu]*Fold[Times,1,MapThread[KroneckerDelta[#1,#2]&,{ii,kk}]]
    ] /; MatchQ[h,Part]||MatchQ[h, Subscript];
Diff[h_[e1_?AtomQ, i__], h_[e2_?AtomQ, k__]] :=
    0 /; (!MatchQ[e1,e2] && (MatchQ[h,Part]||MatchQ[h, Subscript]));
Diff[h_[e1_, i__],e2_] :=
    h[SimpDiff[e1,e2],i] /; !MyFreeQ[e1,e2] && !MatchQ[Head[e1],DDList] && (MatchQ[h,Part]||MatchQ[h, Subscript]);
Diff[Part[DDList[e_,l_], i__], h_[e_,k__]] :=
    Module[ {indexDeltas,il = {},ii,kk,ll},
        ii = {i};
        kk = {k};
        ll = l;
        indexDeltas = MapThread[(Which[ 
        							 ListQ[#1] && ii!={},
                                     AppendTo[il,First[ii]];
                                     ii = Rest[ii],
                                     !ListQ[#1],
                                     AppendTo[il,#1]
                                 ];
                                 KroneckerDelta[Last[il],#2])&,
                                {ll,kk}
                      ];
        Part[Ch[e], ListToSequence[il]] * Times[ListToSequence[indexDeltas]]
    ] /; MatchQ[h,Part]||MatchQ[h,Subscript];
Diff[Part[DDList[e1_,l_],i__], h_[e2_,k__]] :=
    0 /; (MatchQ[h,Part] || MatchQ[h,Subscript]) && !MatchQ[e1,e2];
Diff[Part[list_,spec__], v_] :=
	Module[{specd, newspec},
		specd = Map[SimpDiff[#,v]&,{spec}];
		newspec = MapThread[#1 + #2&, {{spec},specd}];
		Part[list,ListToSequence[newspec]] - Part[list,spec]
	] /; MyFreeQ[list,v] && !MyFreeQ[{spec},v];
Diff[Log[h_[e_,i_]], h_[e_,k_]] :=
    (Log[h[e,k]+h[Ch[e],k]]-Log[h[e,k]])*KroneckerDelta[i,k] /; MatchQ[h,Part]||MatchQ[h, Subscript];
Diff[Log[Factorial[e1_]], v_] := 
    Module[{d,it,check},
    	d = SimpDiff[e1, v];
    	If[!MatchQ[d,0],
	    	it = IteratorFor[e1];
    		check = If[d>0,1,0];
    		((check*2)-1)Sum[Log[it], Evaluate[{it, (e1+1+(1-check)d),(e1+(check*d))}]],
    		0
    	]
    ];
Diff[Log[e_], v_] := 
    Log[e + SimpDiff[e,v]] - Log[e];
Diff[Pochhammer[e1_,e2_], v_] := 
	SimpDiff[Factorial[e1+e2-1]/Factorial[e1-1],v];
Diff[Factorial[e1_], v_] := 
    Module[{d,it,check},
    	d = SimpDiff[e1, v];
    	it = IteratorFor[e1];
    	check = If[d>0,1,0];
    	If[!MatchQ[d,0],
    		check(Factorial[e1](Product[it,Evaluate[{it,e1+1,e1+d}]]-1)) + (1-check)(Factorial[e1+d](1-Product[it,Evaluate[{it,e1+1+d,e1}]])),
    		0
    	]
    ];
Diff[Power[e1_, e3_?IntegerQ], v_] :=
	Module[{tmp},
		tmp = SimpDiff[Power[e1,Abs[e3]],v];
    	-tmp / (Power[e1,2*Abs[e3]] + (tmp * Power[e1,Abs[e3]]))
	] /; e3 < 0;
Diff[Power[e1_, e3_?IntegerQ], v_] :=
	Module[{tmp},
		tmp = SimpDiff[e1,v];
	    Sum[Binomial[e3, iterz]*Power[e1,e3-iterz]*Power[tmp,iterz], {iterz, 1, e3}]
	];
Diff[Power[e1_, e2_],v_] :=
    (Power[e1, SimpDiff[e2,v]]-1)Power[e1, e2] /; MyFreeQ[e1,v] && !MyFreeQ[e2,v];
Diff[Power[e1_,e2_],v_] := 	
    Module[{tmp},
    	tmp = SimpDiff[e1,v];
		Power[(e1+tmp),e2] * Power[(e1+tmp), SimpDiff[e2,v]] - Power[e1,e2]
	];
Diff[If[Greater[e1_, 0], 1, 0], v_] :=
    Module[{d,e},
    	d = (If[e1+SimpDiff[e1,v] > 0, 1, 0] - 
    		 If[e1 > 0, 1, 0]);
        e = Simp[Evaluate[d]];
        e
    ];    
Diff[DDList[e1_, e2_], e3_[e1_,e4__]] :=
    Module[ {indexDeltas},
        indexDeltas = MapThread[If[ MatchQ[#1,_List],
                                    1,
                                    KroneckerDelta[#1,#2]
                                ]&,
                                {e2, {e4}}
                      ];
        Part[Ch[e1], ListToSequence[IndListToPSpec[e2]]] * Times[ListToSequence[indexDeltas]]
    ] /; MatchQ[e3,Part] || MatchQ[e3,Subscript];
Diff[Abs[e1_], e2_] :=
	Module[{tmp},
		tmp = SimpDiff[e1,e2];
	    -tmp - 2*e1*If[e1 > 0, 1, 0] + 2*(e1 + tmp) If[e1 + tmp > 0, 1, 0]
	];
Diff[head_[args__], v_] :=
    Module[ {newargs,seq},
		newargs = Map[(# + SimpDiff[#,v])&, {args}];
		seq = ListToSequence[newargs];
		(* I hope this doesn't break anything? *)
		If[Intersection[Attributes[head],{HoldFirst,Hold,HoldAll}]=={},
			head[seq] - head[args],
			head[Evaluate[seq]] - head[args]
		]
    ];(* /; !MatchQ[head,ListLength];*)
Diff[e1_, v_] :=
    (cd[e1,v]);
    
On[Part::pspec];

Simp[N[e_]] := Simp[e];
Simp[Log[Product[e1_, e2__]]] :=
    Simp[Sum[Log[e1], e2]];
Simp[Log[Power[E, e1_]]] :=
    Simp[e1];
Simp[Log[e1_, Power[e1_, e2_]]] :=
    Simp[e2];
Simp[Log[es_Times]] :=
	Module[{tmp},
		tmp = TimesToList[es];
		Plus[ListToSequence[Map[Simp[Evaluate[Log[#]]]&,tmp]]]
	];
Simp[Log[Power[e1_, e2_]]] :=
    Simp[e2] * Simp[Log[e1]];
Simp[Log[e1s_Times + e2_]] :=
    Module[ {heads,list,ks,ksl={},e1l,e1,s,exp},
		ksl = Cases[TimesToList[e1s],_KroneckerDelta];
		e1l = Complement[TimesToList[e1s],ksl]; 
		ks = Times[ListToSequence[ksl]];
		e1 = Times[ListToSequence[e1l]];
        If[ ksl!={},
        	s = Log[e2] + ks(Log[e2 + e1]-Log[e2]);
        	Simp[Evaluate[s]],
			exp = Expand[e1s];
			If[MatchQ[Head[exp],Times],
	        	Log[Simp[e1s + e2]],
	        	Simp[Evaluate[Log[exp+e2]]]
			]
        ]
    ] /; !FreeQ[{e1s},KroneckerDelta];
Simp[Log[e1_]] :=
    Log[Simp[e1]];
Simp[Product[e1_,e2_]] :=
    Product[Simp[e1],e2];
Simp[Product[e1_,e2s__]] :=
    Module[ {e},
        e = Fold[Product[#1,#2]&,Product[e1,Evaluate[First[Reverse[{e2s}]]]],Rest[Reverse[{e2s}]]];
        Simp[Evaluate[e]]
    ];
Simp[Power[Power[e1_, e2_], e3_]] :=
    Simp[Power[e1,e2*e3]];
Simp[Power[e1_+e2_,i_?IntegerQ]] :=
	Module[{foo},
		foo = Sum[Binomial[i, iterz]*Power[e1,i-iterz]*Power[e2,iterz], {iterz, 0, i}];
		Simp[Evaluate[foo]]
	] /; i>0;
Simp[Power[KroneckerDelta[e1_, e2_], e3_]] :=
    KroneckerDelta[e1,e2] /; e3 > 0;
Simp[Power[e1s_Times + e2_ ,pow_?NumberQ]] :=
    Module[ {heads,list,ks,ksl={},e1l,e1,s,exp},
		ksl = Cases[TimesToList[e1s],_KroneckerDelta];
		e1l = Complement[TimesToList[e1s],ksl]; 
		ks = Times[ListToSequence[ksl]];
		e1 = Times[ListToSequence[e1l]];
        If[ ksl!={},
        	s = Power[e2,pow] + ks(Power[e2 + e1,pow]-Power[e2,pow]);
        	Simp[Evaluate[s]],
			exp = Expand[e1s];
			If[MatchQ[Head[exp],Times],
	        	Power[Simp[e1s + e2],pow],
	        	Simp[Evaluate[Power[exp+e2,pow]]]
			]
        ]
    ] /; !FreeQ[{e1s},KroneckerDelta];
Simp[Power[e1_, e2_]] :=
    Simp[e1]^Simp[e2];
Simp[Sum[e1_, e2s__List]] :=
    Module[ {e},
        e = Fold[Sum[#1,#2]&, Sum[e1, Evaluate[First[Reverse[{e2s}]]]], Rest[Reverse[{e2s}]]];
        Simp[Evaluate[e]]
    ] /; Length[{e2s}] > 1;
Simp[Sum[es_Plus, {e3_,it1_,it2_}]] :=
	Module[{each, sech,list},
		list = PlusToList[es];
		each = Map[Sum[#, {e3,it1,it2}]&, list];
		sech = Map[Simp, each];
		Plus[ListToSequence[sech]]
    ];
Simp[Sum[KroneckerDelta[e1_, e2_], {e4_,min_,max_}]] :=
   Simp[Sum[Times[1,KroneckerDelta[e1, e2]], {e4,min,max}]];
Simp[Sum[KroneckerDelta[e1_, e2_] * e3_, {e4_,min_,max_}]] :=
    Module[ {p,q,r,s},
        Which[
	        MatchQ[e1,e4],
    	    r = ReplaceAll[e3, e1->e2];
        	Simp[Evaluate[r]],
        	MatchQ[e2,e4], 
	        r = ReplaceAll[e3, e2->e1];
    	    Simp[Evaluate[r]],
        	True,
        	p = Position[Unevaluated[e1],Unevaluated[e4]];
        	q = Position[Unevaluated[e2],Unevaluated[e4]];
	       	Which[        		
		        p !=  {} && q !=  {},
        		Sum[Simp[KroneckerDelta[e1,e2]*e3],{e4,min,max}], 
		        p!=  {} || q!= {}, 
        		r = Solve[e1== e2,e4];
		        s = ReplaceAll[e3,r[[1,1]]];
        		Simp[Evaluate[s]]
        	]
        ]
    ] /; Position[Unevaluated[e1],Unevaluated[e4]]!={} || Position[Unevaluated[e2],Unevaluated[e4]]!={};
Simp[Sum[e1_ * e2_, {e3_,it1_,it2_}]] :=
    Module[ {r},
        r = Simp[Sum[e1,{e3,it1,it2}]];
        Simp[Evaluate[e2*r]]
    ] /; MyFreeQ[e2,e3];

Simp[Sum[Times[e1__], {e2_,min_,max_}]] :=
    Module[ {ex},
        ex = Expand[Times[e1]];
        If[ !MatchQ[Head[ex],Times],
            Simp[Evaluate[Sum[ex, {e2,min,max}]]],
            Sum[Simp[Times[e1]],{e2,min,max}]
        ]
    ];
Simp[Sum[e1_, {e2_,min_,max_}]] :=
	Module[{r,pass,parts,nums,nmz,RepNum,newr,sss,BackNum},
		r = Simp[e1];
		If[SameQ[r,e1],			
		    Sum[r, {e2,min,max}],
			(* this next hunk is an attempt to work around a Mathematica bug where 
			   sometimes symbolic sums with "numbers" in them cause infinite recursion. Seriously. 
			   It's been documented since at least 2005: 
			      http://forums.wolfram.com/mathgroup/archive/2005/Sep/msg00686.html
			*)		    
			parts = Position[r, piece_ /; NumberQ[piece] && !IntegerQ[piece]];
			nums  = Map[Part[r, ListToSequence[#]] &, parts];
			nmz = {};
			RepNum[exp_, pos_] :=
				Module[{nm},
					nm = SymbolNotInExpression[r, "n"];
					AppendTo[nmz, nm];
					ReplacePart[exp, pos -> nm]
				];
			newr = Fold[RepNum, r, parts];
			sss = Sum[newr, Evaluate[{e2,min,max}]];
			BackNum[exp_, {rep_, nr_}] :=
				Module[{pos},
					pos = Position[exp, rep];
					ReplacePart[exp, pos -> nr]
				];
			pass = Fold[BackNum, sss, MapThread[{#1, #2} &, {nmz, nums}]];
			(* end hack around Mathematica sucking. *)
		    		    	    
		    Simp[Evaluate[pass]]
		]
	] ;
Simp[If[e1_*KroneckerDelta[e2_,e3_]>0, 1, 0]] :=
    KroneckerDelta[e2,e3]*Simp[If[e1>0, 1, 0 ]];
Simp[If[h1_[Ch[e1_], e2__]*ks_ + h2_[e1_,e3__] > 0, 1, 0] - If[h2_[e1_,e3__]>0, 1, 0]] :=
    Module[ {heads,list},
        If[ MatchQ[Head[ks],Times],
            list = ReplacePart[ks,0->List],
            list = {ks}
        ];
        heads = Union[Map[Head,list]];
        If[ MatchQ[heads,{KroneckerDelta}],
            (If[h1[Ch[e1],e2] + h2[e1,e3] > 0, 1, 0] - 
             If[h2[e1,e3] > 0, 1, 0]) * ks,
            If[h1[Ch[e1],e2]*ks + h2[e1,e3] > 0, 1, 0] - 
            If[h2[e1,e3] > 0, 1, 0]
        ]
    ] /; (MatchQ[h1,Part]||MatchQ[h1, Subscript])&&(MatchQ[h2,Part]||MatchQ[h2, Subscript]);
Simp[If[Times[e1s__] + e2_ > 0, 1, 0]] :=
    Module[ {heads,list,ks,ksl={},e1l,e1,s},
		ksl = Cases[{e1s},_KroneckerDelta];
		e1l = Complement[{e1s},ksl]; 
		ks = Times[ListToSequence[ksl]];
		e1 = Times[ListToSequence[e1l]];
        If[ ksl!={},
        	s = Heaviside[e2] + ks(Heaviside[e2 + e1]-Heaviside[e2]);
        	Simp[Evaluate[s]],
        	Heaviside[Simp[Times[e1s] + e2]]
        ]
    ] /; !FreeQ[{e1s},KroneckerDelta];    
Simp[If[Times[-1,e1_]>0,1,0]] := Simp[1-Heaviside[e1]];
Simp[es_Plus] :=
	Module[{temp},
		temp = PlusToList[es];
	    Plus[ListToSequence[Map[Simp[Evaluate[#]]&, temp]]]
	];
Simp[KroneckerDelta[e1_,e2_]] :=
    Module[ {s},
        s = Quiet[Solve[e1==e2]];
        If[ s == {},
            0,
            KroneckerDelta[e1,e2]
        ]
    ];
Simp[es_Times] :=
	Module[{tmp},
		tmp = TimesToList[es];
	    Times[ListToSequence[Map[Simp,tmp]]]	    
	];
Simp[Abs[e1_*KroneckerDelta[e2_,e3_]]] :=
    Abs[e1]*KroneckerDelta[e2,e3];
Simp[Abs[e1_*e2_]] :=
    Simp[Abs[e1]]*Simp[Abs[e2]];
Simp[Abs[KroneckerDelta[e2_,e3_]]] :=
    KroneckerDelta[e2,e3];
Simp[Abs[KroneckerDelta[i_,y_]-KroneckerDelta[j_,y_]]] :=
    KroneckerDelta[i,y]+KroneckerDelta[j,y]-(2KroneckerDelta[i,j]KroneckerDelta[i,y]KroneckerDelta[j,y]);
Simp[Abs[e1_Plus]] :=
    Module[ {fac},
        fac = Factor[e1];
        If[ MatchQ[Head[fac],Plus],
            Abs[Simp[Plus[e1]]],
            Simp[Evaluate[Abs[fac]]]
        ]
    ];
Simp[Abs[e1_]] :=
    Abs[Simp[e1]];
Simp[h_[h_[e1_,e2__],e3__]] :=
    Simp[h[e1,e2,e3]] /; MatchQ[h,Part]||MatchQ[h,Subscript];
Simp[h_[Plus[e1_,e2_],e3__]] :=
    Simp[h[e1,e3]+h[e2,e3]] /; MatchQ[h,Part]||MatchQ[h,Subscript];
Simp[h_[Times[e1_?NumberQ,e2_],e3__]] :=
    Simp[e1*h[e2,e3]] /; MatchQ[h,Part]||MatchQ[h,Subscript];
Simp[Part[list_, spec__]] :=
    Module[ {heads,ks,ksl={},e1l,e1,s,exp,processone,processed,org,aft,kr},
        processone = 
        	Function[{arg},
        		Module[{e1s,e2},
        			If[MatchQ[arg, (e1sm_Times + e2m_) /; !FreeQ[{e1sm}, KroneckerDelta]&&((e1s=e1sm)||True)&&((e2=e2m)||True)],
						ksl = Cases[TimesToList[e1s],_KroneckerDelta];
						e1l = Complement[TimesToList[e1s],ksl]; 
						ks = Times[ListToSequence[ksl]];
						e1 = Times[ListToSequence[e1l]];
       	 				If[ ksl!={},
       	 					{e2,e1,ks},
       	 					{e2,e1s,1}       	 					
       	 				],
       	 				{arg,0,1}
        			]
        		]
        	];   
        processed = Map[processone,{spec}];
        org = Map[#[[1]]&,processed];
        aft = Map[#[[1]]+#[[2]]&,processed];
        kr = Map[#[[3]]&,processed];
        Part[list,ListToSequence[org]] + Times[ListToSequence[kr]]*(Part[list,ListToSequence[aft]]-Part[list,ListToSequence[org]])
    ] /; Cases[{spec}, (e1s_Times + e2_) /; !FreeQ[{e1s}, KroneckerDelta]]!={};
Simp[f_[h1_[Ch[e1_], e2__]*ks_ + h2_[e1_,e3__]] - f_[h2_[e1_,e3__]]] :=
    Module[ {heads,list},
        If[ MatchQ[Head[ks],Times],
            list = ReplacePart[ks,0->List],
            list = {ks}
        ];
        heads = Union[Map[Head,list]];
        If[ MatchQ[heads,{KroneckerDelta}],
            (f[h1[Ch[e1],e2] + h2[e1,e3] > 0, 1, 0] - 
             f[h2[e1,e3] > 0, 1, 0]) * ks,
            f[h1[Ch[e1],e2]*ks + h2[e1,e3] > 0, 1, 0] - 
            f[h2[e1,e3] > 0, 1, 0]
        ]
    ] /; (MatchQ[h1,Part]||MatchQ[h1, Subscript])&&(MatchQ[h2,Part]||MatchQ[h2, Subscript]);
Simp[e1_] := e1;
(*end Finite Difference Engine code *)

(*begin DD utilities code *)
MyFreeQ[e1_,e2_?AtomQ] := FreeQ[e1,e2];
MyFreeQ[Sum[e1_,{e2_,min_,max_}],e3_] :=
    MyFreeQ[e1,e3] && MyFreeQ[min,e3] && MyFreeQ[max,e3];
MyFreeQ[Product[e1_,{e2_,min_,max_}],e3_] :=
    MyFreeQ[e1,e3] && MyFreeQ[min,e3] && MyFreeQ[max,e3];
MyFreeQ[e1_+e2_,e3_] :=
    MyFreeQ[e1,e3] && MyFreeQ[e2,e3];
MyFreeQ[e1_*e2_,e3_] :=
    MyFreeQ[e1,e3]&& MyFreeQ[e2,e3];
MyFreeQ[Power[e1_,e2_],e3_] :=
    MyFreeQ[e1,e3] && MyFreeQ[e2,e3];
MyFreeQ[Part[e1_,p1__],Part[e1_,p2__]] :=
    False;
MyFreeQ[Subscript[e1_,p1__],Subscript[e1_,p2__]] :=
    False;
MyFreeQ[DDList[e1_,e2_],Part[e1_,e3__]] :=
    False;
MyFreeQ[DDList[e1_,e2_],Subscript[e1_,e3__]] :=
    False;
MyFreeQ[e1_,e2_?NumberQ] :=
    False;
MyFreeQ[e3_,Part[e1_,p2__]] :=
    MyFreeQ[e3,e1];    
MyFreeQ[Log[e1_],e2_] :=
    MyFreeQ[e1,e2];
MyFreeQ[e3_[e1__],e2_] :=
    !MatchQ[e3,e2]&&Union[Map[MyFreeQ[#1,e2]&,{e1}]]== {True};
MyFreeQ[e1_,e2_] :=
    FreeQ[e1,e2];
(* end dd utilities code *)

End[] (* End Private Context *)
EndPackage[];