(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[    707275,      13128]
NotebookOptionsPosition[    704483,      13042]
NotebookOutlinePosition[    704844,      13058]
CellTagsIndexPosition[    704801,      13055]
WindowFrame->Generic
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["\<\
The idea with this notebook is that we have some data generated by a \
stochstic process on three components, A, B and C, where A + b  --> C at some \
rate k1, and C-->A+B at some rate k2, and we want to find a likely pair of \
values for these variables. So we say that there are three distributions p1, \
p2, and p3, where p1 is over the amount of A at time t+\[Tau], given the \
amounts of A, B, and C at time t in addition to k1 and k2, and similary with \
p2 and B and p3 and C. Then we have a distribution on p1 and p2, given a time \
series for A, B, and C. We do MCMC on this to find a distribution over the \
likely values, and simulated annealing to find maximum likelihood solutions. \
\
\>", "Input",
 CellChangeTimes->{{3.418695909658144*^9, 3.418696149077655*^9}},
 FontSize->16,
 Background->RGBColor[0.09411764705882353, 0.5333333333333333, 1.]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", "\"\<Global`*\>\"", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Remove", "[", "\"\<Global`*\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"<<", "Combinatorica`"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"<<", "DDUtils.m"}], "\[IndentingNewLine]", 
 RowBox[{"<<", "FunctionMaker.m"}], "\[IndentingNewLine]", 
 RowBox[{"<<", "DependencyDiagrams.m"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"<<", "MultivariateStatistics`"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"Part", "::", "pspec"}], "]"}], ";"}]}], "Input",
 CellChangeTimes->{
  3.403835446961486*^9, {3.418658074390162*^9, 3.418658101258098*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Remove", "::", "\<\"rmnsm\"\>"}], ":", 
  " ", "\<\"There are no symbols matching \\\"\\!\\(\\\"Global`*\\\"\\)\\\". \
\\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\\\", \
ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Remove/rmnsm\\\", \
ButtonNote -> \\\"Remove::rmnsm\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.424803219695069*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"(*", " ", 
  RowBox[{
   RowBox[{"Start", " ", "off", " ", "by", " ", "defining", " ", "the", " ", 
    RowBox[{"graph", ".", " ", 
     RowBox[{"We", "'"}]}], "ll", " ", "display", " ", "it", " ", "at", " ", 
    "the", " ", "bottom"}], ",", " ", 
   RowBox[{
   "if", " ", "you", " ", "want", " ", "to", " ", "see", " ", "what", " ", 
    "it", " ", "looks", " ", 
    RowBox[{"like", ".", " ", "This"}], " ", "graph", " ", "specifies", " ", 
    "the", " ", "probability", " ", "of", " ", "the", " ", "values", " ", 
    "of", " ", "variables", " ", "A"}], ",", " ", "B", ",", " ", 
   RowBox[{
    RowBox[{"and", " ", "C", " ", "at", " ", "time", " ", "t"}], "+", "TAU"}],
    ",", " ", 
   RowBox[{
   "given", " ", "their", " ", "values", " ", "at", " ", "time", " ", "t", 
    " ", "along", " ", "with", " ", "two", " ", "rate", " ", 
    RowBox[{"parameters", "."}]}]}], " ", "*)"}]], "Input",
 CellChangeTimes->{{3.386944559631293*^9, 3.386944624056317*^9}, {
  3.386944668298732*^9, 3.386944671090572*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"RateGraph", " ", "=", " ", 
   RowBox[{"EmptyGraph", "[", 
    RowBox[{"14", ",", 
     RowBox[{"Type", "\[Rule]", "Directed"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RateGraph", " ", "=", " ", 
   RowBox[{"SetVertexLabels", "[", 
    RowBox[{"RateGraph", ",", " ", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"IndexN", "[", "i", "]"}], ",", " ", 
       RowBox[{"IndexN", "[", "j", "]"}], ",", " ", 
       RowBox[{"ConstraintN", "[", "PlusOne", "]"}], ",", " ", 
       RowBox[{"ConstantN", "[", "Aa", "]"}], ",", 
       RowBox[{"ConstantN", "[", "Bb", "]"}], ",", " ", 
       RowBox[{"ConstantN", "[", "Cc", "]"}], ",", " ", 
       RowBox[{"VariableN", "[", "k1", "]"}], ",", 
       RowBox[{"VariableN", "[", "k2", "]"}], ",", 
       RowBox[{"ConstantN", "[", "Aa", "]"}], ",", 
       RowBox[{"ConstantN", "[", "Bb", "]"}], ",", " ", 
       RowBox[{"ConstantN", "[", "Cc", "]"}], ",", 
       RowBox[{"FunctionN", "[", "p1", "]"}], ",", 
       RowBox[{"FunctionN", "[", "p2", "]"}], ",", 
       RowBox[{"FunctionN", "[", "p3", "]"}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"edges", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", "2"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "5"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "6"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "9"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "10"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2", ",", "11"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12", ",", "4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13", ",", "4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14", ",", "4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12", ",", "5"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13", ",", "5"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14", ",", "5"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12", ",", "6"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13", ",", "6"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14", ",", "6"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12", ",", "7"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13", ",", "7"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14", ",", "7"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12", ",", "8"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13", ",", "8"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14", ",", "8"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12", ",", "9"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13", ",", "10"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14", ",", "11"}], "}"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"edgeLabs", " ", "=", " ", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "IndexCons", ",", "IndexCons", ",", "IndexL", ",", "IndexL", ",", 
      "IndexL", ",", "IndexL", ",", "IndexL", ",", "IndexL", ",", "CondDep", 
      ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
      "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
      "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
      "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
      "CondDep"}], "}"}], "/.", 
    RowBox[{"CondDep", "\[Rule]", "FactorEdge"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ee", " ", "=", " ", 
   RowBox[{"MapThread", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"{", 
       RowBox[{"#1", ",", " ", 
        RowBox[{"EdgeLabel", "\[Rule]", "#2"}], ",", " ", 
        RowBox[{"EdgeDirection", "\[Rule]", "True"}]}], "}"}], "&"}], ",", 
     RowBox[{"{", 
      RowBox[{"edges", ",", "edgeLabs"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Set", "[", 
       RowBox[{"RateGraph", ",", 
        RowBox[{"AddEdge", "[", 
         RowBox[{"RateGraph", ",", "#1"}], "]"}]}], "]"}], "&"}], ",", "ee"}],
     "]"}], ";"}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"Functions", " ", "which", " ", "specify", " ", "p", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"A_t2", "|", "A_t1"}], ",", "B_t1", ",", "C_t1", ",", "k1", 
       ",", "k2"}], ")"}], " ", "and", " ", "similarly", " ", "for", " ", "B",
      " ", "and", " ", 
     RowBox[{"C", ".", " ", "Basically"}]}], ",", " ", 
    RowBox[{
    "the", " ", "distributions", " ", "are", " ", "assumed", " ", "to", " ", 
     "be", " ", "normal", " ", "around", " ", "the", " ", "mean", " ", 
     "given", " ", "by", " ", "the", " ", "change", " ", "predicted", " ", 
     "by", " ", "the", " ", "rate", " ", 
     RowBox[{"parameters", "."}]}]}], " ", "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"p1", "[", 
     RowBox[{
     "a1_", ",", "a2_", ",", "b1_", ",", "c1_", ",", "k1_", ",", "k2_"}], 
     "]"}], ":=", "\[IndentingNewLine]", 
    RowBox[{"E", "^", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"a2", "-", 
          RowBox[{"(", 
           RowBox[{"a1", "-", 
            RowBox[{"(", 
             RowBox[{"a1", "*", "b1", "*", 
              RowBox[{"(", 
               RowBox[{"k1", "*", "TAU"}], ")"}]}], ")"}], "+", 
            RowBox[{"(", 
             RowBox[{"c1", "*", 
              RowBox[{"(", 
               RowBox[{"k2", "*", "TAU"}], ")"}]}], ")"}]}], ")"}]}], ")"}], 
        "^", "2"}], "*", 
       RowBox[{"-", "SIGMA"}]}], ")"}]}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"p2", "[", 
     RowBox[{
     "a1_", ",", "b1_", ",", "b2_", ",", "c1_", ",", "k1_", ",", "k2_"}], 
     "]"}], ":=", "\[IndentingNewLine]", 
    RowBox[{"E", "^", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"b2", "-", 
          RowBox[{"(", 
           RowBox[{"b1", "-", 
            RowBox[{"(", 
             RowBox[{"a1", "*", "b1", "*", 
              RowBox[{"(", 
               RowBox[{"k1", "*", "TAU"}], ")"}]}], ")"}], "+", 
            RowBox[{"(", 
             RowBox[{"c1", "*", 
              RowBox[{"(", 
               RowBox[{"k2", "*", "TAU"}], ")"}]}], ")"}]}], ")"}]}], ")"}], 
        "^", "2"}], "*", 
       RowBox[{"-", "SIGMA"}]}], ")"}]}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"p3", "[", 
    RowBox[{
    "a1_", ",", "b1_", ",", "c1_", ",", "c2_", ",", "k1_", ",", "k2_"}], 
    "]"}], ":=", "\[IndentingNewLine]", 
   RowBox[{"E", "^", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"c2", "-", 
         RowBox[{"(", 
          RowBox[{"c1", "+", 
           RowBox[{"(", 
            RowBox[{"a1", "*", "b1", "*", 
             RowBox[{"(", 
              RowBox[{"k1", "*", "TAU"}], ")"}]}], ")"}], "-", 
           RowBox[{"(", 
            RowBox[{"c1", "*", 
             RowBox[{"(", 
              RowBox[{"k2", "*", "TAU"}], ")"}]}], ")"}]}], ")"}]}], ")"}], 
       "^", "2"}], "*", 
      RowBox[{"-", "SIGMA"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.386619827930681*^9, 3.386620354202311*^9}, {
  3.386620392626559*^9, 3.386620507727502*^9}, {3.386620554565209*^9, 
  3.386620559642351*^9}, {3.386620673227277*^9, 3.386620678523681*^9}, {
  3.386621353688524*^9, 3.386621354142046*^9}, {3.386621407267304*^9, 
  3.386621438089724*^9}, {3.386621783693342*^9, 3.386621793780831*^9}, {
  3.386621844817494*^9, 3.38662184946043*^9}, {3.41865796982361*^9, 
  3.418657990897271*^9}, {3.418658052332537*^9, 3.41865805833057*^9}, {
  3.418658188338639*^9, 3.418658189492585*^9}, {3.418658463526882*^9, 
  3.418658596840254*^9}, {3.418658646303038*^9, 3.41865868041095*^9}, {
  3.418660576456519*^9, 3.418660579115175*^9}, {3.418660649459992*^9, 
  3.418660652452828*^9}, {3.418669074191256*^9, 3.418669103566634*^9}}],

Cell["\<\
This is the original time series data. Each list in x is an amount of {A, B, \
C}.\
\>", "Input",
 CellChangeTimes->{{3.418696190728093*^9, 3.418696213218832*^9}},
 FontSize->16],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{"k1", ",", "k2", ",", "Aa", ",", "Bb", ",", "Cc"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"200", ",", "100", ",", "100"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"200", ",", "100", ",", "100"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"200", ",", "100", ",", "100"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"199", ",", "99", ",", "101"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"199", ",", "99", ",", "101"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"198", ",", "98", ",", "102"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"194", ",", "94", ",", "106"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"194", ",", "94", ",", "106"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"193", ",", "93", ",", "107"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"193", ",", "93", ",", "107"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"191", ",", "91", ",", "109"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"191", ",", "91", ",", "109"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"190", ",", "90", ",", "110"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"189", ",", "89", ",", "111"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"188", ",", "88", ",", "112"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"187", ",", "87", ",", "113"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"187", ",", "87", ",", "113"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"187", ",", "87", ",", "113"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"187", ",", "87", ",", "113"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"187", ",", "87", ",", "113"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"186", ",", "86", ",", "114"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"186", ",", "86", ",", "114"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"185", ",", "85", ",", "115"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"183", ",", "83", ",", "117"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"183", ",", "83", ",", "117"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"183", ",", "83", ",", "117"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"182", ",", "82", ",", "118"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"180", ",", "80", ",", "120"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"178", ",", "78", ",", "122"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"178", ",", "78", ",", "122"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"178", ",", "78", ",", "122"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"175", ",", "75", ",", "125"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"175", ",", "75", ",", "125"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"174", ",", "74", ",", "126"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"170", ",", "70", ",", "130"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"169", ",", "69", ",", "131"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"166", ",", "66", ",", "134"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"165", ",", "65", ",", "135"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"165", ",", "65", ",", "135"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"166", ",", "66", ",", "134"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"165", ",", "65", ",", "135"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"165", ",", "65", ",", "135"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"165", ",", "65", ",", "135"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"164", ",", "64", ",", "136"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"163", ",", "63", ",", "137"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"163", ",", "63", ",", "137"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"163", ",", "63", ",", "137"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"164", ",", "64", ",", "136"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"164", ",", "64", ",", "136"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"164", ",", "64", ",", "136"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"162", ",", "62", ",", "138"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"161", ",", "61", ",", "139"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"160", ",", "60", ",", "140"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"160", ",", "60", ",", "140"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"160", ",", "60", ",", "140"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"160", ",", "60", ",", "140"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"159", ",", "59", ",", "141"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"159", ",", "59", ",", "141"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"158", ",", "58", ",", "142"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"157", ",", "57", ",", "143"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"157", ",", "57", ",", "143"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"156", ",", "56", ",", "144"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"154", ",", "54", ",", "146"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"154", ",", "54", ",", "146"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"154", ",", "54", ",", "146"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"153", ",", "53", ",", "147"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"153", ",", "53", ",", "147"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"151", ",", "51", ",", "149"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"150", ",", "50", ",", "150"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"150", ",", "50", ",", "150"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"148", ",", "48", ",", "152"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"148", ",", "48", ",", "152"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"148", ",", "48", ",", "152"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"147", ",", "47", ",", "153"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"147", ",", "47", ",", "153"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"147", ",", "47", ",", "153"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"147", ",", "47", ",", "153"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"147", ",", "47", ",", "153"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"145", ",", "45", ",", "155"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"145", ",", "45", ",", "155"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"145", ",", "45", ",", "155"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"142", ",", "42", ",", "158"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"141", ",", "41", ",", "159"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"141", ",", "41", ",", "159"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"141", ",", "41", ",", "159"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"141", ",", "41", ",", "159"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140", ",", "40", ",", "160"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140", ",", "40", ",", "160"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"139", ",", "39", ",", "161"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"139", ",", "39", ",", "161"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"139", ",", "39", ",", "161"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140", ",", "40", ",", "160"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140", ",", "40", ",", "160"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140", ",", "40", ",", "160"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140", ",", "40", ",", "160"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"139", ",", "39", ",", "161"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"138", ",", "38", ",", "162"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"138", ",", "38", ",", "162"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"138", ",", "38", ",", "162"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"138", ",", "38", ",", "162"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137", ",", "37", ",", "163"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"136", ",", "36", ",", "164"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135", ",", "35", ",", "165"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135", ",", "35", ",", "165"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135", ",", "35", ",", "165"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135", ",", "35", ",", "165"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135", ",", "35", ",", "165"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"136", ",", "36", ",", "164"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"136", ",", "36", ",", "164"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135", ",", "35", ",", "165"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134", ",", "34", ",", "166"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133", ",", "33", ",", "167"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133", ",", "33", ",", "167"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132", ",", "32", ",", "168"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132", ",", "32", ",", "168"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132", ",", "32", ",", "168"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131", ",", "31", ",", "169"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131", ",", "31", ",", "169"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131", ",", "31", ",", "169"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132", ",", "32", ",", "168"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133", ",", "33", ",", "167"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133", ",", "33", ",", "167"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133", ",", "33", ",", "167"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133", ",", "33", ",", "167"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132", ",", "32", ",", "168"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132", ",", "32", ",", "168"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131", ",", "31", ",", "169"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131", ",", "31", ",", "169"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131", ",", "31", ",", "169"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131", ",", "31", ",", "169"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130", ",", "30", ",", "170"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"129", ",", "29", ",", "171"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"129", ",", "29", ",", "171"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128", ",", "28", ",", "172"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127", ",", "27", ",", "173"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"119", ",", "19", ",", "181"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"119", ",", "19", ",", "181"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"119", ",", "19", ",", "181"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"119", ",", "19", ",", "181"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126", ",", "26", ",", "174"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120", ",", "20", ",", "180"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121", ",", "21", ",", "179"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122", ",", "22", ",", "178"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125", ",", "25", ",", "175"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124", ",", "24", ",", "176"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123", ",", "23", ",", "177"}], "}"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"ListPlot", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"#", "[", 
        RowBox[{"[", "1", "]"}], "]"}], "&"}], ",", "x"}], "]"}], ",", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"#", "[", 
        RowBox[{"[", "2", "]"}], "]"}], "&"}], ",", "x"}], "]"}], ",", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"#", "[", 
        RowBox[{"[", "3", "]"}], "]"}], "&"}], ",", "x"}], "]"}]}], "}"}], 
  "]"}]}], "Input",
 CellChangeTimes->{{3.386944634821483*^9, 3.38694470411019*^9}, {
   3.418659516580608*^9, 3.418659553078867*^9}, {3.418659596106399*^9, 
   3.418659637592385*^9}, {3.418659732804596*^9, 3.418659800511845*^9}, {
   3.418659893148766*^9, 3.418659940511068*^9}, {3.41865997360844*^9, 
   3.418659978410039*^9}, 3.418666417982255*^9, {3.418668140898595*^9, 
   3.418668171107244*^9}, {3.418668540466921*^9, 3.418668543485522*^9}, {
   3.418668776231439*^9, 3.418668779652106*^9}, 3.418668892324281*^9, {
   3.418669112057874*^9, 3.41866911242257*^9}, {3.418673905838418*^9, 
   3.418673907601975*^9}, {3.418674032701164*^9, 3.418674047966196*^9}, {
   3.418674096153595*^9, 3.418674121496903*^9}, {3.418674896804655*^9, 
   3.418674917852981*^9}, {3.418675223951637*^9, 3.418675224279381*^9}, {
   3.418675651277346*^9, 3.418675747931109*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, 
   {Hue[0.67, 0.6, 0.6], PointBox[CompressedData["
1:eJxd10mS5TQQgOGEFUsvWLBgIQiCIAiCMPNMC2jmyTRjMzSiih7opijTxdBN
UaAj+Ag6go+gI/gIOoK37AjCmSbyfxvF53QqJU/Su+/g+MLlO0XknztE/mu3
33pua3+IeiB63wV3W9uuqe+G79naar53a6M5wPdvbTA/AD+o/V9VPwQ/rPXM
j2xtMT+6tdncb20yPwY/Dj8BPwk/pfMxPw0/o/MxP6vjv6J+Dn4efkHnZ35R
52N+SeuZX4bPwVH7v6z9wPKK1rO4Olr8VZ2PxdX1e42/trVFnWE578/P5xF/
HfmwvKHHLa5OFn/TO8PyFvJheRtxdbT4O1sbLK4Wi7/rnWF5D3F1O9Tj73tn
dbX4B1tbLA7Lh3rc4upo8cE7wlkd1BWWj3R8B5qnruoMV3VWywXvCGd1snxY
PvaOcIarOlj+J94Rzur6nbbqopZPvSOc4arOlv+Zd1Qny4crLJ8jH85wheUL
319UR8uHKywXtzZYPpwvIh9x+RL5cIYrLF8hH85qsXx1S3r8a+8IZ7jC8o0e
t3w4wxWWS8iHM1xh+Rb5cIYrLMk7wBFOcIYLvNfR699gHt/vs9WHE2zPsdUt
sL23Ns4Gy4G3vfc2zwgneP9OWf0D1IcbLIfe9l21/iKcDlEfLnCFbR2x+2rr
jjnAEU5whgtc4QaL7hvsubN9hD2XEU5qe44z4gXxCjdYrngH2PZJ9h4lOMMF
rvC+j7P5X/W2faFdj4h4gjNc4Ao3WK55BzjCCc5qu14FrnCD7X/Efv3hqN6v
P5zhAlf7n2L1Ybmu5+l3NVz38QgnOMMF/VW1fceb2r7zcsM7wBFOcL7h+y9w
veHH02D50TvAEU5whgtc4QbLEeof+fHHIz/fBGe4wFVt62yD5SfvAEc4qW0d
z3CBK/IbLMd+vOEY9eEEZ7gcoz7iDZbR1+/goLb70cMRHuAEj3CGJ7jAM1zh
BeNv8ArLzz6/gwPcq/f3Ax7gBI9whie4wDNc4QVu8Krev383vTs4wD0c4eGm
r5fgEedneIKL2va1s3r/vsML3OAVlhPv7sTXD3APxxM/vwFO8Gj17P7DE1zU
9rzOcIUXuMErLL94d3CAezjCA5zgEc7wBBd4Vu/vP+IL3OAV+fKrdwcHuIej
er//cFLb8zLCGZ7gAs9whRfUb/CK8+U3705t71uAezjCA5zgEc5qe/8muMAz
XOEF/Td4heV37w4OcA9HeFDv+0V4hDM8wQX9z3CFF7jBKyy3vDu1Xc8A93BE
/nDLjz/BI5zhCS7wjHoVXuAGr7Dc9u7gcNvX7+EID3CCRzjDE1zgGa7wAjd4
heUP7w4OcA9H9X7/4YTzRzjDE1zQ3wxXeFHv6z+8wnLq3cEB7uEID6d+PAke
4Xzq5zshXuAZrvACN3hFPfnTu4OD2taPHo7woN73//AIZ3iCCzyr9/0PvMAN
XmE509bWfzjAPRzhAU5nvt545ueT1fv+Hy7wDFd4sfp2/xFfYfnLn9/BAe7V
+/oHD3CCRzij/wku8AxXeIEbvMLyt3cHB7iHIzyo9/0fPMIZnv73v1zpDBQ=

     "]]}, 
   {Hue[0.9060679774997897, 0.6, 0.6], PointBox[CompressedData["
1:eJxl2Emy3TQUxvEDI4YeMGDAQFAURVEUZfqeiBB6AualIaEJyguhCU1M34OW
4CVoCV6CluAlaAmeMqOoe75H6c+dqH4+PpJlS7J87zq+eXT9VjP7+xazf8vD
bz91KC9HPxB73wYPh7Ject8O33Eos3yn+113gO8+lFG+B77X67/ovg++3+uT
H/B8+UEv5dHzL7gfgh+GH4EfhR/z9uXH4Sf8euQnPf+8+yn4afgZr09+1kv5
OY+fcz8Pn4Kj13/kx2F7wY8rLr/j8dPeH3d258njL3rcnWE705+fzyD+EvJh
e9lLxd35bT/+Su8M26vIh+01xGXlv+7H3/LSbW57o3eG7U3E3fmsHz/bm8f/
lwefXKfisq5j6h3hk+fi/aqw6bn7fdE40H3LcJV1n496R/hk3Plzq7Cd6x3h
DFdZ4+Z87wifzDMfd5qHGqd2oXeEM3yyTij/Ym+tU5onGa6waV1UPpzhCtul
vr6odVrtwxXWe0DzPsL5cn9+RdzeQz6c4Qrb+8iHs1vrVFXc1zH7oHeEM1xh
+9BL5cMZrrBdQT6c4QrbR8iHM1xhS70DHOEEZ7jAVT59KBtsV70etX8V7cMJ
zsr391aBq853N9iOewfZ35MRTnBGfjlG+3CD7VrvcK2vL8JJVvtwgSvcZN8X
2Me9AxzhBGe4wBVusF330vctQfZ9TYSTrH0Q4gXxCjfYPukd4Cj7PizBGS5w
hZusfdunvYPs9yMinuAMF7jCDbbP0D4c4QRn2e9XgSvcYPu8d4CjrPsPZ7jA
VfZ9doPthpe+Lw83+niEE5zhgvqq7N8BTfbvBPuid4AjnOAse/0FrrJfT4Pt
y94BjnCCM1zgCjfYvkL7sl9/lNV/OMMFrrJ/pzXYvu4d4Agn2b8DM1zgivwG
283+esPNPh7hBGe4yGof8Qbb3Lc/wEH25zHCEZ7gBM9whhe4wCtc4Q3X3+Ad
tm/6/AEO8ChrfsATnOAZzvACF3iFK7zBDd5lrX/f9h7gAI9whCdZ/YdnnJ/h
BS6y/y+yylrf4Q1u8A7bd70HWf2HRzjKev5wgmdZ4x9e4CL7eF3hCm9wg3fY
vu89wAEe4QhPcIJnOMMLXOBV1vxHfIMbvCPffug9wAEe4Sjr+cNJ9vEywxle
4AKvcIU3tN/gHefbj70H2edbgEc4whOc4BnOss+/BS7wCld4Q/0N3mH7qfcA
B3iEIzzJ2i/CM5zhBS6of4UrvMEN3mH7ufcg+/0M8AhH5E+y+g/PcIYXuMAr
2qvwBjd4h+2X3gMcZK3/cIQnOMEznOEFLvAKV3iDG7zD9mvvAQ7wCEdZzx9O
OH+GM7zABfWtcIU3We9/eIftt94DHOARjvAkq//wDGdZ/Ue8wCtc4Q1u8I72
7PfeAxxkf3+McIQnWft/eIYzvMAFXmXtf+ANbvAO2x9e+vfQAAd4hCM8wUnW
/kdW/2Xt/+ECr3CFN1nzH/Edtj/78wc4wKOs9x88wQme4Yz6F7jAK1zhDW7w
DttfvQc4wCMc4UnW/g+e4Qwv//kfaqo8Zw==
     "]]}, 
   {Hue[0.14213595499957954`, 0.6, 0.6], PointBox[CompressedData["
1:eJxd1knS3TQQwHHBiqUXLFiwEBRFURRFmXkmAsI8mTAnIREhAySEzyQMCR+B
PoKPoCP4CDqCj6AjeMuOol63qP6/TdfP7VZLfrbsBy4cnbp0dwjhn7tC+C8e
fvuJQ/w66YHkfQ88HGIy3wvfd4hivl99Wh3hBw+xmh+CH9bxz6gfgR/V8cyP
ab35cY1nNY5ab34CfhJ+Cn4afkb7m5+Fn9P5mJ/X+m/UL8Avwi/peOaXNZ7T
+Irmza/CJ+Ck45/X43B4TY9bXp2y5l8/xKwWdbX8G4fYLA+Hk/58OYn8m6iH
w1sav9W8OqrD294Ch3dQD4d3kVcny793iNnyarH8+94Chw+QVxfLf+gt6mr5
jw6xWR4OH2u8oHl1UofJO8GizuoKh090flavrlYPV3X4TuMp7wSLOqorHD71
TrDAVZ2t/jPvBIu6Wr26Wf3n3gkWuKrDRY1feCd1VAtc4fAl6mGBKxy+8uMl
dbJ6uML2HshWD9u+3+uRD6dRDwtc4XAG9bC9B8Tq1cXqz3onWOAKB92Hq9XD
AlfY9u1eDwtc4XAe9bDAFQ7ZO8IJzrDABe77urrBtq/a9bR9ufeHM2z7ZrP+
cN8nL+lx2PZFc1RHdYIzLKi3fbD3hxts+5y572vWH7Z9qfeHC1zhvk/Z+i96
RzjBGRa4wBVucF+X9bdxL/nzBOcVHBcct1jhBofL3hFOl/04GRa4wBVul/18
wxXveMWvMyGfYYELXOEGh+/RH05whkXdrztc4QaHH7wjnNT9+sMCF7iq7X9u
cLiqfdXxqs8nOMMCF4xX1XYfNbXdR+Gad4QTnGG55scvcL3m59Pg8KN3hBOc
YYELXOEGh+vof93PP133682wwAWuantOGxx+8o5wgrPannOBC1xR3+Bw5Ocb
j9AfzrDA5Qj9kW9wmH3/AY5q+z9GOMETnOEZFniBC7zCFd4w/wbvcPjZ1w9w
hEd1fz7gCc7wDAu8wAVe4QpvcIN3dd//bngPcIRHOMHTDd8vwzPOF3iBi9re
26u67+/wBjd4h8NN7+Gm7x/hEU43/fomOMOzut//8AIXtd2vK1zhDW7wDodf
vAc4wiOc4AnO8AwLvMAFXtX9+Ud+gxu8oz786j3AER7hpO7/P5zVdr/MsMAL
XOAVrvCG/g3ecX74zXtQ2/MW4RFO8ARneIZFbc/fAhd4hSu8YfwG73D43XuA
IzzCCZ7Udj0zPMMCL3DB+Ctc4Q1u8A6HW96D2q5nhEc4oX665eef4RkWeIEL
vKJfhTe4wTscbnsPcLzt+49wgic4wzMs8AIXeIUrvMEN3uHwh/cAR3iEk7r/
/3DG+TMs8AIXjLfCFd7U/f0P73A49h7gCI9wgqdjP58Mz7Ac+/UuyBd4hSu8
wQ3e0S/86T3AUW3vjxFO8KTu3//wDAu8wAVe1f37B97gBu9wuKPzsPXDER7h
BE9wvuP7zXf8ekTdv//hAq9whTd1f/6R3+Hwlz9/gCM8qvv7D57gDM+wYPwF
LvAKV3iDG7zD4W/vAY7wCCd4UvfvP3iGBV7+97+/eB+k
     "]]}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  ImageSize->{638., Automatic},
  PlotRange->{{0., 505.}, {0., 200.}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{{3.418675737022345*^9, 3.418675766990401*^9}, 
   3.418676255635548*^9, 3.418695132633783*^9, 3.424801728803699*^9, 
   3.424802588408091*^9, 3.424803148658148*^9, 3.424803230430872*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{
   "Here", " ", "I", " ", "copy", " ", "the", " ", "data", " ", "into", " ", 
    "the", " ", "constants", " ", "expected", " ", "by", " ", "the", " ", 
    "code", " ", "generated", " ", "from", " ", "the", " ", "graph"}], "*)"}],
   "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"l", "=", 
     RowBox[{"Length", "[", "x", "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"NUMDATA", "=", 
     RowBox[{"Min", "[", 
      RowBox[{"l", ",", "200"}], "]"}]}], ";"}], "\[IndentingNewLine]", "\n", 
   
   RowBox[{
    RowBox[{"ITERZ", "=", "250"}], ";"}], "\[IndentingNewLine]", "\n", 
   RowBox[{"(*", 
    RowBox[{
     RowBox[{
     "Defining", " ", "the", " ", "variables", " ", "in", " ", "the", " ", 
      RowBox[{"graph", ".", "Note"}], " ", "that", " ", 
      RowBox[{"I", "'"}], "m", " ", "saying", " ", 
      RowBox[{"they", "'"}], "re", " ", "both", " ", "continuous", " ", 
      "between", " ", "0", " ", "and", " ", "1"}], ",", 
     RowBox[{
     "which", " ", "should", " ", "make", " ", "it", " ", "more", " ", 
      "impressive", " ", "when", " ", "we", " ", "later", " ", "learn", " ", 
      RowBox[{"they", "'"}], "re", " ", "of", " ", "different", " ", "orders",
       " ", "of", " ", 
      RowBox[{"magnitude", "."}]}]}], "*)"}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"VARIABLES", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"k1", ",", "Continuous", ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "1"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"k2", ",", "Continuous", ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "1"}], "}"}]}], "}"}]}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", "\n", 
   RowBox[{"(*", 
    RowBox[{
     RowBox[{"Two", " ", "indices"}], ",", 
     RowBox[{
     "both", " ", "over", " ", "the", " ", "number", " ", "of", " ", "data", 
      " ", "points", " ", "we", " ", "care", " ", 
      RowBox[{"about", "."}]}]}], "*)"}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"INDICES", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"i", ",", "1", ",", "NUMDATA"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"j", ",", "1", ",", "NUMDATA"}], "}"}]}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", "\n", 
   RowBox[{"(*", 
    RowBox[{
    "Then", " ", "generate", " ", "the", " ", "MCMC", " ", "code", " ", "to", 
     " ", "sample", " ", "the", " ", "posterior", " ", "probabilities", " ", 
     "of", " ", "the", " ", "rate", " ", "parameters"}], "*)"}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"code", "=", 
     RowBox[{"GenerateMCMCCode", "[", 
      RowBox[{
      "RateGraph", ",", "VARIABLES", ",", "INDICES", ",", "ITERZ", ",", 
       RowBox[{"{", "}"}], ",", 
       RowBox[{"{", "}"}], ",", "True"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Aa", " ", "=", " ", 
     RowBox[{"Table", "[", 
      RowBox[{"0", ",", 
       RowBox[{"{", "NUMDATA", "}"}]}], "]"}]}], ";"}], "\n", 
   RowBox[{
    RowBox[{"Bb", " ", "=", " ", 
     RowBox[{"Table", "[", 
      RowBox[{"0", ",", 
       RowBox[{"{", "NUMDATA", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]",
    
   RowBox[{
    RowBox[{"Cc", " ", "=", " ", 
     RowBox[{"Table", "[", 
      RowBox[{"0", ",", 
       RowBox[{"{", "NUMDATA", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]",
    
   RowBox[{
    RowBox[{"For", "[", 
     RowBox[{
      RowBox[{"it", "=", "1"}], ",", 
      RowBox[{"it", "\[LessEqual]", "NUMDATA"}], ",", 
      RowBox[{"it", "++"}], ",", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"Part", "[", 
         RowBox[{"Aa", ",", "it"}], "]"}], "=", 
        RowBox[{"x", "[", 
         RowBox[{"[", 
          RowBox[{"it", ",", "1"}], "]"}], "]"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Part", "[", 
         RowBox[{"Bb", ",", "it"}], "]"}], "=", 
        RowBox[{"x", "[", 
         RowBox[{"[", 
          RowBox[{"it", ",", "2"}], "]"}], "]"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Part", "[", 
         RowBox[{"Cc", ",", "it"}], "]"}], "=", 
        RowBox[{"x", "[", 
         RowBox[{"[", 
          RowBox[{"it", ",", "3"}], "]"}], "]"}]}], ";"}]}], 
     "\[IndentingNewLine]", "]"}], ";"}]}]}]], "Input",
 CellChangeTimes->{{3.386944634821483*^9, 3.38694470411019*^9}, {
   3.418659516580608*^9, 3.418659553078867*^9}, {3.418659596106399*^9, 
   3.418659637592385*^9}, {3.418659732804596*^9, 3.418659800511845*^9}, {
   3.418659893148766*^9, 3.418659940511068*^9}, {3.41865997360844*^9, 
   3.418659978410039*^9}, 3.418666417982255*^9, {3.418668140898595*^9, 
   3.418668171107244*^9}, {3.418668540466921*^9, 3.418668543485522*^9}, {
   3.418668776231439*^9, 3.418668779652106*^9}, 3.418668892324281*^9, {
   3.418669112057874*^9, 3.41866911242257*^9}, {3.418673905838418*^9, 
   3.418673907601975*^9}, {3.418674032701164*^9, 3.418674047966196*^9}, {
   3.418674096153595*^9, 3.418674121496903*^9}, {3.418674896804655*^9, 
   3.418674917852981*^9}, {3.418675223951637*^9, 3.418675224279381*^9}, {
   3.418675651277346*^9, 3.418675676814069*^9}, {3.418695147070232*^9, 
   3.418695147182612*^9}, {3.418699781240021*^9, 3.418699781635941*^9}, {
   3.424803130930771*^9, 3.424803131179142*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 SuperscriptBox["\[ExponentialE]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "SIGMA"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k10"}], "+", "k11"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         SuperscriptBox[
          RowBox[{
          "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], "2"], 
         " ", 
         RowBox[{
         "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "+", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k10"}], "+", "k11"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{"Aa", "\[LeftDoubleBracket]", 
          RowBox[{"1", "+", "i"}], "\[RightDoubleBracket]"}], " ", 
         RowBox[{
         "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "k10", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k10"}], "+", "k11"}], ")"}]}], "+", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k10"}], "+", "k11"}], ")"}], "2"]}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         SuperscriptBox[
          RowBox[{
          "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], "2"], 
         " ", 
         SuperscriptBox[
          RowBox[{
          "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
          "2"]}]}]}], "-", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k10"}], "+", "k11"}], ")"}], " ", "k2", " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{"Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}]}],
      ")"}]}], "-", 
   RowBox[{"SIGMA", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k10"}], "+", "k11"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         SuperscriptBox[
          RowBox[{
          "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
          "2"]}]}]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "k10", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k10"}], "+", "k11"}], ")"}]}], "+", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k10"}], "+", "k11"}], ")"}], "2"]}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         SuperscriptBox[
          RowBox[{
          "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], "2"], 
         " ", 
         SuperscriptBox[
          RowBox[{
          "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
          "2"]}]}]}], "+", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k10"}], "+", "k11"}], ")"}], " ", "TAU", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          UnderoverscriptBox["\[Sum]", 
           RowBox[{"i", "=", "imin0"}], "imax0"], 
          RowBox[{
           RowBox[{
           "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ", 
           
           RowBox[{
           "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ", 
           
           RowBox[{"Bb", "\[LeftDoubleBracket]", 
            RowBox[{"1", "+", "i"}], "\[RightDoubleBracket]"}]}]}], "-", 
         RowBox[{"k2", " ", "TAU", " ", 
          RowBox[{
           UnderoverscriptBox["\[Sum]", 
            RowBox[{"i", "=", "imin0"}], "imax0"], 
           RowBox[{
            RowBox[{
            "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ",
             
            RowBox[{
            "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ",
             
            RowBox[{
            "Cc", "\[LeftDoubleBracket]", "i", 
             "\[RightDoubleBracket]"}]}]}]}]}], ")"}]}]}], ")"}]}], "-", 
   RowBox[{"SIGMA", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "k10", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k10"}], "+", "k11"}], ")"}]}], "+", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k10"}], "+", "k11"}], ")"}], "2"]}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         SuperscriptBox[
          RowBox[{
          "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], "2"], 
         " ", 
         SuperscriptBox[
          RowBox[{
          "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
          "2"]}]}]}], "-", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k10"}], "+", "k11"}], ")"}], " ", "TAU", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          RowBox[{
           UnderoverscriptBox["\[Sum]", 
            RowBox[{"i", "=", "imin0"}], "imax0"], 
           RowBox[{
            RowBox[{
            "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ",
             
            RowBox[{
            "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ",
             
            RowBox[{
            "Cc", "\[LeftDoubleBracket]", "i", 
             "\[RightDoubleBracket]"}]}]}]}], "+", 
         RowBox[{"k2", " ", "TAU", " ", 
          RowBox[{
           UnderoverscriptBox["\[Sum]", 
            RowBox[{"i", "=", "imin0"}], "imax0"], 
           RowBox[{
            RowBox[{
            "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ",
             
            RowBox[{
            "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ",
             
            RowBox[{
            "Cc", "\[LeftDoubleBracket]", "i", 
             "\[RightDoubleBracket]"}]}]}]}], "+", 
         RowBox[{
          UnderoverscriptBox["\[Sum]", 
           RowBox[{"i", "=", "imin0"}], "imax0"], 
          RowBox[{
           RowBox[{
           "Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ", 
           
           RowBox[{
           "Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], " ", 
           
           RowBox[{"Cc", "\[LeftDoubleBracket]", 
            RowBox[{"1", "+", "i"}], "\[RightDoubleBracket]"}]}]}]}], 
        ")"}]}]}], ")"}]}]}]]], "Print",
 CellChangeTimes->{3.424801764628841*^9, 3.424802608229808*^9, 
  3.424803164580593*^9, 3.424803246626459*^9}],

Cell[BoxData[
 SuperscriptBox["\[ExponentialE]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "SIGMA"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "-", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", 
          RowBox[{"1", "+", "i"}], "\[RightDoubleBracket]"}], " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "-", 
      RowBox[{"2", " ", "k1", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{"Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "k20", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k20"}], "+", "k21"}], ")"}]}], "+", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k20"}], "+", "k21"}], ")"}], "2"]}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        SuperscriptBox[
         RowBox[{"Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          "2"]}]}]}], ")"}]}], "-", 
   RowBox[{"SIGMA", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "-", 
      RowBox[{"2", " ", "k1", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{"Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "-", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Bb", "\[LeftDoubleBracket]", 
          RowBox[{"1", "+", "i"}], "\[RightDoubleBracket]"}], " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "k20", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k20"}], "+", "k21"}], ")"}]}], "+", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k20"}], "+", "k21"}], ")"}], "2"]}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        SuperscriptBox[
         RowBox[{"Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          "2"]}]}]}], ")"}]}], "-", 
   RowBox[{"SIGMA", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", "k1", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Aa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{"Bb", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{
         "Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}], 
      "-", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        SuperscriptBox[
         RowBox[{"Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          "2"]}]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "k20", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k20"}], "+", "k21"}], ")"}]}], "+", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "k20"}], "+", "k21"}], ")"}], "2"]}], ")"}], " ", 
       SuperscriptBox["TAU", "2"], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        SuperscriptBox[
         RowBox[{"Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          "2"]}]}], "+", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "k20"}], "+", "k21"}], ")"}], " ", "TAU", " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "imin0"}], "imax0"], 
        RowBox[{
         RowBox[{"Cc", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
          " ", 
         RowBox[{"Cc", "\[LeftDoubleBracket]", 
          RowBox[{"1", "+", "i"}], "\[RightDoubleBracket]"}]}]}]}]}], 
     ")"}]}]}]]], "Print",
 CellChangeTimes->{3.424801764628841*^9, 3.424802608229808*^9, 
  3.424803164580593*^9, 3.424803258447437*^9}]
}, Open  ]]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"SIGMA", "=", "1"}], ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{
   "The", " ", "size", " ", "of", " ", "the", " ", "time", " ", "steps", " ", 
    "in", " ", "the", " ", 
    RowBox[{"data", "."}]}], " ", "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"TAU", " ", "=", " ", ".001"}], ";"}]}], "Input",
 CellChangeTimes->{{3.386621377676868*^9, 3.386621398897713*^9}, {
   3.386621449207001*^9, 3.386621723174256*^9}, {3.386621966742207*^9, 
   3.386621967619972*^9}, {3.386622982742033*^9, 3.386623160572022*^9}, {
   3.386682669788158*^9, 3.386682763222602*^9}, 3.386695228272828*^9, {
   3.386695406740401*^9, 3.386695406829602*^9}, {3.386696297319094*^9, 
   3.386696344747774*^9}, {3.386696399360704*^9, 3.386696564216738*^9}, 
   3.386707841493029*^9, {3.386707944124362*^9, 3.386707944402568*^9}, {
   3.386715004189672*^9, 3.386715004455254*^9}, {3.386715039364132*^9, 
   3.386715039652016*^9}, {3.386944717481733*^9, 3.386944965080542*^9}, {
   3.386948535265229*^9, 3.386948663922406*^9}, {3.386948700421768*^9, 
   3.386948730293284*^9}, {3.386948775235851*^9, 3.38694877745197*^9}, 
   3.403835298208848*^9, 3.403835416935682*^9, {3.418658635629011*^9, 
   3.418658640576658*^9}, 3.418659648138493*^9}],

Cell[BoxData[
 RowBox[{"(*", " ", 
  RowBox[{
   RowBox[{
   "Now", " ", "we", " ", "actually", " ", "execute", " ", "the", " ", 
    "generated", " ", "code"}], ",", " ", 
   RowBox[{
   "and", " ", "then", " ", "display", " ", "the", " ", "sampled", " ", 
    "values", " ", "of", " ", "k1", " ", "and", " ", 
    RowBox[{"k2", "."}]}]}], " ", "*)"}]], "Input",
 CellChangeTimes->{{3.386944968656947*^9, 3.386944996041679*^9}, {
  3.386947160851194*^9, 3.386947163372285*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Print", "[", "\"\<Here we go!\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"code", "[", 
    RowBox[{"[", "1", "]"}], "]"}], "[", "]"}], ";"}], "\[IndentingNewLine]", 

 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"code", "[", 
     RowBox[{"[", "2", "]"}], "]"}], "[", "True", "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Total", "[", "k1", "]"}], "/", 
  RowBox[{"Length", "[", "k1", "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Total", "[", "k2", "]"}], "/", 
  RowBox[{"Length", "[", "k2", "]"}]}]}], "Input",
 CellChangeTimes->{{3.386696594976131*^9, 3.386696609250415*^9}, {
   3.386707928661264*^9, 3.38670792917469*^9}, {3.386766997677039*^9, 
   3.386767033611331*^9}, {3.386767309037521*^9, 3.386767382013767*^9}, {
   3.418658709138191*^9, 3.418658721910278*^9}, {3.418660074875778*^9, 
   3.418660080517048*^9}, {3.418668509906231*^9, 3.418668521204674*^9}, {
   3.418674433651921*^9, 3.418674435376356*^9}, {3.418674608763772*^9, 
   3.418674643090346*^9}, {3.418674950877165*^9, 3.418674951540242*^9}, 
   3.418676308937147*^9}],

Cell[BoxData["\<\"Here we go!\"\>"], "Print",
 CellChangeTimes->{3.418695228878855*^9, 3.424802013175918*^9, 
  3.42480263600983*^9, 3.424803192281104*^9, 3.424803288643474*^9}],

Cell[BoxData["0.05088369935707732`"], "Output",
 CellChangeTimes->{3.418695253529966*^9, 3.42480202567719*^9, 
  3.424802648470366*^9, 3.424803199229625*^9, 3.424803297400901*^9}],

Cell[BoxData["0.29034211830911916`"], "Output",
 CellChangeTimes->{3.418695253529966*^9, 3.42480202567719*^9, 
  3.424802648470366*^9, 3.424803199229625*^9, 3.424803297663449*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"ListPlot", "[", 
  RowBox[{"k1", ",", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"All", ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "1"}], "}"}]}], "}"}]}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListPlot", "[", 
  RowBox[{"k2", ",", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"All", ",", "All"}], "}"}]}]}], 
  "]"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.386696594976131*^9, 3.386696609250415*^9}, {
  3.386707928661264*^9, 3.38670792917469*^9}, {3.386766997677039*^9, 
  3.386767033611331*^9}, {3.386767309037521*^9, 3.386767382013767*^9}, {
  3.418658709138191*^9, 3.418658721910278*^9}, {3.418660074875778*^9, 
  3.418660080517048*^9}, {3.418668509906231*^9, 3.418668521204674*^9}, {
  3.418674433651921*^9, 3.418674435376356*^9}, {3.418674608763772*^9, 
  3.418674643090346*^9}, {3.418674950877165*^9, 3.418674951540242*^9}, {
  3.418676308937147*^9, 3.418676314792336*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, 
   {Hue[0.67, 0.6, 0.6], PointBox[CompressedData["
1:eJxd0rlqVkEYBuAPbSwtAlpY/BFxw2jiHo1m3Pct7msmahKjiV6AzVQKll7B
AUHBxkYwnXMVlvGUltqopFLxF/mfgcPwnJeXt5n+u0/GJpdExOLv78/993wd
HenMTq24vzDa/ZEGfw7Nf/nxoetleDnuSwPj6/q+PXvV9Uq8CnfwarwGr8Xr
8Ua8CW/Gg3gIb8Xb8Ha8A+/Eu/BuPIz34L14BO/D+/FoWtzy8fPzLW+6Tr0u
OA6Q4zhIjuMQOY7D5DiOkOM4So7jGDmO4+Q4TpDjOEmO4xQ5jtPkOM6Q4zhL
juMcOY7zvU644IrjAn1ccMUxRh8XXHFcpI8Lrjgu0ccFVxyX6eOCK44r9HHB
FcdV+rjgiuMafVxwxXGdPi644rhBHxdccdykjwuuOG7RxwVXHLfp44Irjjv0
ccEVxzh9XHDFkXvdwQlnXHCDK25xTLCPE8644GYivVg6v/C+/213H7c47va6
gxPOuOAGV9ziuMc+TjjjghtccYvjPvs44YwLbnDFLY5J9nHCGRfc4IpbHFPs
44QzLrjBFbc4ptnHCWdccIMrbnE8YB8nnHHBDa64xTHDPk44z6QlA8Ovpz+9
6+7jBlfc4njY6w5OOOOCG1xxi+MR+zjhjAtucMUtjln2ccIZF9zgilscc+zj
hDMuc2n25Ybvb57+e3+44hbH4153cMIZF9zgilscT9jH6b9/AWnIyxk=
     "]]}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  ImageSize->{642., Automatic},
  PlotRange->{All, {0, 1}},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.386714370224535*^9, 3.386766962747554*^9, 3.386767010004538*^9, 
   3.418660059451291*^9, 3.418660143769401*^9, 3.418660480659309*^9, 
   3.418664865941501*^9, 3.418667136030786*^9, 3.418668259900182*^9, 
   3.41866875200754*^9, 3.41866884184705*^9, 3.418669193341372*^9, 
   3.418673993328145*^9, 3.418674266859619*^9, 3.418674496931694*^9, 
   3.418674704715281*^9, 3.418674980063578*^9, 3.418675314232502*^9, 
   3.418675846187217*^9, {3.41867630048011*^9, 3.418676315286428*^9}, 
   3.418695286576042*^9, 3.424802155129928*^9, 3.424802678047709*^9, 
   3.424803498388786*^9}],

Cell[BoxData[
 GraphicsBox[{{}, 
   {Hue[0.67, 0.6, 0.6], PointBox[CompressedData["
1:eJxd13tUFNcdB/CpoFJF0cRYYjFZX4giKBGxAspXxbcIy1tQGBaW5bEsWIWo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     "]]}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  ImageSize->{642., Automatic},
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.386714370224535*^9, 3.386766962747554*^9, 3.386767010004538*^9, 
   3.418660059451291*^9, 3.418660143769401*^9, 3.418660480659309*^9, 
   3.418664865941501*^9, 3.418667136030786*^9, 3.418668259900182*^9, 
   3.41866875200754*^9, 3.41866884184705*^9, 3.418669193341372*^9, 
   3.418673993328145*^9, 3.418674266859619*^9, 3.418674496931694*^9, 
   3.418674704715281*^9, 3.418674980063578*^9, 3.418675314232502*^9, 
   3.418675846187217*^9, {3.41867630048011*^9, 3.418676315286428*^9}, 
   3.418695286576042*^9, 3.424802155129928*^9, 3.424802678047709*^9, 
   3.42480349880128*^9}]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ANNEALITS", "=", "20"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"SIGMA", "=", "1"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"k1z", "=", 
   RowBox[{"{", "}"}]}], ";", 
  RowBox[{"k2z", "=", 
   RowBox[{"{", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"code", "[", 
    RowBox[{"[", "2", "]"}], "]"}], "[", "True", "]"}], ";"}], "\n", 
 RowBox[{"For", "[", 
  RowBox[{
   RowBox[{"it", "=", "1"}], ",", 
   RowBox[{"it", "\[LessEqual]", "ANNEALITS"}], ",", 
   RowBox[{"it", "++"}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"For", "[", 
     RowBox[{
      RowBox[{"it2", "=", "1"}], ",", 
      RowBox[{"it2", "\[LessEqual]", "ITERZ"}], ",", 
      RowBox[{"it2", "++"}], ",", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"AppendTo", "[", 
        RowBox[{"k1z", ",", 
         RowBox[{"Part", "[", 
          RowBox[{"k1", ",", "it2"}], "]"}]}], "]"}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{"AppendTo", "[", 
        RowBox[{"k2z", ",", 
         RowBox[{"Part", "[", 
          RowBox[{"k2", ",", "it2"}], "]"}]}], "]"}], ";"}]}], 
     "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Part", "[", 
      RowBox[{"k1", ",", "1"}], "]"}], "=", 
     RowBox[{"Part", "[", 
      RowBox[{"k1", ",", "ITERZ"}], "]"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Part", "[", 
      RowBox[{"k2", ",", "1"}], "]"}], "=", 
     RowBox[{"Part", "[", 
      RowBox[{"k2", ",", "ITERZ"}], "]"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"SIGMA", "=", 
     RowBox[{"SIGMA", "*", "1.5"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"code", "[", 
      RowBox[{"[", "2", "]"}], "]"}], "[", "False", "]"}], ";"}]}], 
  "\[IndentingNewLine]", "]"}]}], "Input",
 CellChangeTimes->{{3.418674769385935*^9, 3.418674852169877*^9}, {
   3.4186749622525*^9, 3.418674964760836*^9}, 3.418675187474224*^9, {
   3.418675605404769*^9, 3.418675616367772*^9}, {3.41867606668593*^9, 
   3.418676066792398*^9}, {3.418676281349545*^9, 3.418676285417882*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"ListPlot", "[", 
    RowBox[{"k1z", ",", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", ".1"}], "}"}]}], "}"}]}]}], "]"}], 
   "\[IndentingNewLine]", 
   RowBox[{"ListPlot", "[", 
    RowBox[{"k2z", ",", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}]}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.418674769385935*^9, 3.418674852169877*^9}, {
  3.4186749622525*^9, 3.418674964760836*^9}, {3.418675187474224*^9, 
  3.418675193255241*^9}, {3.418675547970572*^9, 3.418675562513767*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, 
   {Hue[0.67, 0.6, 0.6], PointBox[CompressedData["
1:eJxd3HFQ1A+ZMPBvvzOPjIyMjIxsMzIyss3IyMg2IyMj24yMjGwzfkZGthkZ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     "]]}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  ImageSize->{861., Automatic},
  PlotRange->{All, {0, 0.1}},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.41867487539813*^9, 3.418675093797072*^9, 3.418675193993512*^9, {
   3.41867554987681*^9, 3.41867556306444*^9}, 3.418676034694597*^9, 
   3.418676531217396*^9, 3.41869581866201*^9, 3.424803076954061*^9, 
   3.424803720489786*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]],

Cell[BoxData[
 GraphicsBox[{{}, 
   {Hue[0.67, 0.6, 0.6], PointBox[CompressedData["
1:eJxc3X9cztf/+PFrFos1QiMWa9asEWvWrFl4skYIIRYaF0Jas0YsSS5JkiSV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     "]]}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  ImageSize->{861., Automatic},
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.41867487539813*^9, 3.418675093797072*^9, 3.418675193993512*^9, {
   3.41867554987681*^9, 3.41867556306444*^9}, 3.418676034694597*^9, 
   3.418676531217396*^9, 3.41869581866201*^9, 3.424803076954061*^9, 
   3.424803721053931*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Total", "[", "k1z", "]"}], "/", 
  RowBox[{"Length", "[", "k1z", "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Total", "[", "k2z", "]"}], "/", 
  RowBox[{"Length", "[", "k2z", "]"}]}]}], "Input",
 CellChangeTimes->{{3.41867517126257*^9, 3.418675174301184*^9}}],

Cell[BoxData["0.060847278604253215`"], "Output",
 CellChangeTimes->{3.418675175072176*^9, 3.418675529834789*^9, 
  3.418676042755795*^9, 3.418676536690821*^9, 3.418695836240827*^9, 
  3.424803086389035*^9, 3.424803725270926*^9}],

Cell[BoxData["0.8215521389464896`"], "Output",
 CellChangeTimes->{3.418675175072176*^9, 3.418675529834789*^9, 
  3.418676042755795*^9, 3.418676536690821*^9, 3.418695836240827*^9, 
  3.424803086389035*^9, 3.424803725457576*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"k1z", ",", "k2z"}], "}"}]], "Input",
 CellChangeTimes->{{3.424803740665163*^9, 3.424803749245742*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0.4955141976102071`", ",", "0.4021844317451879`", ",", 
     "0.12074185753544237`", ",", "0.12074185753544237`", ",", 
     "0.12074185753544237`", ",", "0.08624907361075376`", ",", 
     "0.08624907361075376`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.045241845458850204`", ",", "0.045241845458850204`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05494798736280055`", ",", 
     "0.05494798736280055`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05084524984494941`", ",", "0.05084524984494941`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05189320289785132`", ",", 
     "0.05189320289785132`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05704183572691636`", ",", 
     "0.05704183572691636`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.05798068903546255`", ",", 
     "0.05798068903546255`", ",", "0.0606586653616`", ",", "0.0606586653616`",
      ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.0606586653616`", 
     ",", "0.0606586653616`", ",", "0.0606586653616`", ",", 
     "0.0606586653616`", ",", "0.0606586653616`", ",", "0.05720487891843458`",
      ",", "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.05720487891843458`", ",", "0.05720487891843458`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.0524989695593736`", ",", 
     "0.0524989695593736`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.054505878786409535`", ",", 
     "0.054505878786409535`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06769651862626894`", ",", "0.06769651862626894`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.06496643901581949`", ",", 
     "0.06496643901581949`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05764009454539397`", ",", 
     "0.05764009454539397`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.05897384501501524`", ",", 
     "0.05897384501501524`", ",", "0.0598023931990677`", ",", 
     "0.0598023931990677`", ",", "0.0598023931990677`", ",", 
     "0.0598023931990677`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.0616772742889978`", ",", "0.0616772742889978`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06251204927530954`", ",", "0.06251204927530954`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06211485874371283`", ",", "0.06211485874371283`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06119077336566199`", ",", 
     "0.06119077336566199`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06117759022555151`", ",", 
     "0.06117759022555151`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.06121387702954772`", ",", "0.06121387702954772`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`", ",", 
     "0.0615762670885511`", ",", "0.0615762670885511`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.6688441629810935`", ",", "0.6688441629810935`", ",", 
     "0.6688441629810935`", ",", "1.1478074237456535`", ",", 
     "1.1478074237456535`", ",", "1.1478074237456535`", ",", 
     "1.1478074237456535`", ",", "1.1478074237456535`", ",", 
     "0.887580257877616`", ",", "0.7509290943043182`", ",", 
     "0.7509290943043182`", ",", "0.48471109086732755`", ",", 
     "0.49339878470152226`", ",", "0.4957151688518906`", ",", 
     "0.043912803471164996`", ",", "0.043912803471164996`", ",", 
     "0.1440724614226363`", ",", 
     RowBox[{"-", "0.052658576931598394`"}], ",", "0.2986946597110546`", ",", 
     "0.07580676074108306`", ",", "0.07580676074108306`", ",", 
     "0.22788085317983775`", ",", "0.22788085317983775`", ",", 
     "0.22788085317983775`", ",", "0.22210463301565722`", ",", 
     "0.18860055746161045`", ",", "0.12353639307525495`", ",", 
     "0.12353639307525495`", ",", "0.18779235661171528`", ",", 
     "0.18779235661171528`", ",", 
     RowBox[{"-", "0.04573464117765172`"}], ",", "0.12422357738066511`", ",", 
     "0.05904655573167995`", ",", "0.05904655573167995`", ",", 
     "0.2082802254567898`", ",", "0.29467012805989923`", ",", 
     "0.11716939454958455`", ",", "0.10414738754729022`", ",", 
     "0.10414738754729022`", ",", "0.10414738754729022`", ",", 
     "0.10414738754729022`", ",", "0.2857566119349668`", ",", 
     "0.2857566119349668`", ",", "0.10145958499736787`", ",", 
     "0.16270764735897777`", ",", "0.16270764735897777`", ",", 
     "0.08287977120620311`", ",", "0.08287977120620311`", ",", 
     "0.08287977120620311`", ",", "0.08287977120620311`", ",", 
     "0.12532524448088397`", ",", "0.07659617435061197`", ",", 
     "0.07659617435061197`", ",", "0.07659617435061197`", ",", 
     "0.07659617435061197`", ",", "0.07659617435061197`", ",", 
     "0.07659617435061197`", ",", "0.2729645873325966`", ",", 
     "0.13722988708142186`", ",", "0.018094637854309106`", ",", 
     "0.28947865133470674`", ",", "0.20465457049626423`", ",", 
     "0.10472101835377175`", ",", "0.10472101835377175`", ",", 
     "0.10472101835377175`", ",", "0.23026099535868305`", ",", 
     "0.2512307933344825`", ",", 
     RowBox[{"-", "0.026612907773637784`"}], ",", 
     RowBox[{"-", "0.026612907773637784`"}], ",", "0.15288968170556966`", ",",
      "0.15288968170556966`", ",", "0.15288968170556966`", ",", 
     "0.3330229928344088`", ",", "0.3330229928344088`", ",", 
     "0.04299819817785289`", ",", "0.4920126298390231`", ",", 
     "0.4920126298390231`", ",", "0.4920126298390231`", ",", 
     "0.0775663239518879`", ",", "0.020938326215327777`", ",", 
     "0.020938326215327777`", ",", "0.16208967866373108`", ",", 
     "0.19122035047271915`", ",", "0.2587289956358736`", ",", 
     "0.20040373881338744`", ",", "0.20040373881338744`", ",", 
     RowBox[{"-", "0.0603163523952788`"}], ",", 
     RowBox[{"-", "0.0603163523952788`"}], ",", 
     RowBox[{"-", "0.031622046667005736`"}], ",", 
     RowBox[{"-", "0.031622046667005736`"}], ",", 
     RowBox[{"-", "0.031622046667005736`"}], ",", 
     RowBox[{"-", "0.031622046667005736`"}], ",", "0.03073740529722726`", ",",
      "0.05648327251511086`", ",", "0.05648327251511086`", ",", 
     "0.026519622639955973`", ",", "0.22196054821258288`", ",", 
     "0.22196054821258288`", ",", "0.22515421040608605`", ",", 
     "0.2431714808386385`", ",", "0.06516946086174677`", ",", 
     "0.13319970393484576`", ",", 
     RowBox[{"-", "0.050363251173194984`"}], ",", 
     RowBox[{"-", "0.050363251173194984`"}], ",", 
     RowBox[{"-", "0.050363251173194984`"}], ",", 
     RowBox[{"-", "0.050363251173194984`"}], ",", 
     RowBox[{"-", "0.050363251173194984`"}], ",", 
     RowBox[{"-", "0.050363251173194984`"}], ",", 
     RowBox[{"-", "0.050363251173194984`"}], ",", 
     RowBox[{"-", "0.08080871075814677`"}], ",", "0.2785909904626315`", ",", 
     "0.2785909904626315`", ",", "0.2785909904626315`", ",", 
     "0.2785909904626315`", ",", "0.2785909904626315`", ",", 
     "0.35461230233164864`", ",", "0.3456059894368092`", ",", 
     "0.21139722150366513`", ",", "0.21139722150366513`", ",", 
     "0.03765425645627443`", ",", "0.3938456100646095`", ",", 
     "0.1696622661271895`", ",", "0.2873344305907286`", ",", 
     "0.6120871741765616`", ",", "0.46266982982592153`", ",", 
     "0.46266982982592153`", ",", "0.6041076299415635`", ",", 
     "0.5603402068434437`", ",", "0.5603402068434437`", ",", 
     "0.7721983147882069`", ",", "0.5999707010299209`", ",", 
     "0.5999707010299209`", ",", "0.5803043377216816`", ",", 
     "0.18441810921144947`", ",", "0.18441810921144947`", ",", 
     "0.38498978585337496`", ",", "0.5696402511319868`", ",", 
     "0.5696402511319868`", ",", "0.5696402511319868`", ",", 
     "0.5679800368353959`", ",", "0.5679800368353959`", ",", 
     "0.6315519157333416`", ",", "0.6315519157333416`", ",", 
     "0.6315519157333416`", ",", "0.6315519157333416`", ",", 
     "0.3346938757577185`", ",", "0.3346938757577185`", ",", 
     "0.26719368678170624`", ",", "0.12698725281837692`", ",", 
     "0.12698725281837692`", ",", "0.12698725281837692`", ",", 
     "0.4094630909312485`", ",", "0.3802593184492602`", ",", 
     "0.3802593184492602`", ",", "0.47526463352069787`", ",", 
     "0.35322066688196135`", ",", "0.35322066688196135`", ",", 
     "0.41541520574875435`", ",", "0.26919230563985586`", ",", 
     "0.26919230563985586`", ",", "0.26919230563985586`", ",", 
     "0.08460516594468076`", ",", "0.08460516594468076`", ",", 
     "0.1100982999627767`", ",", "0.1100982999627767`", ",", 
     "0.1100982999627767`", ",", "0.1100982999627767`", ",", 
     RowBox[{"-", "0.05143767636283428`"}], ",", 
     RowBox[{"-", "0.05143767636283428`"}], ",", 
     RowBox[{"-", "0.05143767636283428`"}], ",", 
     RowBox[{"-", "0.05143767636283428`"}], ",", "0.45763204083107223`", ",", 
     "0.5943712233492777`", ",", "0.5572286649217781`", ",", 
     "0.6482106674497399`", ",", "0.6482106674497399`", ",", 
     "0.3465047066847188`", ",", "0.345990067482465`", ",", 
     "0.345990067482465`", ",", "0.345990067482465`", ",", 
     "0.345990067482465`", ",", "0.2696322432123583`", ",", 
     "0.2696322432123583`", ",", "0.2696322432123583`", ",", 
     "0.30943294005928945`", ",", "0.30943294005928945`", ",", 
     "0.3557971221315152`", ",", "0.3557971221315152`", ",", 
     "0.509493316579055`", ",", "0.25939700295464685`", ",", 
     "0.25939700295464685`", ",", "0.1924144766181271`", ",", 
     "0.1924144766181271`", ",", "0.5262124717203531`", ",", 
     "0.7083958657594818`", ",", "0.25810047436420624`", ",", 
     "0.18898769002319915`", ",", "0.32968056611357566`", ",", 
     "0.32968056611357566`", ",", "0.32968056611357566`", ",", 
     "0.32968056611357566`", ",", "0.2249306697760749`", ",", 
     "0.5599246230295689`", ",", "0.5599246230295689`", ",", 
     "0.5599246230295689`", ",", "0.5599246230295689`", ",", 
     "0.6210702885347879`", ",", "0.6210702885347879`", ",", 
     "0.27425530561256994`", ",", "0.27425530561256994`", ",", 
     "0.27425530561256994`", ",", "0.27425530561256994`", ",", 
     "0.3387226791954276`", ",", "0.3387226791954276`", ",", 
     "0.4029069260285906`", ",", "0.3833613481938554`", ",", 
     "0.3833613481938554`", ",", "0.43995898275119627`", ",", 
     "0.6615424273681665`", ",", "0.6615424273681665`", ",", 
     "0.6615424273681665`", ",", "0.57533781130221`", ",", 
     "0.57533781130221`", ",", "0.57533781130221`", ",", "0.57533781130221`", 
     ",", "0.57533781130221`", ",", "0.57533781130221`", ",", 
     "0.17969878103144404`", ",", "0.17969878103144404`", ",", 
     "0.3619703117700429`", ",", "0.4516355605513034`", ",", 
     "0.39928292452830666`", ",", "0.39928292452830666`", ",", 
     "0.39928292452830666`", ",", "0.39928292452830666`", ",", 
     "0.3620803888130278`", ",", "0.3136754606873479`", ",", 
     "0.3136754606873479`", ",", "0.3136754606873479`", ",", 
     "0.499352936585969`", ",", "0.43996012355489694`", ",", 
     "0.443682013262412`", ",", "0.443682013262412`", ",", 
     "0.443682013262412`", ",", "0.46894808193877224`", ",", 
     "0.7061661917304312`", ",", "0.7061661917304312`", ",", 
     "0.7061661917304312`", ",", "0.31035135851873336`", ",", 
     "0.31035135851873336`", ",", "0.31035135851873336`", ",", 
     "0.31035135851873336`", ",", "0.5954609729935597`", ",", 
     "0.6116188607333266`", ",", "0.2798382568656204`", ",", 
     "0.2798382568656204`", ",", "0.37591932287223734`", ",", 
     "0.25515168892186796`", ",", "0.25515168892186796`", ",", 
     "0.25515168892186796`", ",", "0.5333262577902653`", ",", 
     "0.5333262577902653`", ",", "0.5333262577902653`", ",", 
     "0.5012560847659238`", ",", "0.5012560847659238`", ",", 
     "0.5012560847659238`", ",", "0.5012560847659238`", ",", 
     "0.5012560847659238`", ",", "0.5012560847659238`", ",", 
     "0.4378589012511358`", ",", "0.4378589012511358`", ",", 
     "0.4378589012511358`", ",", "0.4378589012511358`", ",", 
     "0.4378589012511358`", ",", "0.4378589012511358`", ",", 
     "0.4378589012511358`", ",", "0.4378589012511358`", ",", 
     "0.4378589012511358`", ",", "0.4378589012511358`", ",", 
     "0.4378589012511358`", ",", "0.45319247781621513`", ",", 
     "0.45319247781621513`", ",", "0.4165677144486792`", ",", 
     "0.33527919762812974`", ",", "0.3468281481858955`", ",", 
     "0.3468281481858955`", ",", "0.3468281481858955`", ",", 
     "0.3468281481858955`", ",", "0.3468281481858955`", ",", 
     "0.509863745373826`", ",", "0.6745602889881008`", ",", 
     "0.6745602889881008`", ",", "0.3660191501660697`", ",", 
     "0.6148582696146624`", ",", "0.6148582696146624`", ",", 
     "0.6148582696146624`", ",", "0.6148582696146624`", ",", 
     "0.17568427612026255`", ",", "0.17568427612026255`", ",", 
     "0.17568427612026255`", ",", "0.2932220780751529`", ",", 
     "0.2932220780751529`", ",", "0.4582248600586458`", ",", 
     "0.4582248600586458`", ",", "0.3856641652017696`", ",", 
     "0.18888010085712337`", ",", "0.18888010085712337`", ",", 
     "0.18888010085712337`", ",", "0.18888010085712337`", ",", 
     "0.16825914984796547`", ",", "0.16825914984796547`", ",", 
     "0.19011029505906663`", ",", "0.5666914107426904`", ",", 
     "0.5287743475589932`", ",", "0.5287743475589932`", ",", 
     "0.5287743475589932`", ",", "0.41064184813114263`", ",", 
     "0.41064184813114263`", ",", "0.2832465112157593`", ",", 
     "0.2832465112157593`", ",", "0.2832465112157593`", ",", 
     "0.2832465112157593`", ",", "0.2171499332850249`", ",", 
     "0.5255318479875108`", ",", "0.5255318479875108`", ",", 
     "0.5255318479875108`", ",", "0.5255318479875108`", ",", 
     "0.808791007471765`", ",", "0.808791007471765`", ",", 
     "0.808791007471765`", ",", "0.808791007471765`", ",", 
     "0.7920834815302105`", ",", "0.6690784534497495`", ",", 
     "0.6690784534497495`", ",", "0.6690784534497495`", ",", 
     "0.5648762963571711`", ",", "0.5648762963571711`", ",", 
     "0.9013624547654484`", ",", "0.7638331629272592`", ",", 
     "0.7638331629272592`", ",", "0.7638331629272592`", ",", 
     "0.7638331629272592`", ",", "0.7638331629272592`", ",", 
     "0.7638331629272592`", ",", "0.7638331629272592`", ",", 
     "0.6691417155843876`", ",", "0.6691417155843876`", ",", 
     "0.7264527035023736`", ",", "0.7264527035023736`", ",", 
     "0.7264527035023736`", ",", "0.6405516287058626`", ",", 
     "0.40490616389929707`", ",", "0.48594793831618605`", ",", 
     "0.48594793831618605`", ",", "0.49807776847387847`", ",", 
     "0.59613792971059`", ",", "0.59613792971059`", ",", "0.59613792971059`", 
     ",", "0.59613792971059`", ",", "0.59613792971059`", ",", 
     "0.59613792971059`", ",", "0.5680007040166242`", ",", 
     "0.5103342339176372`", ",", "0.5103342339176372`", ",", 
     "0.5103342339176372`", ",", "0.400325318987156`", ",", 
     "0.7890472104134663`", ",", "0.729573636025447`", ",", 
     "0.729573636025447`", ",", "0.6483932659814606`", ",", 
     "0.6483932659814606`", ",", "0.6483932659814606`", ",", 
     "0.6483932659814606`", ",", "0.6483932659814606`", ",", 
     "0.6483932659814606`", ",", "0.7356356500088762`", ",", 
     "0.6816723148874457`", ",", "0.5300997751130828`", ",", 
     "0.5300997751130828`", ",", "0.5944943843781826`", ",", 
     "0.5944943843781826`", ",", "0.5944943843781826`", ",", 
     "0.786447589560507`", ",", "0.786447589560507`", ",", 
     "0.8049008085692806`", ",", "0.7069093304531312`", ",", 
     "0.5664635464155885`", ",", "0.3151051032377651`", ",", 
     "0.8192301147444855`", ",", "0.8192301147444855`", ",", 
     "0.768493435583046`", ",", "0.578084251948797`", ",", 
     "0.578084251948797`", ",", "0.578084251948797`", ",", 
     "0.8723872479951917`", ",", "0.8723872479951917`", ",", 
     "0.666383851361085`", ",", "0.666383851361085`", ",", 
     "0.666383851361085`", ",", "0.666383851361085`", ",", 
     "0.666383851361085`", ",", "0.666383851361085`", ",", 
     "0.8500705145563348`", ",", "0.8500705145563348`", ",", 
     "0.6810775698018797`", ",", "0.37774200876736963`", ",", 
     "0.37774200876736963`", ",", "0.37774200876736963`", ",", 
     "0.37774200876736963`", ",", "0.37774200876736963`", ",", 
     "0.37774200876736963`", ",", "0.5565080546454843`", ",", 
     "0.5565080546454843`", ",", "0.5860217655204794`", ",", 
     "0.5860217655204794`", ",", "0.5860217655204794`", ",", 
     "0.5860217655204794`", ",", "0.5860217655204794`", ",", 
     "0.5860217655204794`", ",", "0.7282521359588485`", ",", 
     "0.7282521359588485`", ",", "0.7643898777632906`", ",", 
     "0.8533962414699761`", ",", "0.6772914678639524`", ",", 
     "0.5325786091619991`", ",", "0.38499837199469017`", ",", 
     "0.6452765999716719`", ",", "0.4936132634491783`", ",", 
     "0.4936132634491783`", ",", "0.4936132634491783`", ",", 
     "0.4936132634491783`", ",", "0.5233110516897538`", ",", 
     "0.5233110516897538`", ",", "0.6583499031174129`", ",", 
     "0.687577700703295`", ",", "0.687577700703295`", ",", 
     "0.4334824931132731`", ",", "0.4334824931132731`", ",", 
     "0.7982735362229518`", ",", "0.7580212907184287`", ",", 
     "0.7580212907184287`", ",", "0.7580212907184287`", ",", 
     "0.7580212907184287`", ",", "0.7580212907184287`", ",", 
     "0.7580212907184287`", ",", "0.7580212907184287`", ",", 
     "0.8891986030368817`", ",", "0.8891986030368817`", ",", 
     "0.8891986030368817`", ",", "0.7885272179887967`", ",", 
     "0.7885272179887967`", ",", "0.7885272179887967`", ",", 
     "0.6529298135012254`", ",", "0.7744819586797609`", ",", 
     "0.7744819586797609`", ",", "1.0187634894752677`", ",", 
     "1.0187634894752677`", ",", "1.0187634894752677`", ",", 
     "1.0187634894752677`", ",", "1.0501871360214856`", ",", 
     "0.7048335931988752`", ",", "0.7048335931988752`", ",", 
     "0.5160022097523641`", ",", "0.5669696513051404`", ",", 
     "0.5669696513051404`", ",", "0.5016787296703816`", ",", 
     "0.5016787296703816`", ",", "0.6556749848814568`", ",", 
     "0.6556749848814568`", ",", "0.9621361234644905`", ",", 
     "0.9621361234644905`", ",", "0.5586848274748644`", ",", 
     "1.071287789921306`", ",", "1.071287789921306`", ",", 
     "0.6254027788541247`", ",", "0.8358441550618742`", ",", 
     "0.8358441550618742`", ",", "0.7535125220924246`", ",", 
     "0.6363145099105263`", ",", "0.6363145099105263`", ",", 
     "0.6363145099105263`", ",", "0.6363145099105263`", ",", 
     "0.6363145099105263`", ",", "0.6363145099105263`", ",", 
     "0.6363145099105263`", ",", "0.8235691588182683`", ",", 
     "0.8235691588182683`", ",", "0.8235691588182683`", ",", 
     "0.7866555899682417`", ",", "0.9843045155694411`", ",", 
     "0.9843045155694411`", ",", "0.7936610491950187`", ",", 
     "0.7936610491950187`", ",", "0.8616994344721614`", ",", 
     "0.8447493782516061`", ",", "0.8447493782516061`", ",", 
     "0.8447493782516061`", ",", "0.9157831622307382`", ",", 
     "0.9157831622307382`", ",", "0.9157831622307382`", ",", 
     "0.9157831622307382`", ",", "0.6531007278451743`", ",", 
     "0.6531007278451743`", ",", "0.5759441174114467`", ",", 
     "0.5509673983269072`", ",", "0.7268464750878151`", ",", 
     "0.7268464750878151`", ",", "0.7268464750878151`", ",", 
     "0.7268464750878151`", ",", "0.7268464750878151`", ",", 
     "0.7268464750878151`", ",", "0.8613776820021005`", ",", 
     "0.8613776820021005`", ",", "0.9466140752428222`", ",", 
     "0.9466140752428222`", ",", "0.9466140752428222`", ",", 
     "0.9466140752428222`", ",", "0.9752072933990413`", ",", 
     "0.648878777638898`", ",", "0.648878777638898`", ",", 
     "0.8194790979540345`", ",", "0.8194790979540345`", ",", 
     "0.8194790979540345`", ",", "0.8194790979540345`", ",", 
     "0.8194790979540345`", ",", "0.8194790979540345`", ",", 
     "0.8194790979540345`", ",", "0.8194790979540345`", ",", 
     "0.8194790979540345`", ",", "0.8194790979540345`", ",", 
     "0.8662643692552214`", ",", "0.8662643692552214`", ",", 
     "0.7575448737250451`", ",", "0.7575448737250451`", ",", 
     "0.6093260590777604`", ",", "0.6093260590777604`", ",", 
     "0.7060575257707161`", ",", "0.6458055982867935`", ",", 
     "0.6458055982867935`", ",", "0.47385291212674335`", ",", 
     "0.47385291212674335`", ",", "0.5622615392400063`", ",", 
     "0.5622615392400063`", ",", "0.5622615392400063`", ",", 
     "0.7776097751515578`", ",", "0.8225534150761564`", ",", 
     "0.5826061239657786`", ",", "0.5397885084585726`", ",", 
     "0.47448592447809`", ",", "0.5829731234140931`", ",", 
     "0.7304491269681683`", ",", "0.7304491269681683`", ",", 
     "0.6471278860308163`", ",", "0.5118588796237906`", ",", 
     "0.5118588796237906`", ",", "0.8075594773003745`", ",", 
     "0.7055226092627334`", ",", "0.7055226092627334`", ",", 
     "0.7055226092627334`", ",", "0.6490517850041311`", ",", 
     "0.7505216447065846`", ",", "0.7505216447065846`", ",", 
     "0.8246887719567039`", ",", "0.707874533300132`", ",", 
     "0.9175757246052089`", ",", "0.509956082601136`", ",", 
     "0.509956082601136`", ",", "0.509956082601136`", ",", 
     "0.6880281123510312`", ",", "0.8170131563116312`", ",", 
     "0.4951730433495431`", ",", "0.4976566460116394`", ",", 
     "0.4067610828448079`", ",", "0.5821548050560157`", ",", 
     "0.5821548050560157`", ",", "0.654720200908854`", ",", 
     "0.5981482426909687`", ",", "0.6190754726525103`", ",", 
     "0.6190754726525103`", ",", "0.7778168784123968`", ",", 
     "0.7778168784123968`", ",", "0.7778168784123968`", ",", 
     "0.49766238023711795`", ",", "0.49766238023711795`", ",", 
     "0.49766238023711795`", ",", "0.49766238023711795`", ",", 
     "0.49766238023711795`", ",", "0.49766238023711795`", ",", 
     "0.37231127258909635`", ",", "0.5604476036765195`", ",", 
     "0.5604476036765195`", ",", "0.5604476036765195`", ",", 
     "0.5705682720266503`", ",", "0.5705682720266503`", ",", 
     "0.5705682720266503`", ",", "0.5705682720266503`", ",", 
     "0.634571634170102`", ",", "0.634571634170102`", ",", 
     "0.5469369804049586`", ",", "0.6238906597316584`", ",", 
     "0.6238906597316584`", ",", "0.7396229589307489`", ",", 
     "0.7396229589307489`", ",", "0.8669191748862426`", ",", 
     "0.5570146146608841`", ",", "0.6372177510665443`", ",", 
     "0.6372177510665443`", ",", "0.581986296655514`", ",", 
     "0.581986296655514`", ",", "0.581986296655514`", ",", 
     "0.3042361619939554`", ",", "0.3042361619939554`", ",", 
     "0.31705101937796504`", ",", "0.31705101937796504`", ",", 
     "0.31705101937796504`", ",", "0.38867178099514427`", ",", 
     "0.3669049197057113`", ",", "0.3669049197057113`", ",", 
     "0.3669049197057113`", ",", "0.534073612251199`", ",", 
     "0.3664805009209413`", ",", "0.3664805009209413`", ",", 
     "0.596832059961182`", ",", "0.3127051073523657`", ",", 
     "0.3127051073523657`", ",", "0.3127051073523657`", ",", 
     "0.49729755716686563`", ",", "0.49729755716686563`", ",", 
     "0.3219505057799458`", ",", "0.46113034552196874`", ",", 
     "0.46113034552196874`", ",", "0.46113034552196874`", ",", 
     "0.46113034552196874`", ",", "0.46113034552196874`", ",", 
     "0.5022977485335796`", ",", "0.3703820844180348`", ",", 
     "0.3703820844180348`", ",", "0.3703820844180348`", ",", 
     "0.3703820844180348`", ",", "0.5993959296460667`", ",", 
     "0.49643496039972`", ",", "0.39981578716500155`", ",", 
     "0.39981578716500155`", ",", "0.39981578716500155`", ",", 
     "0.39981578716500155`", ",", "0.39981578716500155`", ",", 
     "0.39981578716500155`", ",", "0.2480186299383768`", ",", 
     "0.2480186299383768`", ",", "0.2919358301660415`", ",", 
     "0.38015570363872225`", ",", "0.40357564892615316`", ",", 
     "0.4228463503677451`", ",", "0.5984939650137716`", ",", 
     "0.5984939650137716`", ",", "0.5399922713085755`", ",", 
     "0.48877377756654133`", ",", "0.6833994565972004`", ",", 
     "0.6833994565972004`", ",", "0.475130151515773`", ",", 
     "0.475130151515773`", ",", "0.475130151515773`", ",", 
     "0.475130151515773`", ",", "0.475130151515773`", ",", 
     "0.475130151515773`", ",", "0.6214514767930579`", ",", 
     "0.6214514767930579`", ",", "0.6190984732743644`", ",", 
     "0.468577459754417`", ",", "0.5403969891476085`", ",", 
     "0.5403969891476085`", ",", "0.6835766745706754`", ",", 
     "0.6835766745706754`", ",", "0.6835766745706754`", ",", 
     "0.6835766745706754`", ",", "0.6429475686665607`", ",", 
     "0.6429475686665607`", ",", "0.6429475686665607`", ",", 
     "0.3786316659430184`", ",", "0.6017406275406596`", ",", 
     "0.6017406275406596`", ",", "0.6017406275406596`", ",", 
     "0.6017406275406596`", ",", "0.6017406275406596`", ",", 
     "0.6017406275406596`", ",", "0.6017406275406596`", ",", 
     "0.6175597066875409`", ",", "0.6951353810211578`", ",", 
     "0.6951353810211578`", ",", "0.4228136217625311`", ",", 
     "0.5366458928033033`", ",", "0.5366458928033033`", ",", 
     "0.930443876839668`", ",", "0.8950055549302519`", ",", 
     "0.7808481381677477`", ",", "0.5177710817240941`", ",", 
     "0.5177710817240941`", ",", "0.5177710817240941`", ",", 
     "0.5177710817240941`", ",", "0.5177710817240941`", ",", 
     "0.40011136637820166`", ",", "0.40011136637820166`", ",", 
     "0.40011136637820166`", ",", "0.7852426086197946`", ",", 
     "0.7852426086197946`", ",", "0.7857712478768111`", ",", 
     "0.7857712478768111`", ",", "0.7857712478768111`", ",", 
     "0.7857712478768111`", ",", "1.0882794532914857`", ",", 
     "1.0882794532914857`", ",", "1.0882794532914857`", ",", 
     "1.0030318372703106`", ",", "1.0030318372703106`", ",", 
     "1.3189049160685964`", ",", "1.1828025191047373`", ",", 
     "1.1828025191047373`", ",", "1.0787939977993735`", ",", 
     "1.0787939977993735`", ",", "1.0787939977993735`", ",", 
     "1.0953020414974604`", ",", "1.0953020414974604`", ",", 
     "1.0953020414974604`", ",", "1.1297446878894275`", ",", 
     "1.1297446878894275`", ",", "1.1784482353594867`", ",", 
     "1.1784482353594867`", ",", "1.0204125463158094`", ",", 
     "1.3810750411170605`", ",", "1.3810750411170605`", ",", 
     "1.1007222340627099`", ",", "1.0816878898007196`", ",", 
     "1.197860845071331`", ",", "1.2543081085225554`", ",", 
     "1.2543081085225554`", ",", "1.1671982681324653`", ",", 
     "1.1671982681324653`", ",", "1.1671982681324653`", ",", 
     "1.0173048425992033`", ",", "1.2456226697471195`", ",", 
     "1.140312026337476`", ",", "1.140312026337476`", ",", 
     "1.140312026337476`", ",", "1.140312026337476`", ",", 
     "0.9480220666909628`", ",", "0.9480220666909628`", ",", 
     "1.0079812757756204`", ",", "1.0079812757756204`", ",", 
     "1.0079812757756204`", ",", "1.0079812757756204`", ",", 
     "1.0079812757756204`", ",", "1.0079812757756204`", ",", 
     "0.9969992087902888`", ",", "0.9969992087902888`", ",", 
     "0.9969992087902888`", ",", "1.0283573742461916`", ",", 
     "1.0283573742461916`", ",", "1.1664028332364158`", ",", 
     "1.1664028332364158`", ",", "1.1664028332364158`", ",", 
     "1.1664028332364158`", ",", "1.1664028332364158`", ",", 
     "1.1664028332364158`", ",", "1.1702628703128568`", ",", 
     "1.1702628703128568`", ",", "1.1702628703128568`", ",", 
     "1.1702628703128568`", ",", "1.1702628703128568`", ",", 
     "1.0576909270606762`", ",", "1.0576909270606762`", ",", 
     "1.1186331934083498`", ",", "1.1186331934083498`", ",", 
     "1.1186331934083498`", ",", "1.1186331934083498`", ",", 
     "1.1186331934083498`", ",", "1.0954080142851352`", ",", 
     "1.1905802612322203`", ",", "1.1905802612322203`", ",", 
     "1.1905802612322203`", ",", "1.1905802612322203`", ",", 
     "1.1905802612322203`", ",", "1.1905802612322203`", ",", 
     "1.1905802612322203`", ",", "1.2247927701426136`", ",", 
     "1.2247927701426136`", ",", "1.2247927701426136`", ",", 
     "1.2247927701426136`", ",", "1.2612658983105254`", ",", 
     "1.1363721075471058`", ",", "1.1926701201230945`", ",", 
     "1.1926701201230945`", ",", "1.1926701201230945`", ",", 
     "1.1615805975642437`", ",", "1.2171028751667727`", ",", 
     "1.1330341356832558`", ",", "1.1330341356832558`", ",", 
     "1.1330341356832558`", ",", "1.1330341356832558`", ",", 
     "1.1330341356832558`", ",", "1.1330341356832558`", ",", 
     "1.1330341356832558`", ",", "1.1330341356832558`", ",", 
     "1.0798578854697758`", ",", "1.2098288081217454`", ",", 
     "1.2098288081217454`", ",", "1.2098288081217454`", ",", 
     "1.0968784488277394`", ",", "1.0968784488277394`", ",", 
     "1.0968784488277394`", ",", "1.0968784488277394`", ",", 
     "1.0968784488277394`", ",", "1.2489871417497873`", ",", 
     "0.9989547311091586`", ",", "1.1502240597688076`", ",", 
     "1.1502240597688076`", ",", "1.1502240597688076`", ",", 
     "1.1502240597688076`", ",", "1.1502240597688076`", ",", 
     "1.129657064720254`", ",", "1.129657064720254`", ",", 
     "1.2368244115197866`", ",", "1.2368244115197866`", ",", 
     "1.2368244115197866`", ",", "1.2368244115197866`", ",", 
     "1.2368244115197866`", ",", "1.1551457879217617`", ",", 
     "1.1551457879217617`", ",", "1.2281935859170503`", ",", 
     "1.2281935859170503`", ",", "1.0981446402647912`", ",", 
     "1.0981446402647912`", ",", "1.0981446402647912`", ",", 
     "1.0587411415348333`", ",", "1.0587411415348333`", ",", 
     "1.0587411415348333`", ",", "1.0587411415348333`", ",", 
     "1.023595878930336`", ",", "1.150244951959075`", ",", 
     "1.150244951959075`", ",", "1.1286344898357177`", ",", 
     "1.1286344898357177`", ",", "1.1286344898357177`", ",", 
     "1.1258970562050499`", ",", "0.9384663625289515`", ",", 
     "0.9384663625289515`", ",", "0.9384663625289515`", ",", 
     "0.9384663625289515`", ",", "0.9384663625289515`", ",", 
     "0.9384663625289515`", ",", "0.9384663625289515`", ",", 
     "0.9404929709925756`", ",", "0.9404929709925756`", ",", 
     "1.223508648460918`", ",", "0.8171836420347451`", ",", 
     "0.916222001075494`", ",", "1.0762319087744618`", ",", 
     "1.0969911494831472`", ",", "1.0969911494831472`", ",", 
     "1.0969911494831472`", ",", "1.0969911494831472`", ",", 
     "1.0969911494831472`", ",", "0.860337082443863`", ",", 
     "0.860337082443863`", ",", "1.1927430514710295`", ",", 
     "0.9149530695766155`", ",", "0.9149530695766155`", ",", 
     "0.9149530695766155`", ",", "0.9149530695766155`", ",", 
     "0.9149530695766155`", ",", "0.9149530695766155`", ",", 
     "0.9149530695766155`", ",", "0.9030069893663546`", ",", 
     "0.951712399135165`", ",", "0.951712399135165`", ",", 
     "1.102159334376898`", ",", "1.102159334376898`", ",", 
     "1.102159334376898`", ",", "1.102159334376898`", ",", 
     "0.9555497904345296`", ",", "0.9555497904345296`", ",", 
     "0.9383662132483328`", ",", "0.9053596530775335`", ",", 
     "0.9053596530775335`", ",", "0.8554402892890176`", ",", 
     "0.9725712847057302`", ",", "1.0541916896761134`", ",", 
     "1.0541916896761134`", ",", "1.0541916896761134`", ",", 
     "1.0899356965929945`", ",", "1.1439476178889973`", ",", 
     "1.1439476178889973`", ",", "1.2809175712350307`", ",", 
     "1.0465686294734031`", ",", "1.0465686294734031`", ",", 
     "1.0465686294734031`", ",", "1.0465686294734031`", ",", 
     "1.0465686294734031`", ",", "0.8393086294062587`", ",", 
     "0.9842418153154198`", ",", "0.9842418153154198`", ",", 
     "1.1695056438956426`", ",", "0.8423719860128329`", ",", 
     "0.858971878628558`", ",", "0.858971878628558`", ",", 
     "0.9238204105953135`", ",", "0.9238204105953135`", ",", 
     "1.1879733613666275`", ",", "0.9700662212471156`", ",", 
     "1.1068736363893237`", ",", "1.0646294415989945`", ",", 
     "0.9367734607981895`", ",", "0.9367734607981895`", ",", 
     "1.1009847950142002`", ",", "1.1009847950142002`", ",", 
     "1.1009847950142002`", ",", "1.1009847950142002`", ",", 
     "0.8941315687241456`", ",", "0.945433039252175`", ",", 
     "0.903630614935721`", ",", "0.903630614935721`", ",", 
     "0.903630614935721`", ",", "0.903630614935721`", ",", 
     "0.903630614935721`", ",", "0.903630614935721`", ",", 
     "1.0362645454489618`", ",", "1.0362645454489618`", ",", 
     "1.0362645454489618`", ",", "1.074979455820284`", ",", 
     "1.0157144640664266`", ",", "1.0157144640664266`", ",", 
     "1.104742631258603`", ",", "1.104742631258603`", ",", 
     "1.104742631258603`", ",", "0.9236230556469437`", ",", 
     "1.0946396452683635`", ",", "1.0946396452683635`", ",", 
     "1.0946396452683635`", ",", "0.854839202604982`", ",", 
     "0.854839202604982`", ",", "0.854839202604982`", ",", 
     "0.9728562025640439`", ",", "0.9728562025640439`", ",", 
     "0.9345880164736555`", ",", "0.9345880164736555`", ",", 
     "0.6907639138118514`", ",", "0.6907639138118514`", ",", 
     "0.6907639138118514`", ",", "0.6907639138118514`", ",", 
     "0.6907639138118514`", ",", "0.6907639138118514`", ",", 
     "0.6907639138118514`", ",", "0.6907639138118514`", ",", 
     "0.8041162987528758`", ",", "0.8041162987528758`", ",", 
     "0.7496883880698049`", ",", "0.6841819438546064`", ",", 
     "0.7413294739939366`", ",", "0.7413294739939366`", ",", 
     "0.7413294739939366`", ",", "0.7876808966461828`", ",", 
     "0.7876808966461828`", ",", "0.7876808966461828`", ",", 
     "0.7876808966461828`", ",", "0.7876808966461828`", ",", 
     "0.7804274405250444`", ",", "0.9375732676564822`", ",", 
     "0.942410343955175`", ",", "0.942410343955175`", ",", 
     "0.579267027213962`", ",", "0.579267027213962`", ",", 
     "0.579267027213962`", ",", "0.579267027213962`", ",", 
     "0.8393034046310671`", ",", "0.8169079820132462`", ",", 
     "0.8169079820132462`", ",", "0.6438262336036928`", ",", 
     "0.6438262336036928`", ",", "0.6438262336036928`", ",", 
     "0.6438262336036928`", ",", "0.6438262336036928`", ",", 
     "0.6438262336036928`", ",", "0.6871074805571473`", ",", 
     "0.8349538330111059`", ",", "0.8349538330111059`", ",", 
     "0.8349538330111059`", ",", "0.8349538330111059`", ",", 
     "0.8284508889601088`", ",", "0.8284508889601088`", ",", 
     "0.8284508889601088`", ",", "0.6550716166310652`", ",", 
     "0.6550716166310652`", ",", "0.815585350016977`", ",", 
     "0.7446632972496867`", ",", "0.764934416806068`", ",", 
     "0.7328390707311971`", ",", "0.8564357351571985`", ",", 
     "0.8564357351571985`", ",", "0.8564357351571985`", ",", 
     "0.8564357351571985`", ",", "0.5939041032513026`", ",", 
     "0.5939041032513026`", ",", "0.5939041032513026`", ",", 
     "0.5939041032513026`", ",", "0.5939041032513026`", ",", 
     "0.956218557274092`", ",", "0.8234858236049727`", ",", 
     "0.740281675994314`", ",", "0.740281675994314`", ",", 
     "0.8046558386854923`", ",", "0.8046558386854923`", ",", 
     "0.6933348677506117`", ",", "0.6933348677506117`", ",", 
     "0.6933348677506117`", ",", "0.6933348677506117`", ",", 
     "0.6634084853014209`", ",", "0.6634084853014209`", ",", 
     "0.6634084853014209`", ",", "0.6634084853014209`", ",", 
     "0.6634084853014209`", ",", "0.6634084853014209`", ",", 
     "0.7454126887957722`", ",", "0.7395117785489651`", ",", 
     "0.8055784845383867`", ",", "0.8055784845383867`", ",", 
     "0.8055784845383867`", ",", "0.6336304687999179`", ",", 
     "0.6336304687999179`", ",", "0.6336304687999179`", ",", 
     "0.6336304687999179`", ",", "0.6336304687999179`", ",", 
     "0.6336304687999179`", ",", "0.6336304687999179`", ",", 
     "0.7759754932562994`", ",", "0.7759754932562994`", ",", 
     "0.7759754932562994`", ",", "0.7759754932562994`", ",", 
     "0.7759754932562994`", ",", "0.6402355175620247`", ",", 
     "0.8618011118424433`", ",", "0.8618011118424433`", ",", 
     "0.8618011118424433`", ",", "0.8618011118424433`", ",", 
     "0.8618011118424433`", ",", "0.8618011118424433`", ",", 
     "0.8618011118424433`", ",", "0.8618011118424433`", ",", 
     "0.8618011118424433`", ",", "0.8618011118424433`", ",", 
     "0.8618011118424433`", ",", "0.8618011118424433`", ",", 
     "0.8618011118424433`", ",", "0.7456968606015629`", ",", 
     "0.7456968606015629`", ",", "0.8783890299291334`", ",", 
     "0.8783890299291334`", ",", "0.9857090563023315`", ",", 
     "0.8706699230427246`", ",", "0.8706699230427246`", ",", 
     "0.8706699230427246`", ",", "0.8706699230427246`", ",", 
     "0.8706699230427246`", ",", "0.8706699230427246`", ",", 
     "0.8706699230427246`", ",", "0.8778528357815818`", ",", 
     "0.8778528357815818`", ",", "0.8778528357815818`", ",", 
     "0.8778528357815818`", ",", "0.8778528357815818`", ",", 
     "0.8778528357815818`", ",", "0.8778528357815818`", ",", 
     "0.8778528357815818`", ",", "0.8778528357815818`", ",", 
     "0.8778528357815818`", ",", "0.7668982340854407`", ",", 
     "0.9404885081383451`", ",", "0.9404885081383451`", ",", 
     "0.9404885081383451`", ",", "0.7738396497096686`", ",", 
     "0.7738396497096686`", ",", "0.9930196946097956`", ",", 
     "0.9172189987126272`", ",", "0.9172189987126272`", ",", 
     "0.8548912376010785`", ",", "0.8548912376010785`", ",", 
     "0.8548912376010785`", ",", "0.9181494341946528`", ",", 
     "0.9181494341946528`", ",", "0.9181494341946528`", ",", 
     "0.920473673985082`", ",", "0.8536712085948561`", ",", 
     "0.8536712085948561`", ",", "0.8684032641119832`", ",", 
     "0.8684032641119832`", ",", "0.8684032641119832`", ",", 
     "0.8684032641119832`", ",", "0.8684032641119832`", ",", 
     "0.8684032641119832`", ",", "0.8684032641119832`", ",", 
     "0.8684032641119832`", ",", "0.8684032641119832`", ",", 
     "0.8684032641119832`", ",", "0.8684032641119832`", ",", 
     "0.8684032641119832`", ",", "0.8684032641119832`", ",", 
     "0.8684032641119832`", ",", "0.9138089251327527`", ",", 
     "0.8666690742892906`", ",", "0.8666690742892906`", ",", 
     "0.8666690742892906`", ",", "0.8666690742892906`", ",", 
     "0.7900238551516073`", ",", "0.7900238551516073`", ",", 
     "0.7900238551516073`", ",", "0.7900238551516073`", ",", 
     "0.9253502417177324`", ",", "0.9253502417177324`", ",", 
     "0.9123394303089031`", ",", "0.9123394303089031`", ",", 
     "0.9123394303089031`", ",", "0.9123394303089031`", ",", 
     "0.9123394303089031`", ",", "0.8712773370605982`", ",", 
     "0.8712773370605982`", ",", "0.8712773370605982`", ",", 
     "0.8712773370605982`", ",", "0.8712773370605982`", ",", 
     "0.8712773370605982`", ",", "0.8712773370605982`", ",", 
     "0.8712773370605982`", ",", "0.7265622001512283`", ",", 
     "0.7265622001512283`", ",", "0.7265622001512283`", ",", 
     "0.7265622001512283`", ",", "0.7265622001512283`", ",", 
     "0.7265622001512283`", ",", "0.9624678274268975`", ",", 
     "0.8355447555261288`", ",", "0.6996966279047865`", ",", 
     "0.7346673507409893`", ",", "0.8798684629145397`", ",", 
     "0.8798684629145397`", ",", "0.8798684629145397`", ",", 
     "0.8798684629145397`", ",", "0.7845138311937093`", ",", 
     "0.7845138311937093`", ",", "0.7829847233468972`", ",", 
     "0.7829847233468972`", ",", "0.7945036670857122`", ",", 
     "0.7945036670857122`", ",", "0.7945036670857122`", ",", 
     "0.7945036670857122`", ",", "0.7945036670857122`", ",", 
     "0.7231705613096832`", ",", "0.7231705613096832`", ",", 
     "0.7231705613096832`", ",", "0.7231705613096832`", ",", 
     "0.8852352693845283`", ",", "0.8852352693845283`", ",", 
     "0.8852352693845283`", ",", "0.8852352693845283`", ",", 
     "0.8852352693845283`", ",", "0.8882795830221925`", ",", 
     "0.8882795830221925`", ",", "0.8882795830221925`", ",", 
     "0.8882795830221925`", ",", "0.8599406265690483`", ",", 
     "0.8599406265690483`", ",", "0.8599406265690483`", ",", 
     "0.8599406265690483`", ",", "0.8599406265690483`", ",", 
     "0.8588227126132628`", ",", "0.8588227126132628`", ",", 
     "0.9572572988641074`", ",", "0.9572572988641074`", ",", 
     "0.9572572988641074`", ",", "0.9572572988641074`", ",", 
     "0.9289190960434089`", ",", "0.9289190960434089`", ",", 
     "0.9289190960434089`", ",", "0.9289190960434089`", ",", 
     "0.9289190960434089`", ",", "0.9289190960434089`", ",", 
     "0.9289190960434089`", ",", "1.0217446101536671`", ",", 
     "1.0217446101536671`", ",", "1.0217446101536671`", ",", 
     "0.9438968571221729`", ",", "0.9438968571221729`", ",", 
     "0.9438968571221729`", ",", "0.9438968571221729`", ",", 
     "0.8297618971890062`", ",", "0.8297618971890062`", ",", 
     "0.8297618971890062`", ",", "0.8297618971890062`", ",", 
     "0.828806815520727`", ",", "0.828806815520727`", ",", 
     "0.828806815520727`", ",", "0.828806815520727`", ",", 
     "0.828806815520727`", ",", "0.8759448350132046`", ",", 
     "0.8759448350132046`", ",", "0.8759448350132046`", ",", 
     "0.8759448350132046`", ",", "0.8759448350132046`", ",", 
     "0.8759448350132046`", ",", "0.8759448350132046`", ",", 
     "0.8759448350132046`", ",", "0.7806381147291777`", ",", 
     "0.7806381147291777`", ",", "0.7806381147291777`", ",", 
     "0.7806381147291777`", ",", "0.8626215363359585`", ",", 
     "0.8626215363359585`", ",", "0.8626215363359585`", ",", 
     "0.8626215363359585`", ",", "0.8626215363359585`", ",", 
     "0.9421103593231004`", ",", "0.9421103593231004`", ",", 
     "0.9421103593231004`", ",", "0.9421103593231004`", ",", 
     "0.9421103593231004`", ",", "0.901136524137979`", ",", 
     "0.901136524137979`", ",", "0.6777824933715231`", ",", 
     "0.6777824933715231`", ",", "0.8347883697537113`", ",", 
     "0.8390184151127883`", ",", "0.8390184151127883`", ",", 
     "0.8390184151127883`", ",", "0.8390184151127883`", ",", 
     "0.9068104827081018`", ",", "0.9068104827081018`", ",", 
     "0.9068104827081018`", ",", "0.9068104827081018`", ",", 
     "0.9068104827081018`", ",", "0.8187656304550743`", ",", 
     "0.8187656304550743`", ",", "0.9531341753469529`", ",", 
     "0.9531341753469529`", ",", "0.8262618828827719`", ",", 
     "0.8586266292383924`", ",", "0.8586266292383924`", ",", 
     "0.8586266292383924`", ",", "0.8586266292383924`", ",", 
     "0.8586266292383924`", ",", "0.8586266292383924`", ",", 
     "0.8586266292383924`", ",", "0.7600002050373317`", ",", 
     "0.7990609723518004`", ",", "0.9962734901703961`", ",", 
     "0.9962734901703961`", ",", "0.8173365577379872`", ",", 
     "0.9733570452723843`", ",", "0.9733570452723843`", ",", 
     "0.8047772856789043`", ",", "0.8047772856789043`", ",", 
     "0.8047772856789043`", ",", "0.8047772856789043`", ",", 
     "0.980008408190078`", ",", "0.980008408190078`", ",", 
     "0.980008408190078`", ",", "0.7514388733067331`", ",", 
     "0.9185374973596496`", ",", "0.9185374973596496`", ",", 
     "0.9185374973596496`", ",", "0.9185374973596496`", ",", 
     "0.9185374973596496`", ",", "0.9873622418843341`", ",", 
     "0.9873622418843341`", ",", "0.98449310044091`", ",", 
     "0.8526096724484493`", ",", "0.8526096724484493`", ",", 
     "0.8526096724484493`", ",", "0.8526096724484493`", ",", 
     "0.8090432419863559`", ",", "0.8090432419863559`", ",", 
     "0.7807434979448477`", ",", "0.9700050086068992`", ",", 
     "0.89352851675192`", ",", "0.89352851675192`", ",", "0.89352851675192`", 
     ",", "0.8966182054961761`", ",", "0.7201223895596812`", ",", 
     "0.6622441625742863`", ",", "0.6622441625742863`", ",", 
     "0.6622441625742863`", ",", "0.6622441625742863`", ",", 
     "0.9631566271958303`", ",", "0.9631566271958303`", ",", 
     "0.9631566271958303`", ",", "0.9631566271958303`", ",", 
     "0.9631566271958303`", ",", "0.9337822766456845`", ",", 
     "0.9337822766456845`", ",", "0.9337822766456845`", ",", 
     "0.9337822766456845`", ",", "0.9337822766456845`", ",", 
     "0.9337822766456845`", ",", "0.9337822766456845`", ",", 
     "0.9337822766456845`", ",", "0.9337822766456845`", ",", 
     "0.9337822766456845`", ",", "0.9369761839654742`", ",", 
     "0.9369761839654742`", ",", "0.9369761839654742`", ",", 
     "0.9369761839654742`", ",", "0.925057726788959`", ",", 
     "0.925057726788959`", ",", "0.7262927378156769`", ",", 
     "0.7744965145790026`", ",", "0.7744965145790026`", ",", 
     "0.9110994229658812`", ",", "0.9110994229658812`", ",", 
     "0.9110994229658812`", ",", "0.8315024410464849`", ",", 
     "0.8315024410464849`", ",", "0.8315024410464849`", ",", 
     "0.8315024410464849`", ",", "0.8315024410464849`", ",", 
     "0.8315024410464849`", ",", "0.8177414498758009`", ",", 
     "0.8177414498758009`", ",", "0.8177414498758009`", ",", 
     "0.8177414498758009`", ",", "0.8177414498758009`", ",", 
     "0.8177414498758009`", ",", "0.8177414498758009`", ",", 
     "0.8177414498758009`", ",", "0.8177414498758009`", ",", 
     "0.8177414498758009`", ",", "0.8812997184956184`", ",", 
     "0.8812997184956184`", ",", "0.9211498721014595`", ",", 
     "0.9211498721014595`", ",", "0.9211498721014595`", ",", 
     "0.9211498721014595`", ",", "0.9211498721014595`", ",", 
     "0.8477626748325771`", ",", "0.8249011888841365`", ",", 
     "0.8249011888841365`", ",", "0.8249011888841365`", ",", 
     "0.8249011888841365`", ",", "0.8249011888841365`", ",", 
     "0.8249011888841365`", ",", "0.8249011888841365`", ",", 
     "0.8249011888841365`", ",", "0.8249011888841365`", ",", 
     "0.8758286316319592`", ",", "0.8758286316319592`", ",", 
     "0.8758286316319592`", ",", "0.8758286316319592`", ",", 
     "0.8758286316319592`", ",", "0.8758286316319592`", ",", 
     "0.8758286316319592`", ",", "0.8758286316319592`", ",", 
     "0.8282023742619213`", ",", "0.8282023742619213`", ",", 
     "0.8282023742619213`", ",", "0.8282023742619213`", ",", 
     "0.8282023742619213`", ",", "0.8282023742619213`", ",", 
     "0.8282023742619213`", ",", "0.8282023742619213`", ",", 
     "0.8282023742619213`", ",", "0.8282023742619213`", ",", 
     "0.7927161941044294`", ",", "0.7927161941044294`", ",", 
     "0.7927161941044294`", ",", "0.7927161941044294`", ",", 
     "0.7818643912259128`", ",", "0.7818643912259128`", ",", 
     "0.9309304551000995`", ",", "0.9309304551000995`", ",", 
     "0.9309304551000995`", ",", "0.9309304551000995`", ",", 
     "0.7420953619306659`", ",", "0.7420953619306659`", ",", 
     "0.8008773655460878`", ",", "0.9934830737594273`", ",", 
     "0.9934830737594273`", ",", "0.9123432605229647`", ",", 
     "0.9123432605229647`", ",", "0.8851355852908197`", ",", 
     "0.8851355852908197`", ",", "0.8851355852908197`", ",", 
     "0.8851355852908197`", ",", "0.8851355852908197`", ",", 
     "0.8106617798283116`", ",", "0.9048988104119126`", ",", 
     "0.9048988104119126`", ",", "0.794525163098345`", ",", 
     "0.794525163098345`", ",", "0.794525163098345`", ",", 
     "0.7185158967002041`", ",", "0.7341023245247512`", ",", 
     "0.8440668154276797`", ",", "0.8440668154276797`", ",", 
     "0.8446850905760589`", ",", "0.8446850905760589`", ",", 
     "0.8446850905760589`", ",", "0.8446850905760589`", ",", 
     "0.8446850905760589`", ",", "0.8446850905760589`", ",", 
     "0.8668980096239793`", ",", "0.8668980096239793`", ",", 
     "0.8668980096239793`", ",", "0.8668980096239793`", ",", 
     "0.8668980096239793`", ",", "0.8668980096239793`", ",", 
     "0.8668980096239793`", ",", "0.9922545870324443`", ",", 
     "0.9922545870324443`", ",", "0.9922545870324443`", ",", 
     "0.9922545870324443`", ",", "0.9922545870324443`", ",", 
     "0.9922545870324443`", ",", "0.9922545870324443`", ",", 
     "0.7558521599698228`", ",", "0.7558521599698228`", ",", 
     "0.9621344987472563`", ",", "0.9621344987472563`", ",", 
     "0.9621344987472563`", ",", "0.9621344987472563`", ",", 
     "0.9621344987472563`", ",", "0.9621344987472563`", ",", 
     "0.9621344987472563`", ",", "1.032537939749777`", ",", 
     "1.032537939749777`", ",", "0.8398036669273287`", ",", 
     "0.9850248004186853`", ",", "0.9850248004186853`", ",", 
     "0.9850248004186853`", ",", "0.9850248004186853`", ",", 
     "1.0207138407168521`", ",", "0.9938195734967812`", ",", 
     "0.9938195734967812`", ",", "0.9938195734967812`", ",", 
     "0.9938195734967812`", ",", "0.940226499085348`", ",", 
     "0.940226499085348`", ",", "0.940226499085348`", ",", 
     "0.940226499085348`", ",", "0.940226499085348`", ",", 
     "0.856165611983807`", ",", "0.856165611983807`", ",", 
     "0.856165611983807`", ",", "0.856165611983807`", ",", 
     "0.856165611983807`", ",", "0.8423014063661546`", ",", 
     "0.7199005842570132`", ",", "0.7199005842570132`", ",", 
     "0.7199005842570132`", ",", "0.7199005842570132`", ",", 
     "0.7199005842570132`", ",", "0.7199005842570132`", ",", 
     "0.7199005842570132`", ",", "0.7199005842570132`", ",", 
     "0.7833801729898051`", ",", "0.7833801729898051`", ",", 
     "0.7807679548678691`", ",", "0.7807679548678691`", ",", 
     "0.7807679548678691`", ",", "0.8058121496658694`", ",", 
     "0.8593669588093208`", ",", "0.8593669588093208`", ",", 
     "0.8593669588093208`", ",", "0.8593669588093208`", ",", 
     "0.8593669588093208`", ",", "0.8593669588093208`", ",", 
     "0.8593669588093208`", ",", "0.8593669588093208`", ",", 
     "0.934336832831933`", ",", "0.934336832831933`", ",", 
     "0.934336832831933`", ",", "0.934336832831933`", ",", 
     "0.934336832831933`", ",", "0.934336832831933`", ",", 
     "1.0321951745507296`", ",", "1.0321951745507296`", ",", 
     "0.9909308119506425`", ",", "0.9909308119506425`", ",", 
     "0.9909308119506425`", ",", "0.9909308119506425`", ",", 
     "0.8116760539427833`", ",", "0.8116760539427833`", ",", 
     "0.8077375520205642`", ",", "0.8077375520205642`", ",", 
     "0.8424975452075418`", ",", "0.8424975452075418`", ",", 
     "0.8424975452075418`", ",", "0.8424975452075418`", ",", 
     "0.8424975452075418`", ",", "0.8424975452075418`", ",", 
     "0.9427793376312231`", ",", "0.9427793376312231`", ",", 
     "0.9427793376312231`", ",", "0.9427793376312231`", ",", 
     "0.9427793376312231`", ",", "0.9427793376312231`", ",", 
     "0.8946432492629711`", ",", "0.8946432492629711`", ",", 
     "0.8946432492629711`", ",", "0.8946432492629711`", ",", 
     "0.8946432492629711`", ",", "0.9252467089649452`", ",", 
     "0.9252467089649452`", ",", "0.9252467089649452`", ",", 
     "0.9252467089649452`", ",", "0.9160582385642004`", ",", 
     "0.9160582385642004`", ",", "0.9160582385642004`", ",", 
     "0.9160582385642004`", ",", "0.9160582385642004`", ",", 
     "0.9160582385642004`", ",", "0.9160582385642004`", ",", 
     "0.9160582385642004`", ",", "0.9160582385642004`", ",", 
     "0.9160582385642004`", ",", "0.9160582385642004`", ",", 
     "0.9160582385642004`", ",", "0.9160582385642004`", ",", 
     "0.9160582385642004`", ",", "0.7763308254982106`", ",", 
     "0.7763308254982106`", ",", "0.7763308254982106`", ",", 
     "0.7763308254982106`", ",", "0.7763308254982106`", ",", 
     "0.7763308254982106`", ",", "0.7763308254982106`", ",", 
     "0.7763308254982106`", ",", "0.7763308254982106`", ",", 
     "0.8963999624400187`", ",", "0.8963999624400187`", ",", 
     "0.8963999624400187`", ",", "0.8963999624400187`", ",", 
     "0.8963999624400187`", ",", "0.8963999624400187`", ",", 
     "0.8963999624400187`", ",", "0.8963999624400187`", ",", 
     "0.8963999624400187`", ",", "0.833025882810429`", ",", 
     "0.833025882810429`", ",", "0.8518965382053589`", ",", 
     "0.8518965382053589`", ",", "0.8518965382053589`", ",", 
     "0.8518965382053589`", ",", "0.8312513754254418`", ",", 
     "0.8073269398565792`", ",", "0.9816284983802076`", ",", 
     "0.9816284983802076`", ",", "0.9816284983802076`", ",", 
     "0.9816284983802076`", ",", "0.9741417353328132`", ",", 
     "0.9741417353328132`", ",", "0.9741417353328132`", ",", 
     "0.9741417353328132`", ",", "0.9024896752408618`", ",", 
     "0.8534244230208208`", ",", "0.8950695840398492`", ",", 
     "0.809431157404559`", ",", "0.809431157404559`", ",", 
     "0.809431157404559`", ",", "0.809431157404559`", ",", 
     "0.809431157404559`", ",", "0.809431157404559`", ",", 
     "0.9786537801475002`", ",", "0.8554806156611685`", ",", 
     "0.8554806156611685`", ",", "0.8554806156611685`", ",", 
     "0.8554806156611685`", ",", "0.8554806156611685`", ",", 
     "0.8554806156611685`", ",", "0.8554806156611685`", ",", 
     "0.9050001536175534`", ",", "0.8994056215100459`", ",", 
     "0.8994056215100459`", ",", "0.8994056215100459`", ",", 
     "0.8994056215100459`", ",", "0.8257996566555034`", ",", 
     "0.8257996566555034`", ",", "0.8257996566555034`", ",", 
     "0.8257996566555034`", ",", "0.912586487800592`", ",", 
     "0.7972562117479662`", ",", "0.7972562117479662`", ",", 
     "0.7972562117479662`", ",", "0.7972562117479662`", ",", 
     "0.7972562117479662`", ",", "0.7972562117479662`", ",", 
     "0.7972562117479662`", ",", "0.7972562117479662`", ",", 
     "0.7972562117479662`", ",", "0.7972562117479662`", ",", 
     "0.7972562117479662`", ",", "1.0047334499314327`", ",", 
     "0.7306344764342747`", ",", "0.7306344764342747`", ",", 
     "0.7306344764342747`", ",", "0.7306344764342747`", ",", 
     "0.8972251513450107`", ",", "0.8972251513450107`", ",", 
     "0.8972251513450107`", ",", "0.8972251513450107`", ",", 
     "0.8972251513450107`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8762293308306099`", ",", 
     "0.8762293308306099`", ",", "0.8283081860287752`", ",", 
     "0.8283081860287752`", ",", "0.953185799158399`", ",", 
     "0.953185799158399`", ",", "0.8973784893968935`", ",", 
     "0.8973784893968935`", ",", "0.8973784893968935`", ",", 
     "0.8973784893968935`", ",", "0.8973784893968935`", ",", 
     "0.8973784893968935`", ",", "0.8973784893968935`", ",", 
     "0.9091253657171481`", ",", "0.9091253657171481`", ",", 
     "0.9091253657171481`", ",", "0.9091253657171481`", ",", 
     "0.9091253657171481`", ",", "0.9091253657171481`", ",", 
     "0.9091253657171481`", ",", "0.881113359369732`", ",", 
     "0.881113359369732`", ",", "0.881113359369732`", ",", 
     "0.881113359369732`", ",", "0.881113359369732`", ",", 
     "0.8635624470602784`", ",", "0.8635624470602784`", ",", 
     "0.8635624470602784`", ",", "0.8635624470602784`", ",", 
     "0.8635624470602784`", ",", "0.8635624470602784`", ",", 
     "0.8635624470602784`", ",", "0.8635624470602784`", ",", 
     "0.8635624470602784`", ",", "0.8293660603309272`", ",", 
     "0.8293660603309272`", ",", "0.8293660603309272`", ",", 
     "0.8293660603309272`", ",", "0.8293660603309272`", ",", 
     "0.8293660603309272`", ",", "0.8784332265979978`", ",", 
     "0.8784332265979978`", ",", "0.8784332265979978`", ",", 
     "0.8784332265979978`", ",", "0.8784332265979978`", ",", 
     "0.8784332265979978`", ",", "0.9296563084310686`", ",", 
     "0.9296563084310686`", ",", "0.9296563084310686`", ",", 
     "0.9296563084310686`", ",", "0.841801504348049`", ",", 
     "0.841801504348049`", ",", "0.841801504348049`", ",", 
     "0.841801504348049`", ",", "0.841801504348049`", ",", 
     "0.841801504348049`", ",", "0.841801504348049`", ",", 
     "0.841801504348049`", ",", "0.841801504348049`", ",", 
     "0.841801504348049`", ",", "0.841801504348049`", ",", 
     "0.841801504348049`", ",", "0.7790871897580925`", ",", 
     "0.7790871897580925`", ",", "0.7790871897580925`", ",", 
     "0.9178820689769195`", ",", "0.9178820689769195`", ",", 
     "0.9178820689769195`", ",", "0.9178820689769195`", ",", 
     "0.9178820689769195`", ",", "0.8553007020345287`", ",", 
     "0.8553007020345287`", ",", "0.8553007020345287`", ",", 
     "0.8553007020345287`", ",", "0.8553007020345287`", ",", 
     "0.8553007020345287`", ",", "0.8553007020345287`", ",", 
     "0.8553007020345287`", ",", "0.8553007020345287`", ",", 
     "0.8553007020345287`", ",", "0.8627172157832488`", ",", 
     "0.8627172157832488`", ",", "0.8627172157832488`", ",", 
     "0.8627172157832488`", ",", "0.8627172157832488`", ",", 
     "0.9764841022910986`", ",", "0.9764841022910986`", ",", 
     "0.9764841022910986`", ",", "0.9211418924591437`", ",", 
     "0.9211418924591437`", ",", "0.9211418924591437`", ",", 
     "0.9119634864222395`", ",", "0.9119634864222395`", ",", 
     "0.9119634864222395`", ",", "0.8842558969291714`", ",", 
     "0.903974219047887`", ",", "0.903974219047887`", ",", 
     "0.903974219047887`", ",", "0.903974219047887`", ",", 
     "0.903974219047887`", ",", "0.903974219047887`", ",", 
     "0.903974219047887`", ",", "0.903974219047887`", ",", 
     "0.903974219047887`", ",", "0.903974219047887`", ",", 
     "0.903974219047887`", ",", "0.903974219047887`", ",", 
     "0.903974219047887`", ",", "0.9198631330059325`", ",", 
     "0.9198631330059325`", ",", "0.9198631330059325`", ",", 
     "0.7977324083277089`", ",", "0.7977324083277089`", ",", 
     "0.7977324083277089`", ",", "0.7977324083277089`", ",", 
     "0.7977324083277089`", ",", "0.7977324083277089`", ",", 
     "0.7977324083277089`", ",", "0.8764341227006245`", ",", 
     "0.8764341227006245`", ",", "0.8764341227006245`", ",", 
     "0.8764341227006245`", ",", "0.8111768489370709`", ",", 
     "0.8111768489370709`", ",", "0.8111768489370709`", ",", 
     "0.8111768489370709`", ",", "0.8899784420323231`", ",", 
     "0.8899784420323231`", ",", "0.8899784420323231`", ",", 
     "0.8899784420323231`", ",", "0.8899784420323231`", ",", 
     "0.8899784420323231`", ",", "0.8899784420323231`", ",", 
     "0.8899784420323231`", ",", "0.8899784420323231`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.8792805208830891`", ",", "0.8792805208830891`", ",", 
     "0.9278655501760277`", ",", "0.9278655501760277`", ",", 
     "0.9278655501760277`", ",", "0.9278655501760277`", ",", 
     "0.9278655501760277`", ",", "0.9278655501760277`", ",", 
     "0.8195207789082088`", ",", "0.8195207789082088`", ",", 
     "0.8195207789082088`", ",", "0.8195207789082088`", ",", 
     "0.8195207789082088`", ",", "0.8021884948955538`", ",", 
     "0.8021884948955538`", ",", "0.8021884948955538`", ",", 
     "0.8021884948955538`", ",", "0.8021884948955538`", ",", 
     "0.8021884948955538`", ",", "0.8021884948955538`", ",", 
     "0.8021884948955538`", ",", "0.955525235350493`", ",", 
     "0.955525235350493`", ",", "0.955525235350493`", ",", 
     "0.955525235350493`", ",", "0.955525235350493`", ",", 
     "0.9151564600708718`", ",", "0.9151564600708718`", ",", 
     "0.8940018507543445`", ",", "0.8940018507543445`", ",", 
     "0.8940018507543445`", ",", "0.8940018507543445`", ",", 
     "0.8940018507543445`", ",", "0.8940018507543445`", ",", 
     "0.8940018507543445`", ",", "0.8940018507543445`", ",", 
     "0.8940018507543445`", ",", "0.8940018507543445`", ",", 
     "0.816889580068024`", ",", "0.8517619953980016`", ",", 
     "0.8517619953980016`", ",", "0.8517619953980016`", ",", 
     "0.8517619953980016`", ",", "0.8517619953980016`", ",", 
     "0.8517619953980016`", ",", "0.8517619953980016`", ",", 
     "0.8517619953980016`", ",", "0.8375978678682549`", ",", 
     "0.8375978678682549`", ",", "0.8375978678682549`", ",", 
     "0.8375978678682549`", ",", "0.8375978678682549`", ",", 
     "0.8375978678682549`", ",", "0.8375978678682549`", ",", 
     "0.8375978678682549`", ",", "0.8375978678682549`", ",", 
     "0.8375978678682549`", ",", "0.8735875459547988`", ",", 
     "0.9368345680894112`", ",", "0.9368345680894112`", ",", 
     "0.9368345680894112`", ",", "0.9368345680894112`", ",", 
     "0.9305018278658`", ",", "0.9290421725301017`", ",", 
     "0.924200182163452`", ",", "0.924200182163452`", ",", 
     "0.924200182163452`", ",", "0.8953500508588166`", ",", 
     "0.8953500508588166`", ",", "0.8953500508588166`", ",", 
     "0.8953500508588166`", ",", "0.8953500508588166`", ",", 
     "0.8953500508588166`", ",", "0.8953500508588166`", ",", 
     "0.8953500508588166`", ",", "0.8174262322660453`", ",", 
     "0.8174262322660453`", ",", "0.8174262322660453`", ",", 
     "0.8174262322660453`", ",", "0.8174262322660453`", ",", 
     "0.87016327370827`", ",", "0.87016327370827`", ",", "0.87016327370827`", 
     ",", "0.87016327370827`", ",", "0.87016327370827`", ",", 
     "0.87016327370827`", ",", "0.87016327370827`", ",", "0.87016327370827`", 
     ",", "0.87016327370827`", ",", "0.87016327370827`", ",", 
     "0.87016327370827`", ",", "0.87016327370827`", ",", "0.87016327370827`", 
     ",", "0.87016327370827`", ",", "0.87016327370827`", ",", 
     "0.87016327370827`", ",", "0.87016327370827`", ",", "0.87016327370827`", 
     ",", "0.87016327370827`", ",", "0.87016327370827`", ",", 
     "0.9410350908432255`", ",", "0.9410350908432255`", ",", 
     "0.8746119113294621`", ",", "0.8746119113294621`", ",", 
     "0.9043018362737487`", ",", "0.8432840234785126`", ",", 
     "0.8432840234785126`", ",", "0.8432840234785126`", ",", 
     "0.8432840234785126`", ",", "0.8432840234785126`", ",", 
     "0.9258467800856062`", ",", "0.9258467800856062`", ",", 
     "0.9258467800856062`", ",", "0.9258467800856062`", ",", 
     "0.9258467800856062`", ",", "0.9258467800856062`", ",", 
     "0.9258467800856062`", ",", "0.9258467800856062`", ",", 
     "0.9258467800856062`", ",", "0.9258467800856062`", ",", 
     "0.9258467800856062`", ",", "0.9192325020473111`", ",", 
     "0.9192325020473111`", ",", "0.9192325020473111`", ",", 
     "0.9192325020473111`", ",", "0.9192325020473111`", ",", 
     "0.9192325020473111`", ",", "0.9192325020473111`", ",", 
     "0.9192325020473111`", ",", "0.9192325020473111`", ",", 
     "0.9192325020473111`", ",", "0.9192325020473111`", ",", 
     "0.9192325020473111`", ",", "0.9192325020473111`", ",", 
     "0.8673155735575688`", ",", "0.8673155735575688`", ",", 
     "0.8673155735575688`", ",", "0.8673155735575688`", ",", 
     "0.8673155735575688`", ",", "0.790886835619342`", ",", 
     "0.790886835619342`", ",", "0.790886835619342`", ",", 
     "0.790886835619342`", ",", "0.8816947384547921`", ",", 
     "0.8816947384547921`", ",", "0.8311544528790196`", ",", 
     "0.8311544528790196`", ",", "0.9058159637597228`", ",", 
     "0.9058159637597228`", ",", "0.8105976891585727`", ",", 
     "0.8105976891585727`", ",", "0.8105976891585727`", ",", 
     "0.8105976891585727`", ",", "0.8105976891585727`", ",", 
     "0.8105976891585727`", ",", "0.8105976891585727`", ",", 
     "0.8105976891585727`", ",", "0.8585656542867717`", ",", 
     "0.8585656542867717`", ",", "0.8585656542867717`", ",", 
     "0.8585656542867717`", ",", "0.7924845716465327`", ",", 
     "0.7924845716465327`", ",", "0.7924845716465327`", ",", 
     "0.8599032720126327`", ",", "0.8599032720126327`", ",", 
     "0.8599032720126327`", ",", "0.8599032720126327`", ",", 
     "0.8599032720126327`", ",", "0.8599032720126327`", ",", 
     "0.8599032720126327`", ",", "0.8599032720126327`", ",", 
     "0.8599032720126327`", ",", "0.8599032720126327`", ",", 
     "0.8599032720126327`", ",", "0.8979366893299047`", ",", 
     "0.8979366893299047`", ",", "0.8979366893299047`", ",", 
     "0.8979366893299047`", ",", "0.8215033983608238`", ",", 
     "0.8215033983608238`", ",", "0.8325652432925713`", ",", 
     "0.8325652432925713`", ",", "0.8325652432925713`", ",", 
     "0.8325652432925713`", ",", "0.8538236538635068`", ",", 
     "0.8538236538635068`", ",", "0.8538236538635068`", ",", 
     "0.8538236538635068`", ",", "0.8538236538635068`", ",", 
     "0.8538236538635068`", ",", "0.9044240690599017`", ",", 
     "0.9044240690599017`", ",", "0.9044240690599017`", ",", 
     "0.9044240690599017`", ",", "0.9044240690599017`", ",", 
     "0.9044240690599017`", ",", "0.9044240690599017`", ",", 
     "0.9044240690599017`", ",", "0.820078385597538`", ",", 
     "0.820078385597538`", ",", "0.820078385597538`", ",", 
     "0.820078385597538`", ",", "0.820078385597538`", ",", 
     "0.820078385597538`", ",", "0.820078385597538`", ",", 
     "0.820078385597538`", ",", "0.8966701021218827`", ",", 
     "0.8966701021218827`", ",", "0.8966701021218827`", ",", 
     "0.8966701021218827`", ",", "0.8966701021218827`", ",", 
     "0.8902432767476998`", ",", "0.8902432767476998`", ",", 
     "0.8902432767476998`", ",", "0.8902432767476998`", ",", 
     "0.8902432767476998`", ",", "0.8902432767476998`", ",", 
     "0.8829979067846422`", ",", "0.8829979067846422`", ",", 
     "0.8829979067846422`", ",", "0.8829979067846422`", ",", 
     "0.8829979067846422`", ",", "0.9027596649472226`", ",", 
     "0.9027596649472226`", ",", "0.9027596649472226`", ",", 
     "0.9027596649472226`", ",", "0.9027596649472226`", ",", 
     "0.9027596649472226`", ",", "0.9027596649472226`", ",", 
     "0.9027596649472226`", ",", "0.9027596649472226`", ",", 
     "0.9027596649472226`", ",", "0.9027596649472226`", ",", 
     "0.9027596649472226`", ",", "0.9027596649472226`", ",", 
     "0.9097049213385452`", ",", "0.8808202057129697`", ",", 
     "0.8808202057129697`", ",", "0.8808202057129697`", ",", 
     "0.909168325876275`", ",", "0.909168325876275`", ",", 
     "0.909168325876275`", ",", "0.909168325876275`", ",", 
     "0.909168325876275`", ",", "0.909168325876275`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8678334334958295`", ",", "0.8678334334958295`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8761042922138093`", ",", 
     "0.8761042922138093`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8970677893058201`", ",", "0.8970677893058201`", ",", 
     "0.8343128879468324`", ",", "0.8233916818827667`", ",", 
     "0.8233916818827667`", ",", "0.8233916818827667`", ",", 
     "0.8233916818827667`", ",", "0.8233916818827667`", ",", 
     "0.8233916818827667`", ",", "0.8233916818827667`", ",", 
     "0.8233916818827667`", ",", "0.8233916818827667`", ",", 
     "0.8629801345513415`", ",", "0.8629801345513415`", ",", 
     "0.8629801345513415`", ",", "0.8629801345513415`", ",", 
     "0.8629801345513415`", ",", "0.8629801345513415`", ",", 
     "0.8629801345513415`", ",", "0.8629801345513415`", ",", 
     "0.8629801345513415`", ",", "0.8876840857131181`", ",", 
     "0.8876840857131181`", ",", "0.8821726341175833`", ",", 
     "0.8821726341175833`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.8829939381941182`", ",", 
     "0.8829939381941182`", ",", "0.834367837120993`", ",", 
     "0.834367837120993`", ",", "0.834367837120993`", ",", 
     "0.834367837120993`", ",", "0.834367837120993`", ",", 
     "0.834367837120993`", ",", "0.834367837120993`", ",", 
     "0.834367837120993`", ",", "0.834367837120993`", ",", 
     "0.8854225362526758`", ",", "0.8854225362526758`", ",", 
     "0.8854225362526758`", ",", "0.8854225362526758`", ",", 
     "0.8854225362526758`", ",", "0.8854225362526758`", ",", 
     "0.8854225362526758`", ",", "0.9595741282239989`", ",", 
     "0.9595741282239989`", ",", "0.9595741282239989`", ",", 
     "0.9595741282239989`", ",", "0.9595741282239989`", ",", 
     "0.9595741282239989`", ",", "0.9051693094396794`", ",", 
     "0.9051693094396794`", ",", "0.9051693094396794`", ",", 
     "0.9051693094396794`", ",", "0.9051693094396794`", ",", 
     "0.9051693094396794`", ",", "0.9051693094396794`", ",", 
     "0.9051693094396794`", ",", "0.9051693094396794`", ",", 
     "0.9051693094396794`", ",", "0.9051693094396794`", ",", 
     "0.9051693094396794`", ",", "0.9051693094396794`", ",", 
     "0.9051693094396794`", ",", "0.9051693094396794`", ",", 
     "0.9291384824293573`", ",", "0.8969675351853725`", ",", 
     "0.8969675351853725`", ",", "0.8953284700712206`", ",", 
     "0.8953284700712206`", ",", "0.8953284700712206`", ",", 
     "0.8953284700712206`", ",", "0.8953284700712206`", ",", 
     "0.9031375488528192`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.8942891576643224`", ",", 
     "0.8942891576643224`", ",", "0.9274161839351315`", ",", 
     "0.9274161839351315`", ",", "0.9274161839351315`", ",", 
     "0.9274161839351315`", ",", "0.9274161839351315`", ",", 
     "0.88424815769141`", ",", "0.88424815769141`", ",", "0.88424815769141`", 
     ",", "0.88424815769141`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.9072674980907545`", ",", "0.9072674980907545`", ",", 
     "0.8712600405426771`", ",", "0.8233680578773774`", ",", 
     "0.8233680578773774`", ",", "0.8233680578773774`", ",", 
     "0.8233680578773774`", ",", "0.8233680578773774`", ",", 
     "0.8233680578773774`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.9280099238952273`", ",", 
     "0.9280099238952273`", ",", "0.8964376629907466`", ",", 
     "0.8964376629907466`", ",", "0.8964376629907466`", ",", 
     "0.8964376629907466`", ",", "0.8964376629907466`", ",", 
     "0.8964376629907466`", ",", "0.8964376629907466`", ",", 
     "0.8964376629907466`", ",", "0.8964376629907466`", ",", 
     "0.8964376629907466`", ",", "0.8964376629907466`", ",", 
     "0.876590753981502`", ",", "0.8679362248176901`", ",", 
     "0.9796612291027514`", ",", "0.9796612291027514`", ",", 
     "0.9796612291027514`", ",", "0.9796612291027514`", ",", 
     "0.9796612291027514`", ",", "0.9796612291027514`", ",", 
     "0.9796612291027514`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9950156427316482`", ",", 
     "0.9950156427316482`", ",", "0.9789544841103694`", ",", 
     "0.9789544841103694`", ",", "0.9789544841103694`", ",", 
     "0.9789544841103694`", ",", "0.9789544841103694`", ",", 
     "0.9789544841103694`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.8623724005302398`", ",", 
     "0.8623724005302398`", ",", "0.9527699837492861`", ",", 
     "0.9527699837492861`", ",", "0.8649835057045345`", ",", 
     "0.8649835057045345`", ",", "0.894086568433468`", ",", 
     "0.894086568433468`", ",", "0.894086568433468`", ",", 
     "0.894086568433468`", ",", "0.894086568433468`", ",", 
     "0.894086568433468`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.8913897488725255`", ",", 
     "0.8913897488725255`", ",", "0.9014560500229499`", ",", 
     "0.8937233629324097`", ",", "0.8937233629324097`", ",", 
     "0.8937233629324097`", ",", "0.8937233629324097`", ",", 
     "0.8937233629324097`", ",", "0.8937233629324097`", ",", 
     "0.8937233629324097`", ",", "0.9069104544342836`", ",", 
     "0.9069104544342836`", ",", "0.9069104544342836`", ",", 
     "0.9069104544342836`", ",", "0.9069104544342836`", ",", 
     "0.9069104544342836`", ",", "0.9069104544342836`", ",", 
     "0.9069104544342836`", ",", "0.9069104544342836`", ",", 
     "0.9069104544342836`", ",", "0.9069104544342836`", ",", 
     "0.9069104544342836`", ",", "0.9069104544342836`", ",", 
     "0.9069104544342836`", ",", "0.9069104544342836`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9029207773041406`", ",", "0.9029207773041406`", ",", 
     "0.9291829911290522`", ",", "0.9291829911290522`", ",", 
     "0.9013104460566135`", ",", "0.9013104460566135`", ",", 
     "0.9013104460566135`", ",", "0.9387855159925974`", ",", 
     "0.9387855159925974`", ",", "0.8831346372165112`", ",", 
     "0.8831346372165112`", ",", "0.8831346372165112`", ",", 
     "0.929445167305133`", ",", "0.929445167305133`", ",", 
     "0.9413487060721443`", ",", "0.9413487060721443`", ",", 
     "0.9413487060721443`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.8740365993439463`", ",", "0.8740365993439463`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.9593002626903377`", ",", "0.9593002626903377`", ",", 
     "0.8725347254141471`", ",", "0.8725347254141471`", ",", 
     "0.8725347254141471`", ",", "0.8725347254141471`", ",", 
     "0.8725347254141471`", ",", "0.8725347254141471`", ",", 
     "0.8725347254141471`", ",", "0.9079845766331167`", ",", 
     "0.9079845766331167`", ",", "0.9079845766331167`", ",", 
     "0.9079845766331167`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8758468002539007`", ",", 
     "0.8758468002539007`", ",", "0.8779739712464325`", ",", 
     "0.8779739712464325`", ",", "0.8779739712464325`", ",", 
     "0.8779739712464325`", ",", "0.8779739712464325`", ",", 
     "0.9493574244463434`", ",", "0.9199100391892682`", ",", 
     "0.9199100391892682`", ",", "0.9199100391892682`", ",", 
     "0.9199100391892682`", ",", "0.9332004279662124`", ",", 
     "0.9332004279662124`", ",", "0.9332004279662124`", ",", 
     "0.9332004279662124`", ",", "0.9332004279662124`", ",", 
     "0.8950834826483633`", ",", "0.8950834826483633`", ",", 
     "0.8950834826483633`", ",", "0.8950834826483633`", ",", 
     "0.8950834826483633`", ",", "0.9572137014363467`", ",", 
     "0.9572137014363467`", ",", "0.9572137014363467`", ",", 
     "0.9572137014363467`", ",", "0.9572137014363467`", ",", 
     "0.9572137014363467`", ",", "0.9572137014363467`", ",", 
     "0.9572137014363467`", ",", "0.9572137014363467`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9157273213426294`", ",", 
     "0.9157273213426294`", ",", "0.9165619778413134`", ",", 
     "0.9165619778413134`", ",", "0.9165619778413134`", ",", 
     "0.9165619778413134`", ",", "0.9165619778413134`", ",", 
     "0.8963506971556346`", ",", "0.8963506971556346`", ",", 
     "0.8963506971556346`", ",", "0.8963506971556346`", ",", 
     "0.8963506971556346`", ",", "0.8963506971556346`", ",", 
     "0.9058353796547991`", ",", "0.9058353796547991`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.8985921370412977`", ",", 
     "0.8985921370412977`", ",", "0.9261402834517812`", ",", 
     "0.9261402834517812`", ",", "0.9261402834517812`", ",", 
     "0.9261402834517812`", ",", "0.9261402834517812`", ",", 
     "0.9261402834517812`", ",", "0.9261402834517812`", ",", 
     "0.9261402834517812`", ",", "0.9261402834517812`", ",", 
     "0.9261402834517812`", ",", "0.9261402834517812`", ",", 
     "0.8875290183107772`", ",", "0.8875290183107772`", ",", 
     "0.8875290183107772`", ",", "0.8875290183107772`", ",", 
     "0.8875290183107772`", ",", "0.8875290183107772`", ",", 
     "0.8875290183107772`", ",", "0.8875290183107772`", ",", 
     "0.9243877081958436`", ",", "0.9243877081958436`", ",", 
     "0.9243877081958436`", ",", "0.9243877081958436`", ",", 
     "0.9243877081958436`", ",", "0.9243877081958436`", ",", 
     "0.8653411050961937`", ",", "0.8653411050961937`", ",", 
     "0.8653411050961937`", ",", "0.8653411050961937`", ",", 
     "0.8653411050961937`", ",", "0.8653411050961937`", ",", 
     "0.8653411050961937`", ",", "0.8653411050961937`", ",", 
     "0.8653411050961937`", ",", "0.8635231446796243`", ",", 
     "0.8635231446796243`", ",", "0.8635231446796243`", ",", 
     "0.8635231446796243`", ",", "0.8635231446796243`", ",", 
     "0.8635231446796243`", ",", "0.8635231446796243`", ",", 
     "0.8635231446796243`", ",", "0.8635231446796243`", ",", 
     "0.9546401125384785`", ",", "0.9546401125384785`", ",", 
     "0.9546401125384785`", ",", "0.9297929196920464`", ",", 
     "0.9297929196920464`", ",", "0.9297929196920464`", ",", 
     "0.9297929196920464`", ",", "0.9297929196920464`", ",", 
     "0.9297929196920464`", ",", "0.9297929196920464`", ",", 
     "0.9297929196920464`", ",", "0.9297929196920464`", ",", 
     "0.9297929196920464`", ",", "0.9297929196920464`", ",", 
     "0.9297929196920464`", ",", "0.9297929196920464`", ",", 
     "0.9297929196920464`", ",", "0.9297929196920464`", ",", 
     "0.920217970461895`", ",", "0.920217970461895`", ",", 
     "0.920217970461895`", ",", "0.9232189755428627`", ",", 
     "0.9232189755428627`", ",", "0.9232189755428627`", ",", 
     "0.9232189755428627`", ",", "0.9232189755428627`", ",", 
     "0.8706723117443744`", ",", "0.8706723117443744`", ",", 
     "0.8706723117443744`", ",", "0.8706723117443744`", ",", 
     "0.8706723117443744`", ",", "0.8706723117443744`", ",", 
     "0.8706723117443744`", ",", "0.8706723117443744`", ",", 
     "0.8706723117443744`", ",", "0.8706723117443744`", ",", 
     "0.8706723117443744`", ",", "0.8706723117443744`", ",", 
     "0.8706723117443744`", ",", "0.8706723117443744`", ",", 
     "0.8706723117443744`", ",", "0.8685042698463589`", ",", 
     "0.8612045314351772`", ",", "0.8755568891360568`", ",", 
     "0.8755568891360568`", ",", "0.8755568891360568`", ",", 
     "0.8755568891360568`", ",", "0.8755568891360568`", ",", 
     "0.8755568891360568`", ",", "0.8755568891360568`", ",", 
     "0.8755568891360568`", ",", "0.8755568891360568`", ",", 
     "0.869479410833272`", ",", "0.869479410833272`", ",", 
     "0.8889559821359431`", ",", "0.8889559821359431`", ",", 
     "0.884578116539956`", ",", "0.884578116539956`", ",", 
     "0.884578116539956`", ",", "0.884578116539956`", ",", 
     "0.884578116539956`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9263545246234768`", ",", 
     "0.9263545246234768`", ",", "0.9249696601442523`", ",", 
     "0.9249696601442523`", ",", "0.9249696601442523`", ",", 
     "0.9249696601442523`", ",", "0.9249696601442523`", ",", 
     "0.9249696601442523`", ",", "0.9249696601442523`", ",", 
     "0.9249696601442523`", ",", "0.9249696601442523`", ",", 
     "0.9249696601442523`", ",", "0.9249696601442523`", ",", 
     "0.9083007647071849`", ",", "0.9083007647071849`", ",", 
     "0.9083007647071849`", ",", "0.9083007647071849`", ",", 
     "0.9083007647071849`", ",", "0.9083007647071849`", ",", 
     "0.9170761551421237`", ",", "0.9170761551421237`", ",", 
     "0.9170761551421237`", ",", "0.9173881180885453`", ",", 
     "0.9173881180885453`", ",", "0.9173881180885453`", ",", 
     "0.9173881180885453`", ",", "0.9173881180885453`", ",", 
     "0.8812070402181127`", ",", "0.8834630542687271`", ",", 
     "0.8834630542687271`", ",", "0.8834630542687271`", ",", 
     "0.8527552042820242`", ",", "0.8527552042820242`", ",", 
     "0.8527552042820242`", ",", "0.8527552042820242`", ",", 
     "0.8527552042820242`", ",", "0.8527552042820242`", ",", 
     "0.8527552042820242`", ",", "0.8527552042820242`", ",", 
     "0.8527552042820242`", ",", "0.8527552042820242`", ",", 
     "0.925492958169932`", ",", "0.925492958169932`", ",", 
     "0.925492958169932`", ",", "0.925492958169932`", ",", 
     "0.9234944396167517`", ",", "0.9169216578746004`", ",", 
     "0.9169216578746004`", ",", "0.9169216578746004`", ",", 
     "0.9169216578746004`", ",", "0.9169216578746004`", ",", 
     "0.9169216578746004`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.8806345953612241`", ",", "0.8806345953612241`", ",", 
     "0.9098285133663737`", ",", "0.9098285133663737`", ",", 
     "0.9098285133663737`", ",", "0.9098285133663737`", ",", 
     "0.9098285133663737`", ",", "0.9098285133663737`", ",", 
     "0.9039449236257189`", ",", "0.9039449236257189`", ",", 
     "0.9039449236257189`", ",", "0.9039449236257189`", ",", 
     "0.8791963977025913`", ",", "0.8791963977025913`", ",", 
     "0.8791963977025913`", ",", "0.8692237717439226`", ",", 
     "0.8692237717439226`", ",", "0.8692237717439226`", ",", 
     "0.8692237717439226`", ",", "0.8692237717439226`", ",", 
     "0.8692237717439226`", ",", "0.8692237717439226`", ",", 
     "0.8692237717439226`", ",", "0.8692237717439226`", ",", 
     "0.8692237717439226`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8846698038732707`", ",", 
     "0.8846698038732707`", ",", "0.8254593559667782`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.7965972770934042`", ",", 
     "0.7965972770934042`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8240548401941844`", ",", 
     "0.8240548401941844`", ",", "0.8671342852949664`", ",", 
     "0.8671342852949664`", ",", "0.8671342852949664`", ",", 
     "0.8671342852949664`", ",", "0.8671342852949664`", ",", 
     "0.8671342852949664`", ",", "0.8671342852949664`", ",", 
     "0.8671342852949664`", ",", "0.8671342852949664`", ",", 
     "0.8671342852949664`", ",", "0.8671342852949664`", ",", 
     "0.8712779225279385`", ",", "0.8712779225279385`", ",", 
     "0.8712779225279385`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8163138742149414`", ",", 
     "0.8163138742149414`", ",", "0.8883151961597743`", ",", 
     "0.8883151961597743`", ",", "0.8883151961597743`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8601156686707091`", ",", "0.8601156686707091`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8269739516059893`", ",", "0.8269739516059893`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8515161159008813`", ",", 
     "0.8515161159008813`", ",", "0.8364353648699847`", ",", 
     "0.8364353648699847`", ",", "0.8364353648699847`", ",", 
     "0.845528218052799`", ",", "0.845528218052799`", ",", 
     "0.845528218052799`", ",", "0.845528218052799`", ",", 
     "0.845528218052799`", ",", "0.845528218052799`", ",", 
     "0.8301405785429401`", ",", "0.8301405785429401`", ",", 
     "0.8301405785429401`", ",", "0.8301405785429401`", ",", 
     "0.8301405785429401`", ",", "0.8301405785429401`", ",", 
     "0.8301405785429401`", ",", "0.8301405785429401`", ",", 
     "0.8301405785429401`", ",", "0.8295151678873869`", ",", 
     "0.8295151678873869`", ",", "0.8295151678873869`", ",", 
     "0.8295151678873869`", ",", "0.8295151678873869`", ",", 
     "0.8647036532889791`", ",", "0.8647036532889791`", ",", 
     "0.8647036532889791`", ",", "0.8647036532889791`", ",", 
     "0.8647036532889791`", ",", "0.8647036532889791`", ",", 
     "0.8647036532889791`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8495088615281364`", ",", "0.8495088615281364`", ",", 
     "0.8561155398218506`", ",", "0.8561155398218506`", ",", 
     "0.8561155398218506`", ",", "0.8561155398218506`", ",", 
     "0.8561155398218506`", ",", "0.8561155398218506`", ",", 
     "0.8561155398218506`", ",", "0.8311207603803215`", ",", 
     "0.8311207603803215`", ",", "0.8311207603803215`", ",", 
     "0.8311207603803215`", ",", "0.8311207603803215`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8529511904950671`", ",", 
     "0.8529511904950671`", ",", "0.8545114610718315`", ",", 
     "0.8545114610718315`", ",", "0.8545114610718315`", ",", 
     "0.8545114610718315`", ",", "0.8545114610718315`", ",", 
     "0.8545114610718315`", ",", "0.8545114610718315`", ",", 
     "0.8545114610718315`", ",", "0.8545114610718315`", ",", 
     "0.8545114610718315`", ",", "0.8545114610718315`", ",", 
     "0.8574884693351057`", ",", "0.9013888999956275`", ",", 
     "0.9013888999956275`", ",", "0.9013888999956275`", ",", 
     "0.8803641982192785`", ",", "0.8803641982192785`", ",", 
     "0.878991146380471`", ",", "0.878991146380471`", ",", 
     "0.878991146380471`", ",", "0.878991146380471`", ",", 
     "0.878991146380471`", ",", "0.878991146380471`", ",", 
     "0.878991146380471`", ",", "0.878991146380471`", ",", 
     "0.878991146380471`", ",", "0.878991146380471`", ",", 
     "0.8506941823981342`", ",", "0.8506941823981342`", ",", 
     "0.8506941823981342`", ",", "0.8506941823981342`", ",", 
     "0.8506941823981342`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8420691914513032`", ",", 
     "0.8420691914513032`", ",", "0.8400969780238946`", ",", 
     "0.8400969780238946`", ",", "0.8400969780238946`", ",", 
     "0.8400969780238946`", ",", "0.8400969780238946`", ",", 
     "0.8400969780238946`", ",", "0.8400969780238946`", ",", 
     "0.8400969780238946`", ",", "0.8400969780238946`", ",", 
     "0.8400969780238946`", ",", "0.8526244459056259`", ",", 
     "0.8526244459056259`", ",", "0.8526244459056259`", ",", 
     "0.8526244459056259`", ",", "0.8526244459056259`", ",", 
     "0.8531189676599351`", ",", "0.8531189676599351`", ",", 
     "0.843268664978149`", ",", "0.843268664978149`", ",", 
     "0.843268664978149`", ",", "0.843268664978149`", ",", 
     "0.843268664978149`", ",", "0.843268664978149`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8574741447852378`", ",", 
     "0.8574741447852378`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8310200679091289`", ",", "0.8310200679091289`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8705948982527455`", ",", "0.8705948982527455`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8292998672033015`", ",", 
     "0.8292998672033015`", ",", "0.8616109061980229`", ",", 
     "0.8616109061980229`", ",", "0.8775517420409689`", ",", 
     "0.8775517420409689`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8592516728793673`", ",", "0.8592516728793673`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8391225317604959`", ",", "0.8391225317604959`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8550994721363756`", ",", 
     "0.8550994721363756`", ",", "0.8525797394245619`", ",", 
     "0.8525797394245619`", ",", "0.8525797394245619`", ",", 
     "0.8525797394245619`", ",", "0.8525797394245619`", ",", 
     "0.8525797394245619`", ",", "0.8525797394245619`", ",", 
     "0.8525797394245619`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8580901870128853`", ",", "0.8580901870128853`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8378128780194527`", ",", "0.8378128780194527`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8610015020795446`", ",", 
     "0.8610015020795446`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.8560707316947432`", ",", "0.8560707316947432`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.866593085085629`", ",", "0.866593085085629`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.8389975326081408`", ",", "0.8389975326081408`", ",", 
     "0.838710564215706`", ",", "0.838710564215706`", ",", 
     "0.838710564215706`", ",", "0.838710564215706`", ",", 
     "0.838710564215706`", ",", "0.838710564215706`", ",", 
     "0.838710564215706`", ",", "0.838710564215706`", ",", 
     "0.838710564215706`", ",", "0.838710564215706`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8636535513820945`", ",", 
     "0.8636535513820945`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8593868834015707`", ",", "0.8593868834015707`", ",", 
     "0.8471097785267896`", ",", "0.8471097785267896`", ",", 
     "0.8471097785267896`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8410268191245484`", ",", 
     "0.8410268191245484`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8522939189340631`", ",", "0.8522939189340631`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8388146517024538`", ",", 
     "0.8388146517024538`", ",", "0.8545486313908245`", ",", 
     "0.8545486313908245`", ",", "0.8545486313908245`", ",", 
     "0.8545486313908245`", ",", "0.8545486313908245`", ",", 
     "0.8545486313908245`", ",", "0.8545486313908245`", ",", 
     "0.8545486313908245`", ",", "0.8545486313908245`", ",", 
     "0.8545486313908245`", ",", "0.8545486313908245`", ",", 
     "0.8435428358171204`", ",", "0.8435428358171204`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.837189245911445`", ",", "0.837189245911445`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8346733797518844`", ",", 
     "0.8346733797518844`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8451112921403602`", ",", "0.8451112921403602`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8431372683757726`", ",", 
     "0.8431372683757726`", ",", "0.8412640141925583`", ",", 
     "0.8412640141925583`", ",", "0.8412640141925583`", ",", 
     "0.8412640141925583`", ",", "0.8412640141925583`", ",", 
     "0.8412640141925583`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8380902304583567`", ",", "0.8380902304583567`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8504989175669979`", ",", 
     "0.8504989175669979`", ",", "0.8615010939862904`", ",", 
     "0.8615010939862904`", ",", "0.8615010939862904`", ",", 
     "0.8615010939862904`", ",", "0.8615010939862904`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8531736847881674`", ",", 
     "0.8531736847881674`", ",", "0.8559261456852907`", ",", 
     "0.8559261456852907`", ",", "0.8559261456852907`", ",", 
     "0.8559261456852907`", ",", "0.8559261456852907`", ",", 
     "0.8559261456852907`", ",", "0.8559261456852907`", ",", 
     "0.8559261456852907`", ",", "0.8559261456852907`", ",", 
     "0.8559261456852907`", ",", "0.8559261456852907`", ",", 
     "0.8559261456852907`", ",", "0.8559261456852907`", ",", 
     "0.8559261456852907`", ",", "0.8559261456852907`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8506029820006916`", ",", "0.8506029820006916`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8501968945382179`", ",", "0.8501968945382179`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8454811835682167`", ",", "0.8454811835682167`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.8515031609244323`", ",", 
     "0.8515031609244323`", ",", "0.849505313180239`", ",", 
     "0.849505313180239`", ",", "0.849505313180239`", ",", 
     "0.849505313180239`", ",", "0.849505313180239`", ",", 
     "0.849505313180239`", ",", "0.849505313180239`", ",", 
     "0.849505313180239`", ",", "0.849505313180239`", ",", 
     "0.849505313180239`", ",", "0.849505313180239`", ",", 
     "0.849505313180239`", ",", "0.849505313180239`", ",", 
     "0.8523382098071869`", ",", "0.8523382098071869`", ",", 
     "0.8523382098071869`", ",", "0.8523382098071869`", ",", 
     "0.8523382098071869`", ",", "0.8523382098071869`", ",", 
     "0.8523382098071869`", ",", "0.8523382098071869`", ",", 
     "0.8523382098071869`", ",", "0.8523382098071869`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.856704318589408`", ",", 
     "0.856704318589408`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.848305556817111`", ",", "0.848305556817111`", ",", 
     "0.8808313598459755`", ",", "0.8808313598459755`", ",", 
     "0.8808313598459755`", ",", "0.8808313598459755`", ",", 
     "0.8808313598459755`", ",", "0.8808313598459755`", ",", 
     "0.8808313598459755`", ",", "0.8808313598459755`", ",", 
     "0.8808313598459755`", ",", "0.8808313598459755`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8517185826341401`", ",", "0.8517185826341401`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.8580317558292792`", ",", "0.8580317558292792`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.860222474332145`", ",", 
     "0.860222474332145`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8586504240086652`", ",", "0.8586504240086652`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8720712518569748`", ",", "0.8720712518569748`", ",", 
     "0.8580773452621464`", ",", "0.8580773452621464`", ",", 
     "0.8580773452621464`", ",", "0.8580773452621464`", ",", 
     "0.8580773452621464`", ",", "0.8580773452621464`", ",", 
     "0.8580773452621464`", ",", "0.8580773452621464`", ",", 
     "0.8580773452621464`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8624942758777484`", ",", 
     "0.8624942758777484`", ",", "0.8587037399354532`", ",", 
     "0.8587037399354532`", ",", "0.8587037399354532`", ",", 
     "0.8587037399354532`", ",", "0.8587037399354532`", ",", 
     "0.8587037399354532`", ",", "0.8587037399354532`", ",", 
     "0.8587037399354532`", ",", "0.8587037399354532`", ",", 
     "0.8587037399354532`", ",", "0.8587037399354532`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8731174113709411`", ",", 
     "0.8731174113709411`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`", ",", 
     "0.8690703897178863`", ",", "0.8690703897178863`"}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.424803757474777*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"ClearAll", "[", "\"\<Global`*\>\"", "]"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"Remove", "[", "\"\<Global`*\>\"", "]"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"<<", "Combinatorica`"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"<<", "DDUtils.m"}], "\[IndentingNewLine]", 
   RowBox[{"<<", "FunctionMaker.m"}], "\[IndentingNewLine]", 
   RowBox[{"<<", "DependencyDiagrams.m"}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"RateGraph", " ", "=", " ", 
     RowBox[{"EmptyGraph", "[", 
      RowBox[{"14", ",", 
       RowBox[{"Type", "\[Rule]", "Directed"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"RateGraph", " ", "=", " ", 
     RowBox[{"SetVertexLabels", "[", 
      RowBox[{"RateGraph", ",", " ", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"IndexN", "[", "i", "]"}], ",", " ", 
         RowBox[{"IndexN", "[", "j", "]"}], ",", " ", 
         RowBox[{"ConstraintN", "[", "PlusOne", "]"}], ",", " ", 
         RowBox[{"ConstantN", "[", "Aa", "]"}], ",", 
         RowBox[{"ConstantN", "[", "Bb", "]"}], ",", " ", 
         RowBox[{"ConstantN", "[", "Cc", "]"}], ",", " ", 
         RowBox[{"VariableN", "[", "k1", "]"}], ",", 
         RowBox[{"VariableN", "[", "k2", "]"}], ",", 
         RowBox[{"ConstantN", "[", "Aa", "]"}], ",", 
         RowBox[{"ConstantN", "[", "Bb", "]"}], ",", " ", 
         RowBox[{"ConstantN", "[", "Cc", "]"}], ",", 
         RowBox[{"FunctionN", "[", "p1", "]"}], ",", 
         RowBox[{"FunctionN", "[", "p2", "]"}], ",", 
         RowBox[{"FunctionN", "[", "p3", "]"}]}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"edges", " ", "=", " ", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"3", ",", "1"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3", ",", "2"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "4"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "5"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "6"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2", ",", "9"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2", ",", "10"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2", ",", "11"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12", ",", "4"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13", ",", "4"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14", ",", "4"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12", ",", "5"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13", ",", "5"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14", ",", "5"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12", ",", "6"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13", ",", "6"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14", ",", "6"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12", ",", "7"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13", ",", "7"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14", ",", "7"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12", ",", "8"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13", ",", "8"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14", ",", "8"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12", ",", "9"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13", ",", "10"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14", ",", "11"}], "}"}]}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"edgeLabs", " ", "=", " ", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
       "IndexCons", ",", "IndexCons", ",", "IndexL", ",", "IndexL", ",", 
        "IndexL", ",", "IndexL", ",", "IndexL", ",", "IndexL", ",", "CondDep",
         ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
        "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
        "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
        "CondDep", ",", "CondDep", ",", "CondDep", ",", "CondDep", ",", 
        "CondDep"}], "}"}], "/.", 
      RowBox[{"CondDep", "\[Rule]", "FactorEdge"}]}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ee", " ", "=", " ", 
     RowBox[{"MapThread", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"{", 
         RowBox[{"#1", ",", " ", 
          RowBox[{"EdgeLabel", "\[Rule]", "#2"}], ",", " ", 
          RowBox[{"EdgeDirection", "\[Rule]", "True"}]}], "}"}], "&"}], ",", 
       RowBox[{"{", 
        RowBox[{"edges", ",", "edgeLabs"}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Set", "[", 
        RowBox[{"RateGraph", ",", 
         RowBox[{"AddEdge", "[", 
          RowBox[{"RateGraph", ",", "#1"}], "]"}]}], "]"}], "&"}], ",", 
      "ee"}], "]"}], ";"}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{"PrettyPrint", "[", "RateGraph", "]"}], "\[IndentingNewLine]", 
   RowBox[{"PrettyPrint", "[", 
    RowBox[{"MCMCDiagram", "[", "RateGraph", "]"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.418695546965351*^9, 3.418695607711953*^9}, {
  3.418699786403929*^9, 3.418699788063802*^9}}],

Cell[BoxData[
 GraphicsBox[
  TagBox[{{
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0., 0.44811723608673304`}, {1.1467732421709942`, 0.}}, 0.15],
       InsetBox[
       StyleBox[
        SubscriptBox["\[Delta]", "\[Iota]"],
        StripOnInput->False,
        FontSize->15], {0.5733866210854971, 0.22405861804336652`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0., 0.44811723608673304`}, {0.22798271668821557`, 
       1.614560624838073}}, 0.15], InsetBox[
       StyleBox[
        SubscriptBox["\[Delta]", "\[Iota]"],
        StripOnInput->False,
        FontSize->15], {0.11399135834410778`, 1.031338930462403}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.1467732421709942`, 0.}, {2.2867493367830467`, 
       0.3617083022218458}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.7167612894770206`, 0.1808541511109229}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.1467732421709942`, 0.}, {1.884384130657591, 
       0.3910143594885759}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.5155786864142926`, 0.19550717974428794`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.1467732421709942`, 0.}, {1.5696098224406267`, 
       0.7390129207863054}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.3581915323058105`, 0.3695064603931527}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0.22798271668821557`, 1.614560624838073}, {1.283307193391214,
        2.1516521871945153`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {0.7556449550397147, 1.883106406016294}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0.22798271668821557`, 1.614560624838073}, {
       0.7398082623261878, 2.2588969338913065`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {0.4838954895072017, 1.9367287793646897`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0.22798271668821557`, 1.614560624838073}, {
       0.8857201783990889, 1.189962169035119}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {0.5568514475436522, 1.402261396936596}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.2838279160335184`, 1.3273689788144747`}, {
       2.2867493367830467`, 0.3617083022218458}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.2852886264082826`, 0.8445386405181603}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.2838279160335184`, 1.3273689788144747`}, {
       1.884384130657591, 0.3910143594885759}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.0841060233455546`, 0.8591916691515253}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.2838279160335184`, 1.3273689788144747`}, {
       1.5696098224406267`, 0.7390129207863054}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.9267188692370727`, 1.0331909498003902`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.2838279160335184`, 1.3273689788144747`}, {
       3.013699078714384, 1.2546158831537952`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.648763497373951, 1.290992430984135}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.2838279160335184`, 1.3273689788144747`}, {
       2.7144513775585106`, 1.8777810694175887`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.4991396467960145`, 1.6025750241160317`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.2838279160335184`, 1.3273689788144747`}, {
       1.283307193391214, 2.1516521871945153`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.7835675547123662`, 1.739510583004495}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.9064196113963912`, 1.4241284414270394`}, {
       2.2867493367830467`, 0.3617083022218458}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.096584474089719, 0.8929183718244427}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.9064196113963912`, 1.4241284414270394`}, {
       1.884384130657591, 0.3910143594885759}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.895401871026991, 0.9075714004578077}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.9064196113963912`, 1.4241284414270394`}, {
       1.5696098224406267`, 0.7390129207863054}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.738014716918509, 1.0815706811066725`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.9064196113963912`, 1.4241284414270394`}, {
       3.013699078714384, 1.2546158831537952`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.4600593450553876`, 1.3393721622904173`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.9064196113963912`, 1.4241284414270394`}, {
       2.7144513775585106`, 1.8777810694175887`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.310435494477451, 1.650954755422314}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.9064196113963912`, 1.4241284414270394`}, {
       0.7398082623261878, 2.2588969338913065`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.3231139368612894`, 1.841512687659173}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.214688787649467, 0.9469977291601384}, {2.2867493367830467`,
        0.3617083022218458}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.250719062216257, 0.654353015690992}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.214688787649467, 0.9469977291601384}, {1.884384130657591, 
       0.3910143594885759}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.049536459153529, 0.6690060443243571}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.214688787649467, 0.9469977291601384}, {1.5696098224406267`,
        0.7390129207863054}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.8921493050450469`, 0.8430053249732219}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.214688787649467, 0.9469977291601384}, {3.013699078714384, 
       1.2546158831537952`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.6141939331819257`, 1.100806806156967}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.214688787649467, 0.9469977291601384}, {2.7144513775585106`,
        1.8777810694175887`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.4645700826039887`, 1.4123893992888634`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.214688787649467, 0.9469977291601384}, {0.8857201783990889, 
       1.189962169035119}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"f\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.550204483024278, 1.0684799490976287`}]}}, {
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{-0.0625, 0.573117236086733}, {0.0625, 0.573117236086733}, {
       0.1875, 0.44811723608673304`}, {0.0625, 
       0.32311723608673304`}, {-0.0625, 0.32311723608673304`}, {-0.1875, 
       0.44811723608673304`}}], 
      {RGBColor[1, 0, 0], InsetBox["PlusOne", {0., 0.44811723608673304`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[0, 0, 1]], 
      PolygonBox[{{1.0217732421709942`, 0.125}, {1.2717732421709942`, 
       0.125}, {1.2717732421709942`, -0.125}, {1.035662131059883, -0.125}}], 
      {RGBColor[0, 0, 1], InsetBox[
        StyleBox["i",
         StripOnInput->False,
         FontSize->13], {1.1467732421709942`, 0.}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[0, 0, 1]], 
      PolygonBox[{{0.10298271668821557`, 1.739560624838073}, {
       0.35298271668821557`, 1.739560624838073}, {0.35298271668821557`, 
       1.489560624838073}, {0.11687160557710446`, 1.489560624838073}}], 
      {RGBColor[0, 0, 1], InsetBox[
        StyleBox["j",
         StripOnInput->False,
         FontSize->13], {0.22798271668821557`, 1.614560624838073}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{2.2867493367830467`, 0.3617083022218458}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{2.2867493367830467`, 0.3617083022218458}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Aa",
         StripOnInput->False,
         FontSize->20], {2.2867493367830467`, 0.3617083022218458}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{1.884384130657591, 0.3910143594885759}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{1.884384130657591, 0.3910143594885759}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Bb",
         StripOnInput->False,
         FontSize->20], {1.884384130657591, 0.3910143594885759}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{1.5696098224406267`, 0.7390129207863054}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{1.5696098224406267`, 0.7390129207863054}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Cc",
         StripOnInput->False,
         FontSize->20], {1.5696098224406267`, 0.7390129207863054}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{1.283307193391214, 2.1516521871945153`}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{1.283307193391214, 2.1516521871945153`}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Aa",
         StripOnInput->False,
         FontSize->20], {1.283307193391214, 2.1516521871945153`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{0.7398082623261878, 2.2588969338913065`}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{0.7398082623261878, 2.2588969338913065`}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Bb",
         StripOnInput->False,
         FontSize->20], {0.7398082623261878, 2.2588969338913065`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{0.8857201783990889, 1.189962169035119}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{0.8857201783990889, 1.189962169035119}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Cc",
         StripOnInput->False,
         FontSize->20], {0.8857201783990889, 1.189962169035119}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{2.2213279160335184`, 1.4523689788144747`}, {
       2.3463279160335184`, 1.4523689788144747`}, {2.4713279160335184`, 
       1.3273689788144747`}, {2.3463279160335184`, 1.2023689788144747`}, {
       2.2213279160335184`, 1.2023689788144747`}, {2.0963279160335184`, 
       1.3273689788144747`}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["p1", {2.2838279160335184`, 1.3273689788144747`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{1.8439196113963912`, 1.5491284414270394`}, {
       1.9689196113963912`, 1.5491284414270394`}, {2.093919611396391, 
       1.4241284414270394`}, {1.9689196113963912`, 1.2991284414270394`}, {
       1.8439196113963912`, 1.2991284414270394`}, {1.7189196113963912`, 
       1.4241284414270394`}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["p2", {1.9064196113963912`, 1.4241284414270394`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{2.152188787649467, 1.0719977291601384`}, {
       2.277188787649467, 1.0719977291601384`}, {2.402188787649467, 
       0.9469977291601384}, {2.277188787649467, 0.8219977291601384}, {
       2.152188787649467, 0.8219977291601384}, {2.027188787649467, 
       0.9469977291601384}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["p3", {2.214688787649467, 0.9469977291601384}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{3.013699078714384, 1.2546158831537952`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["k1",
         StripOnInput->False,
         FontSize->20], {3.013699078714384, 1.2546158831537952`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{2.7144513775585106`, 1.8777810694175887`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["k2",
         StripOnInput->False,
         FontSize->20], {2.7144513775585106`, 1.8777810694175887`}]}}}},
   Annotation[#, 
    VertexCoordinateRules -> {{0., 0.44811723608673304`}, {
     1.1467732421709942`, 0.}, {0.22798271668821557`, 1.614560624838073}, {
     2.2867493367830467`, 0.3617083022218458}, {1.884384130657591, 
     0.3910143594885759}, {1.5696098224406267`, 0.7390129207863054}, {
     1.283307193391214, 2.1516521871945153`}, {0.7398082623261878, 
     2.2588969338913065`}, {0.8857201783990889, 1.189962169035119}, {
     2.2838279160335184`, 1.3273689788144747`}, {1.9064196113963912`, 
     1.4241284414270394`}, {2.214688787649467, 0.9469977291601384}, {
     3.013699078714384, 1.2546158831537952`}, {2.7144513775585106`, 
     1.8777810694175887`}}]& ],
  AspectRatio->Automatic,
  FrameTicks->None,
  ImageSize->{942., Automatic},
  PlotRange->All,
  PlotRangePadding->Scaled[0.1]]], "Output",
 CellChangeTimes->{3.418695849415817*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Part", "::", "\<\"pspec\"\>"}], ":", 
  " ", "\<\"Part specification \\!\\(K$2259099\\) is neither an integer nor a \
list of integers. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", \
ButtonStyle->\\\"Link\\\", ButtonFrame->None, \
ButtonData:>\\\"paclet:ref/message/General/pspec\\\", ButtonNote -> \
\\\"Part::pspec\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.418695849973609*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Part", "::", "\<\"pspec\"\>"}], ":", 
  " ", "\<\"Part specification \\!\\(jmin0 + K$2259099\\) is neither an \
integer nor a list of integers. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", \
ButtonStyle->\\\"Link\\\", ButtonFrame->None, \
ButtonData:>\\\"paclet:ref/message/General/pspec\\\", ButtonNote -> \
\\\"Part::pspec\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.41869585011549*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Part", "::", "\<\"pspec\"\>"}], ":", 
  " ", "\<\"Part specification \\!\\(i\\) is neither an integer nor a list of \
integers. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", \
ButtonStyle->\\\"Link\\\", ButtonFrame->None, \
ButtonData:>\\\"paclet:ref/message/General/pspec\\\", ButtonNote -> \
\\\"Part::pspec\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.418695850218201*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"General", "::", "\<\"stop\"\>"}], ":", 
  " ", "\<\"Further output of \\!\\(Part :: \\\"pspec\\\"\\) will be \
suppressed during this calculation. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\
\", ButtonStyle->\\\"Link\\\", ButtonFrame->None, \
ButtonData:>\\\"paclet:ref/message/General/stop\\\", ButtonNote -> \
\\\"General::stop\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.418695850334924*^9}],

Cell[BoxData[
 GraphicsBox[
  TagBox[{{
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.1160536442387603`, 0.8822639597267223}, {
       1.7253912992562772`, 1.3768174441693222`}}, 0.15], InsetBox[
       StyleBox[
        SubscriptBox["\[Delta]", "\[Iota]"],
        StripOnInput->False,
        FontSize->15], {1.4207224717475189`, 1.1295407019480224`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.1160536442387603`, 0.8822639597267223}, {
       1.7229210043250212`, 1.6889235869328707`}}, 0.15], InsetBox[
       StyleBox[
        SubscriptBox["\[Delta]", "\[Iota]"],
        StripOnInput->False,
        FontSize->15], {1.4194873242818908`, 1.2855937733297966`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7253912992562772`, 1.3768174441693222`}, {
       2.714630850033555, 1.3239335905284744`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {2.2200110746449164`, 1.3503755173488983`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7253912992562772`, 1.3768174441693222`}, {
       1.8731604366531194`, 2.624539300717116}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.7992758679546983`, 2.000678372443219}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       2.6627726677412955`, 1.7374948756017623`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {2.1928468360331586`, 1.7132092312673164`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       3.104234614840145, 1.6064058370482113`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {2.4135778095825833`, 1.647664711990541}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       2.897035011130603, 1.1626941407066917`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {2.3099780077278123`, 1.4258088638197812`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       2.019627585309652, 0.38524918995974233`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.8712742948173366`, 1.0370863884463066`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       0.724583375588318, 2.188613130931195}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.2237521899566697`, 1.9387683589320328`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       1.6072209647894065`, 2.984867008647269}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.6650709845572138`, 2.33689529779007}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       1.2421836734587448`, 2.6235687475247027`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.482552338891883, 2.1562461672287867`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.7229210043250212`, 1.6889235869328707`}, {
       0.4224304400098815, 1.8149567421333501`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {1.0726757221674514`, 1.7519401645331105`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.714630850033555, 1.3239335905284744`}, {
       3.5477024588337045`, 0.7580607283147418}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.13116665443363, 1.040997159421608}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.714630850033555, 1.3239335905284744`}, {
       3.0943261394978654`, 2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.9044784947657103`, 1.94349395658895}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.714630850033555, 1.3239335905284744`}, {2.551774196023388, 
       0.6568651632626634}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.633202523028472, 0.9903993768955689}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.6627726677412955`, 1.7374948756017623`}, {
       2.7067358252745586`, 2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.684754246507927, 2.3485290777769077`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.897035011130603, 1.1626941407066917`}, {2.774870336394663, 
       0.}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.835952673762633, 0.5813470703533459}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.897035011130603, 1.1626941407066917`}, {3.574334477996489, 
       1.2819222829874064`}}, 0.15], InsetBox[
       StyleBox["\[Gamma]",
        StripOnInput->False,
        FontSize->15], {3.235684744563546, 1.2223082118470492`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.774870336394663, 0.}, {2.019627585309652, 
       0.38524918995974233`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.3972489608521577`, 0.19262459497987117`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.019627585309652, 0.38524918995974233`}, {2.551774196023388,
        0.6568651632626634}}, 0.15], InsetBox[
       StyleBox["\[Gamma]",
        StripOnInput->False,
        FontSize->15], {2.2857008906665204`, 0.5210571766112029}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.5477024588337045`, 0.7580607283147418}, {3.104234614840145,
        1.6064058370482113`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.3259685368369247`, 1.1822332826814765`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.104234614840145, 1.6064058370482113`}, {
       3.0943261394978654`, 2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.099280377169005, 2.0847300798488186`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.104234614840145, 1.6064058370482113`}, {3.574334477996489, 
       1.2819222829874064`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.339284546418317, 1.4441640600178087`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.0943261394978654`, 2.5630543226494256`}, {
       2.897035011130603, 1.1626941407066917`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.995680575314234, 1.8628742316780587`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.574334477996489, 1.2819222829874064`}, {
       2.6627726677412955`, 1.7374948756017623`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.1185535728688922`, 1.5097085792945844`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.551774196023388, 0.6568651632626634}, {2.6627726677412955`,
        1.7374948756017623`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.6072734318823416`, 1.1971800194322129`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.8731604366531194`, 2.624539300717116}, {1.452302209522084, 
       3.652380055591593}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.6627313230876017`, 3.1384596781543546`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.8731604366531194`, 2.624539300717116}, {
       2.7067358252745586`, 2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.2899481309638388`, 2.7920512903345847`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.8731604366531194`, 2.624539300717116}, {0.637568286084029, 
       2.54118007171051}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.2553643613685743`, 2.582859686213813}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.8731604366531194`, 2.624539300717116}, {
       3.0943261394978654`, 2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.4837432880754924`, 2.593796811683271}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.2421836734587448`, 2.6235687475247027`}, {0., 
       2.42074036734709}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {0.6210918367293724, 2.522154557435896}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.2421836734587448`, 2.6235687475247027`}, {
       0.6893093045457914, 3.0892629333691035`}}, 0.15], InsetBox[
       StyleBox["\[Gamma]",
        StripOnInput->False,
        FontSize->15], {0.9657464890022681, 2.8564158404469033`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0., 2.42074036734709}, {0.4224304400098815, 
       1.8149567421333501`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {0.21121522000494075`, 2.11784855474022}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0.4224304400098815, 1.8149567421333501`}, {0.637568286084029,
        2.54118007171051}}, 0.15], InsetBox[
       StyleBox["\[Gamma]",
        StripOnInput->False,
        FontSize->15], {0.5299993630469553, 2.17806840692193}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.452302209522084, 3.652380055591593}, {1.6072209647894065`, 
       2.984867008647269}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.5297615871557453`, 3.318623532119431}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.6072209647894065`, 2.984867008647269}, {
       2.7067358252745586`, 2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.1569783950319827`, 2.972215144299661}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{1.6072209647894065`, 2.984867008647269}, {0.6893093045457914,
        3.0892629333691035`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.148265134667599, 3.0370649710081863`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.7067358252745586`, 2.959563279952053}, {
       1.2421836734587448`, 2.6235687475247027`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {1.9744597493666518`, 2.791566013738378}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0.6893093045457914, 3.0892629333691035`}, {0.724583375588318,
        2.188613130931195}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {0.7069463400670547, 2.638938032150149}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{0.637568286084029, 2.54118007171051}, {0.724583375588318, 
       2.188613130931195}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {0.6810758308361735, 2.3648966013208526`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.8266350083636596`, 3.0374417598158674`}, {
       3.7398603033362674`, 2.7092866501810953`}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {3.7832476558499635`, 2.8733642049984813`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.8266350083636596`, 3.0374417598158674`}, {
       3.0943261394978654`, 2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        OverscriptBox["\<\"\[Iota]\"\>", "\<\"~\"\>"],
        StripOnInput->False,
        FontSize->15], {3.4604805739307625`, 2.8002480412326465`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.8266350083636596`, 3.0374417598158674`}, {
       2.7067358252745586`, 2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        OverscriptBox["\<\"\[Iota]\"\>", "\<\"~\"\>"],
        StripOnInput->False,
        FontSize->15], {3.266685416819109, 2.99850251988396}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.7398603033362674`, 2.7092866501810953`}, {
       3.0943261394978654`, 2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.4170932214170664`, 2.6361704864152604`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.7398603033362674`, 2.7092866501810953`}, {
       2.7067358252745586`, 2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.223298064305413, 2.834424965066574}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.13422192816675, 3.6329837981887376`}, {2.810485826147222, 
       3.605480555175374}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {2.972353877156986, 3.619232176682056}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.13422192816675, 3.6329837981887376`}, {3.0943261394978654`,
        2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        OverscriptBox["\<\"\[Iota]\"\>", "\<\"~\"\>"],
        StripOnInput->False,
        FontSize->15], {3.1142740338323076`, 3.0980190604190816`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.13422192816675, 3.6329837981887376`}, {2.7067358252745586`,
        2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        OverscriptBox["\<\"\[Iota]\"\>", "\<\"~\"\>"],
        StripOnInput->False,
        FontSize->15], {2.9204788767206544`, 3.296273539070395}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.810485826147222, 3.605480555175374}, {3.0943261394978654`, 
       2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.952405982822544, 3.0842674389124}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{2.810485826147222, 3.605480555175374}, {2.7067358252745586`, 
       2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {2.7586108257108903`, 3.2825219175637135`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.5082465698553347`, 3.248482892630766}, {3.516851872886982, 
       3.556437207256523}}, 0.15], InsetBox[
       StyleBox["\[Iota]",
        StripOnInput->False,
        FontSize->15], {3.5125492213711587`, 3.4024600499436444`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.5082465698553347`, 3.248482892630766}, {
       3.0943261394978654`, 2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        OverscriptBox["\<\"\[Iota]\"\>", "\<\"~\"\>"],
        StripOnInput->False,
        FontSize->15], {3.3012863546766003`, 2.9057686076400957`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.5082465698553347`, 3.248482892630766}, {
       2.7067358252745586`, 2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        OverscriptBox["\<\"\[Iota]\"\>", "\<\"~\"\>"],
        StripOnInput->False,
        FontSize->15], {3.1074911975649466`, 3.1040230862914093`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.516851872886982, 3.556437207256523}, {3.0943261394978654`, 
       2.5630543226494256`}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.305589006192424, 3.0597457649529742`}]}, 
     {GrayLevel[0], Arrowheads[0.015], 
      ArrowBox[{{3.516851872886982, 3.556437207256523}, {2.7067358252745586`, 
       2.959563279952053}}, 0.15], InsetBox[
       StyleBox[
        StyleBox["\<\"d\"\>",
         StripOnInput->False,
         FontSlant->Italic],
        StripOnInput->False,
        FontSize->15], {3.1117938490807706`, 3.258000243604288}]}}, {
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{1.0535536442387603`, 1.0072639597267223`}, {
       1.1785536442387603`, 1.0072639597267223`}, {1.3035536442387603`, 
       0.8822639597267223}, {1.1785536442387603`, 0.7572639597267223}, {
       1.0535536442387603`, 0.7572639597267223}, {0.9285536442387603, 
       0.8822639597267223}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["PlusOne", {1.1160536442387603`, 0.8822639597267223}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[0, 0, 1]], 
      PolygonBox[{{1.6003912992562772`, 1.5018174441693222`}, {
       1.8503912992562772`, 1.5018174441693222`}, {1.8503912992562772`, 
       1.2518174441693222`}, {1.614280188145166, 1.2518174441693222`}}], 
      {RGBColor[0, 0, 1], InsetBox[
        StyleBox["t0",
         StripOnInput->False,
         FontSize->13], {1.7253912992562772`, 1.3768174441693222`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[0, 0, 1]], 
      PolygonBox[{{1.5979210043250212`, 1.8139235869328707`}, {
       1.8479210043250212`, 1.8139235869328707`}, {1.8479210043250212`, 
       1.5639235869328707`}, {1.61180989321391, 1.5639235869328707`}}], 
      {RGBColor[0, 0, 1], InsetBox[
        StyleBox["t1",
         StripOnInput->False,
         FontSize->13], {1.7229210043250212`, 1.6889235869328707`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{2.714630850033555, 1.3239335905284744`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["k1",
         StripOnInput->False,
         FontSize->20], {2.714630850033555, 1.3239335905284744`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{2.6627726677412955`, 1.7374948756017623`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["k1",
         StripOnInput->False,
         FontSize->20], {2.6627726677412955`, 1.7374948756017623`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{2.897035011130603, 1.1626941407066917`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["acceptk10",
         StripOnInput->False,
         FontSize->20], {2.897035011130603, 1.1626941407066917`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{2.712370336394663, 0.125}, {2.837370336394663, 0.125}, {
       2.962370336394663, 0.}, {2.837370336394663, -0.125}, {
       2.712370336394663, -0.125}, {2.587370336394663, 0.}}], 
      {RGBColor[1, 0, 0], InsetBox["IndicOther", {2.774870336394663, 0.}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{2.019627585309652, 0.38524918995974233`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["nAcceptk10",
         StripOnInput->False,
         FontSize->20], {2.019627585309652, 0.38524918995974233`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{3.4852024588337045`, 0.8830607283147418}, {
       3.6102024588337045`, 0.8830607283147418}, {3.7352024588337045`, 
       0.7580607283147418}, {3.6102024588337045`, 0.6330607283147418}, {
       3.4852024588337045`, 0.6330607283147418}, {3.3602024588337045`, 
       0.7580607283147418}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["Proposalk10", {3.5477024588337045`, 0.7580607283147418}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{3.104234614840145, 1.6064058370482113`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["proposedk10",
         StripOnInput->False,
         FontSize->20], {3.104234614840145, 1.6064058370482113`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{3.0318261394978654`, 2.6880543226494256`}, {
       3.1568261394978654`, 2.6880543226494256`}, {3.2818261394978654`, 
       2.5630543226494256`}, {3.1568261394978654`, 2.4380543226494256`}, {
       3.0318261394978654`, 2.4380543226494256`}, {2.9068261394978654`, 
       2.5630543226494256`}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["Fk10", {3.0943261394978654`, 2.5630543226494256`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{3.511834477996489, 1.4069222829874064`}, {
       3.636834477996489, 1.4069222829874064`}, {3.761834477996489, 
       1.2819222829874064`}, {3.636834477996489, 1.1569222829874064`}, {
       3.511834477996489, 1.1569222829874064`}, {3.386834477996489, 
       1.2819222829874064`}}], 
      {RGBColor[1, 0, 0], 
       InsetBox[
        "DeltaCondDist", {3.574334477996489, 1.2819222829874064`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{2.489274196023388, 0.7818651632626634}, {2.614274196023388,
        0.7818651632626634}, {2.739274196023388, 0.6568651632626634}, {
       2.614274196023388, 0.5318651632626634}, {2.489274196023388, 
       0.5318651632626634}, {2.364274196023388, 0.6568651632626634}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["DeltaCondDist", {2.551774196023388, 0.6568651632626634}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{1.8731604366531194`, 2.624539300717116}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["k2",
         StripOnInput->False,
         FontSize->20], {1.8731604366531194`, 2.624539300717116}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{0.724583375588318, 2.188613130931195}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["k2",
         StripOnInput->False,
         FontSize->20], {0.724583375588318, 2.188613130931195}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{1.2421836734587448`, 2.6235687475247027`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["acceptk20",
         StripOnInput->False,
         FontSize->20], {1.2421836734587448`, 2.6235687475247027`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{-0.0625, 2.54574036734709}, {0.0625, 2.54574036734709}, {
       0.1875, 2.42074036734709}, {0.0625, 2.29574036734709}, {-0.0625, 
       2.29574036734709}, {-0.1875, 2.42074036734709}}], 
      {RGBColor[1, 0, 0], InsetBox["IndicOther", {0., 2.42074036734709}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{0.4224304400098815, 1.8149567421333501`}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["nAcceptk20",
         StripOnInput->False,
         FontSize->20], {0.4224304400098815, 1.8149567421333501`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{1.389802209522084, 3.777380055591593}, {1.514802209522084, 
       3.777380055591593}, {1.639802209522084, 3.652380055591593}, {
       1.514802209522084, 3.527380055591593}, {1.389802209522084, 
       3.527380055591593}, {1.264802209522084, 3.652380055591593}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["Proposalk20", {1.452302209522084, 3.652380055591593}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{1.6072209647894065`, 2.984867008647269}, NCache[
       Rational[1, 8], 0.125]], 
      {GrayLevel[0], InsetBox[
        StyleBox["proposedk20",
         StripOnInput->False,
         FontSize->20], {1.6072209647894065`, 2.984867008647269}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{2.6442358252745586`, 3.084563279952053}, {
       2.7692358252745586`, 3.084563279952053}, {2.8942358252745586`, 
       2.959563279952053}, {2.7692358252745586`, 2.834563279952053}, {
       2.6442358252745586`, 2.834563279952053}, {2.5192358252745586`, 
       2.959563279952053}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["Fk20", {2.7067358252745586`, 2.959563279952053}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{0.6268093045457914, 3.2142629333691035`}, {
       0.7518093045457914, 3.2142629333691035`}, {0.8768093045457914, 
       3.0892629333691035`}, {0.7518093045457914, 2.9642629333691035`}, {
       0.6268093045457914, 2.9642629333691035`}, {0.5018093045457914, 
       3.0892629333691035`}}], 
      {RGBColor[1, 0, 0], 
       InsetBox[
        "DeltaCondDist", {0.6893093045457914, 3.0892629333691035`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[1, 0, 0]], 
      PolygonBox[{{0.575068286084029, 2.66618007171051}, {0.700068286084029, 
       2.66618007171051}, {0.825068286084029, 2.54118007171051}, {
       0.700068286084029, 2.41618007171051}, {0.575068286084029, 
       2.41618007171051}, {0.45006828608402905`, 2.54118007171051}}], 
      {RGBColor[1, 0, 0], 
       InsetBox["DeltaCondDist", {0.637568286084029, 2.54118007171051}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[0, 0, 1]], 
      PolygonBox[{{3.7016350083636596`, 3.1624417598158674`}, {
       3.9516350083636596`, 3.1624417598158674`}, {3.9516350083636596`, 
       2.9124417598158674`}, {3.7155238972525484`, 2.9124417598158674`}}], 
      {RGBColor[0, 0, 1], InsetBox[
        StyleBox["i0",
         StripOnInput->False,
         FontSize->13], {3.8266350083636596`, 3.0374417598158674`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{3.7398603033362674`, 2.7092866501810953`}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{3.7398603033362674`, 2.7092866501810953`}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Aa",
         StripOnInput->False,
         FontSize->20], {3.7398603033362674`, 2.7092866501810953`}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[0, 0, 1]], 
      PolygonBox[{{3.00922192816675, 3.7579837981887376`}, {3.25922192816675, 
       3.7579837981887376`}, {3.25922192816675, 3.5079837981887376`}, {
       3.0231108170556387`, 3.5079837981887376`}}], 
      {RGBColor[0, 0, 1], InsetBox[
        StyleBox["i1",
         StripOnInput->False,
         FontSize->13], {3.13422192816675, 3.6329837981887376`}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{2.810485826147222, 3.605480555175374}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{2.810485826147222, 3.605480555175374}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Bb",
         StripOnInput->False,
         FontSize->20], {2.810485826147222, 3.605480555175374}]}}, 
     {GrayLevel[1], EdgeForm[RGBColor[0, 0, 1]], 
      PolygonBox[{{3.3832465698553347`, 3.373482892630766}, {
       3.6332465698553347`, 3.373482892630766}, {3.6332465698553347`, 
       3.123482892630766}, {3.3971354587442235`, 3.123482892630766}}], 
      {RGBColor[0, 0, 1], InsetBox[
        StyleBox["i2",
         StripOnInput->False,
         FontSize->13], {3.5082465698553347`, 3.248482892630766}]}}, 
     {GrayLevel[1], EdgeForm[GrayLevel[0]], 
      DiskBox[{3.516851872886982, 3.556437207256523}, NCache[
       Rational[1, 8], 0.125]], 
      DiskBox[{3.516851872886982, 3.556437207256523}, NCache[
       Rational[1, 9], 0.1111111111111111]], 
      {GrayLevel[0], InsetBox[
        StyleBox["Cc",
         StripOnInput->False,
         FontSize->20], {3.516851872886982, 3.556437207256523}]}}}},
   Annotation[#, 
    VertexCoordinateRules -> {{1.1160536442387603`, 0.8822639597267223}, {
     1.7253912992562772`, 1.3768174441693222`}, {1.7229210043250212`, 
     1.6889235869328707`}, {2.714630850033555, 1.3239335905284744`}, {
     2.6627726677412955`, 1.7374948756017623`}, {2.897035011130603, 
     1.1626941407066917`}, {2.774870336394663, 0.}, {2.019627585309652, 
     0.38524918995974233`}, {3.5477024588337045`, 0.7580607283147418}, {
     3.104234614840145, 1.6064058370482113`}, {3.0943261394978654`, 
     2.5630543226494256`}, {3.574334477996489, 1.2819222829874064`}, {
     2.551774196023388, 0.6568651632626634}, {1.8731604366531194`, 
     2.624539300717116}, {0.724583375588318, 2.188613130931195}, {
     1.2421836734587448`, 2.6235687475247027`}, {0., 2.42074036734709}, {
     0.4224304400098815, 1.8149567421333501`}, {1.452302209522084, 
     3.652380055591593}, {1.6072209647894065`, 2.984867008647269}, {
     2.7067358252745586`, 2.959563279952053}, {0.6893093045457914, 
     3.0892629333691035`}, {0.637568286084029, 2.54118007171051}, {
     3.8266350083636596`, 3.0374417598158674`}, {3.7398603033362674`, 
     2.7092866501810953`}, {3.13422192816675, 3.6329837981887376`}, {
     2.810485826147222, 3.605480555175374}, {3.5082465698553347`, 
     3.248482892630766}, {3.516851872886982, 3.556437207256523}}]& ],
  AspectRatio->Automatic,
  FrameTicks->None,
  ImageSize->{979., Automatic},
  PlotRange->All,
  PlotRangePadding->Scaled[0.1]]], "Output",
 CellChangeTimes->{3.418695851435957*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"l1", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1.`", ",", "0.5902249067844593`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2.`", ",", "0.4084836900611042`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3.`", ",", "0.4084836900611042`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"4.`", ",", "0.4084836900611042`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"5.`", ",", "0.4084836900611042`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"6.`", ",", "0.3001750185132055`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"7.`", ",", "0.3001750185132055`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"8.`", ",", "0.2700885523215208`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"9.`", ",", "0.2700885523215208`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"10.`", ",", "0.2700885523215208`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"11.`", ",", "0.2700885523215208`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12.`", ",", "0.1692606529094739`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13.`", ",", "0.1692606529094739`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14.`", ",", "0.1692606529094739`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"15.`", ",", "0.1692606529094739`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"16.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"17.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"18.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"19.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"20.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"21.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"22.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"23.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"24.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"25.`", ",", "0.11486186618056379`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"26.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"27.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"28.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"29.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"30.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"31.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"32.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"33.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"34.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"35.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"36.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"37.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"38.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"39.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"40.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"41.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"42.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"43.`", ",", "0.026637116568527852`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"44.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"45.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"46.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"47.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"48.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"49.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"50.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"51.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"52.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"53.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"54.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"55.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"56.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"57.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"58.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"59.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"60.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"61.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"62.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"63.`", ",", "0.05921529248777529`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"64.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"65.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"66.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"67.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"68.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"69.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"70.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"71.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"72.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"73.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"74.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"75.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"76.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"77.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"78.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"79.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"80.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"81.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"82.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"83.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"84.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"85.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"86.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"87.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"88.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"89.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"90.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"91.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"92.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"93.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"94.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"95.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"96.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"97.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"98.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"99.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"100.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"101.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"102.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"103.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"104.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"105.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"106.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"107.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"108.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"109.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"110.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"111.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"112.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"113.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"114.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"115.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"116.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"117.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"118.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"119.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"129.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"136.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"138.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"139.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"141.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"142.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"143.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"144.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"145.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"146.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"147.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"148.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"149.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"150.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"151.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"152.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"153.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"154.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"155.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"156.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"157.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"158.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"159.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"160.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"161.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"162.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"163.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"164.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"165.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"166.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"167.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"168.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"169.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"170.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"171.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"172.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"173.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"174.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"175.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"176.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"177.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"178.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"179.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"180.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"181.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"182.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"183.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"184.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"185.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"186.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"187.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"188.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"189.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"190.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"191.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"192.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"193.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"194.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"195.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"196.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"197.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"198.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"199.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"200.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"201.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"202.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"203.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"204.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"205.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"206.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"207.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"208.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"209.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"210.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"211.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"212.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"213.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"214.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"215.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"216.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"217.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"218.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"219.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"220.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"221.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"222.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"223.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"224.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"225.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"226.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"227.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"228.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"229.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"230.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"231.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"232.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"233.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"234.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"235.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"236.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"237.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"238.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"239.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"240.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"241.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"242.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"243.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"244.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"245.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"246.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"247.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"248.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"249.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"250.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"251.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"252.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"253.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"254.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"255.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"256.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"257.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"258.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"259.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"260.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"261.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"262.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"263.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"264.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"265.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"266.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"267.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"268.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"269.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"270.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"271.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"272.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"273.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"274.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"275.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"276.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"277.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"278.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"279.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"280.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"281.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"282.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"283.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"284.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"285.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"286.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"287.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"288.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"289.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"290.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"291.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"292.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"293.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"294.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"295.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"296.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"297.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"298.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"299.`", ",", "0.06248537373601094`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"300.`", ",", "0.06248537373601094`"}], "}"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"l2", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1.`", ",", "0.4674480627976655`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"2.`", ",", "0.4674480627976655`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3.`", ",", "0.5214324875458713`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"4.`", ",", "0.5214324875458713`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"5.`", ",", "0.5214324875458713`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"6.`", ",", "0.5214324875458713`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"7.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"8.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"9.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"10.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"11.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"12.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"14.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"15.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"16.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"17.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"18.`", ",", "0.6285747777111945`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"19.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"20.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"21.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"22.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"23.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"24.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"25.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"26.`", ",", "0.9210368498363979`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"27.`", ",", "0.8795362483375808`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"28.`", ",", "0.8795362483375808`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"29.`", ",", "0.7577071190983092`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"30.`", ",", "0.7577071190983092`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"31.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"32.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"33.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"34.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"35.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"36.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"37.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"38.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"39.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"40.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"41.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"42.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"43.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"44.`", ",", "0.7215210045418189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"45.`", ",", "0.6493306002081732`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"46.`", ",", "0.6493306002081732`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"47.`", ",", "0.6493306002081732`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"48.`", ",", "0.6493306002081732`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"49.`", ",", "0.7147131191125957`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"50.`", ",", "0.7147131191125957`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"51.`", ",", "0.6973632665592006`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"52.`", ",", "0.6973632665592006`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"53.`", ",", "0.7634304961828312`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"54.`", ",", "0.7634304961828312`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"55.`", ",", "0.7540428386333728`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"56.`", ",", "0.7540428386333728`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"57.`", ",", "0.7562981601421697`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"58.`", ",", "0.7562981601421697`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"59.`", ",", "0.7562981601421697`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"60.`", ",", "0.7562981601421697`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"61.`", ",", "0.7562981601421697`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"62.`", ",", "0.7562981601421697`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"63.`", ",", "0.782696210217232`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"64.`", ",", "0.782696210217232`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"65.`", ",", "0.782696210217232`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"66.`", ",", "0.782696210217232`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"67.`", ",", "0.7467992407926125`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"68.`", ",", "0.7467992407926125`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"69.`", ",", "0.7467992407926125`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"70.`", ",", "0.7467992407926125`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"71.`", ",", "0.6715091912044298`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"72.`", ",", "0.6715091912044298`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"73.`", ",", "0.6866358138643825`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"74.`", ",", "0.6866358138643825`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"75.`", ",", "0.6670220153788822`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"76.`", ",", "0.6670220153788822`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"77.`", ",", "0.5188702228517857`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"78.`", ",", "0.5188702228517857`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"79.`", ",", "0.5188702228517857`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"80.`", ",", "0.5188702228517857`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"81.`", ",", "0.5188702228517857`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"82.`", ",", "0.5188702228517857`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"83.`", ",", "0.5882563416675257`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"84.`", ",", "0.5882563416675257`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"85.`", ",", "0.6402365276923645`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"86.`", ",", "0.6402365276923645`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"87.`", ",", "0.5747904298593296`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"88.`", ",", "0.5747904298593296`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"89.`", ",", "0.6001929566717117`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"90.`", ",", "0.6001929566717117`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"91.`", ",", "0.6477184766091844`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"92.`", ",", "0.6477184766091844`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"93.`", ",", "0.6505912347254365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"94.`", ",", "0.6505912347254365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"95.`", ",", "0.46718429320204646`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"96.`", ",", "0.46718429320204646`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"97.`", ",", "0.5413671047001588`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"98.`", ",", "0.5413671047001588`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"99.`", ",", "0.6464211390163187`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"100.`", ",", "0.6464211390163187`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"101.`", ",", "0.6464211390163187`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"102.`", ",", "0.6464211390163187`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"103.`", ",", "0.6464211390163187`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"104.`", ",", "0.6464211390163187`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"105.`", ",", "0.7089297390139375`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"106.`", ",", "0.7089297390139375`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"107.`", ",", "0.7167039826954964`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"108.`", ",", "0.7167039826954964`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"109.`", ",", "0.6288382522157888`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"110.`", ",", "0.6288382522157888`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"111.`", ",", "0.6288382522157888`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"112.`", ",", "0.6288382522157888`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"113.`", ",", "0.6288382522157888`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"114.`", ",", "0.6288382522157888`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"115.`", ",", "0.6318055981210986`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"116.`", ",", "0.6318055981210986`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"117.`", ",", "0.6318055981210986`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"118.`", ",", "0.6318055981210986`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"119.`", ",", "0.7304271823055701`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"120.`", ",", "0.7304271823055701`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"121.`", ",", "0.7304271823055701`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"122.`", ",", "0.7304271823055701`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"123.`", ",", "0.7304271823055701`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"124.`", ",", "0.7304271823055701`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"125.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"126.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"127.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"128.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"129.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"130.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"131.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"132.`", ",", "0.631311216638792`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"133.`", ",", "0.6585161500704335`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"134.`", ",", "0.6585161500704335`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"135.`", ",", "0.6585161500704335`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"136.`", ",", "0.6585161500704335`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"137.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"138.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"139.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"140.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"141.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"142.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"143.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"144.`", ",", "0.7168515774867849`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"145.`", ",", "0.7652202326626563`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"146.`", ",", "0.7652202326626563`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"147.`", ",", "0.6985419532871002`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"148.`", ",", "0.6985419532871002`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"149.`", ",", "0.7060818040196851`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"150.`", ",", "0.7060818040196851`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"151.`", ",", "0.6802629834067673`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"152.`", ",", "0.6802629834067673`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"153.`", ",", "0.6802629834067673`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"154.`", ",", "0.6802629834067673`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"155.`", ",", "0.6582636076149051`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"156.`", ",", "0.6582636076149051`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"157.`", ",", "0.6121507973119436`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"158.`", ",", "0.6121507973119436`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"159.`", ",", "0.5837760693295917`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"160.`", ",", "0.5837760693295917`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"161.`", ",", "0.5837760693295917`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"162.`", ",", "0.5837760693295917`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"163.`", ",", "0.615294966433025`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"164.`", ",", "0.615294966433025`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"165.`", ",", "0.615294966433025`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"166.`", ",", "0.615294966433025`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"167.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"168.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"169.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"170.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"171.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"172.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"173.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"174.`", ",", "0.6666633870962037`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"175.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"176.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"177.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"178.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"179.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"180.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"181.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"182.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"183.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"184.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"185.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"186.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"187.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"188.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"189.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"190.`", ",", "0.6601453473109535`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"191.`", ",", "0.6875933189426752`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"192.`", ",", "0.6875933189426752`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"193.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"194.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"195.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"196.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"197.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"198.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"199.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"200.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"201.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"202.`", ",", "0.6581867374969947`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"203.`", ",", "0.6534211982676978`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"204.`", ",", "0.6534211982676978`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"205.`", ",", "0.6998961868027908`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"206.`", ",", "0.6998961868027908`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"207.`", ",", "0.6378365813780121`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"208.`", ",", "0.6378365813780121`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"209.`", ",", "0.6970984948741803`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"210.`", ",", "0.6970984948741803`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"211.`", ",", "0.6461041082292815`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"212.`", ",", "0.6461041082292815`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"213.`", ",", "0.6461041082292815`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"214.`", ",", "0.6461041082292815`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"215.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"216.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"217.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"218.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"219.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"220.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"221.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"222.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"223.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"224.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"225.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"226.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"227.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"228.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"229.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"230.`", ",", "0.6600548602970879`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"231.`", ",", "0.6918403679604189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"232.`", ",", "0.6918403679604189`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"233.`", ",", "0.6954655415265901`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"234.`", ",", "0.6954655415265901`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"235.`", ",", "0.6954655415265901`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"236.`", ",", "0.6954655415265901`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"237.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"238.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"239.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"240.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"241.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"242.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"243.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"244.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"245.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"246.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"247.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"248.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"249.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"250.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"251.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"252.`", ",", "0.6737100107365`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"253.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"254.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"255.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"256.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"257.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"258.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"259.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"260.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"261.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"262.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"263.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"264.`", ",", "0.6934212570699281`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"265.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"266.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"267.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"268.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"269.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"270.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"271.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"272.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"273.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"274.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"275.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"276.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"277.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"278.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"279.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"280.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"281.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"282.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"283.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"284.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"285.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"286.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"287.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"288.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"289.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"290.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"291.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"292.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"293.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"294.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"295.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"296.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"297.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"298.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"299.`", ",", "0.6688691085703977`"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"300.`", ",", "0.6688691085703977`"}], "}"}]}], "}"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.431453932710302*^9, 3.431453993914078*^9}, {
  3.431454030053331*^9, 3.431454035315575*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"pl", "=", 
  RowBox[{"Plot", "[", 
   RowBox[{
    RowBox[{"Sin", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "0", ",", 
      RowBox[{"2", "Pi"}]}], "}"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"Export", "[", 
  RowBox[{"\"\</tmp/foo3.pdf\>\"", ",", "pl"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"Export", "[", 
  RowBox[{"\"\</tmp/foo3.jpg\>\"", ",", "pl"}], "]"}], "\[IndentingNewLine]", 

 RowBox[{
  RowBox[{"Export", "[", 
   RowBox[{"\"\</tmp/foo3.eps\>\"", ",", "pl"}], "]"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"plot", " ", "=", " ", 
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"l1", ",", "l2"}], "}"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "1"}], "}"}]}], ",", 
     RowBox[{"ImageSize", "\[Rule]", "1000"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", "Thick"}], ",", 
     RowBox[{"AxesStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"Thick", ",", 
        RowBox[{"FontWeight", "\[Rule]", "Bold"}], ",", 
        RowBox[{"FontSize", "\[Rule]", "18"}]}], "]"}]}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"Map", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"#", ",", 
           RowBox[{"FontWeight", "\[Rule]", "Bold"}], ",", 
           RowBox[{"FontSize", "\[Rule]", "28"}]}], "]"}], "&"}], ",", 
        RowBox[{"{", 
         RowBox[{"\"\<Iterations\>\"", ",", "\"\<Estimates\>\""}], "}"}]}], 
       "]"}]}]}], "]"}]}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"Export", "[", 
  RowBox[{
  "\"\</shared/uci/research/eric/writing/dd_jrnl_paper/figures/paramrez.pdf\>\
\"", ",", "plot"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"Export", "[", 
  RowBox[{
  "\"\</shared/uci/research/eric/writing/dd_jrnl_paper/figures/paramrez.jpg\>\
\"", ",", "plot"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"Export", "[", 
  RowBox[{
  "\"\</shared/uci/research/eric/writing/dd_jrnl_paper/figures/paramrez.eps\>\
\"", ",", "plot"}], "]"}]}], "Input",
 CellChangeTimes->{{3.431454041149852*^9, 3.431454084281561*^9}, {
  3.431454117171075*^9, 3.431454180445082*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwl2Xk4VN//AHBrITFzh1Ap2heSUlT0PpWtSClpE0JCskZIKm1IElLJlrKU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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->
   NCache[{{0, 2 Pi}, {-0.9999998592131705, 0.9999998782112116}}, {{
     0, 6.283185307179586}, {-0.9999998592131705, 0.9999998782112116}}],
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.431454180927252*^9}}],

Cell[BoxData["\<\"/tmp/foo3.pdf\"\>"], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.43145418124025*^9}}],

Cell[BoxData["\<\"/tmp/foo3.jpg\"\>"], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.431454181504693*^9}}],

Cell[BoxData["\<\"/tmp/foo3.eps\"\>"], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.43145418173177*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, 
   {Hue[0.67, 0.6, 0.6], Thickness[Large], PointBox[CompressedData["
1:eJxdzzurnUUUBuClVcoIRiwEdySISJDjLRpvZ6IxMRp1J5qrl4zR3Lxla2E6
Wf6DU4goNkNKkXg6K2ERULAQI3ZWH4qtnD+giGSDnGea4ZmXl5fZfvri4TM3
RsTOGyL+u6+fjdUv1v4+dvsfv68uH9qWq4sfPt/x29Jb8FZ8M761zX/8c/9N
u35d+jY8a99+/fPH2w/8svQdeAe+E9/VDn50au+Rr75f+m68E9+DV9qVT386
+tm575a+F9+H78cP4AfxLvwQfhjvxo+0se2Wy1f3fbn0o/gx/Dh+Aq/ittmJ
Yw85jifJcTxFjmMvOY6nyXHsa5f+mn9y6dCVZY5jPzmOZ8hxHCDH8Sw5jufI
cRwkx/E8OY4XyHG8SI5j3i4fP7F72z/fXHfDiQvHIfo4ceE4TB8nLhwv0ceJ
C8fL9HHiwnGEPk5cOI7Sx4kLxzH6OHHhOE4fJy4cJ+jjxIXjJH2cuHC8Qh8n
Lhyv0seJC8dr9HHiwvE6fZy4cJyijxMXjr7ZM9xwx4kHLjzheIN93HDHiQcu
POE4zT5uuOPEAxeecLzJPm6448QDF55wvMU+brjjxAMXnnCcYR833HHigQtP
OM6yjxvuOPHAhScc59jHDXeceODCE47z7OOGO048cOEJxwX2ccMdJx648ITj
bfZxwx0nHrjwhOMd9nHDHSceuPCE4132ccMdJx648ITjPfZxwx0nHrjwhON9
9nHDHSceuPCE4yL7uOGOEw9ceMKx2OyteIZXcMNz3PECJ17DA6/jwtfwhDdw
fMD/8Qyv4IbnuOMFTryGB17Hha/hCW/g+JD/4xlewQ3PcccLnHgND7yO63//
C0P3XgY=
     "]]}, 
   {Hue[0.9060679774997897, 0.6, 0.6], Thickness[Large], 
    PointBox[CompressedData["
1:eJxd1H1MlWUYx/G7Np1bL5A1y6X5iET4sjqBKZjKZb6L4hEFCVQeD5AeFEVm
bi20uwyzTGOGpvbPo2Ol2YRpOTNtN9NFc/2hUNmb9qhpalpMWzrbXGs9d+3+
nn/OPs9v167db1f/xNLCqjuVUkPuUOqf/39/XXlvfLdx4u7Lp/KiD+K6h2S9
m/bE2y1h5FT4AfghuI+sbRj8/aS+P0f24DQ4Hc6AM+FB8BD4cTgGPwlnwdny
yq4N49NeuBp5KPwUPAweDufAufAI+Gk5NvSL7IGPXok8Eh4lPzyWogbkXYw8
Gs6T7PxDw/f1+iWyuNawGoMcVs8gh9VY5LAahxxW45HDagLyCfLXHw034u32
vCa61rCahHySTFh3/92/nb4Q5ZNd68mSt//S0S/jNp/iWk+R8lufHuu2xO53
vmudL3s7V9Q90t/mU13rqZJx+FbWlmybT3OtYVWAHFbTkU+XI21ZV76961KU
x10LrGETlzU59Q3t5+3+z3AtsIbNDJly4/xJtfJ8VF/oWgql5vLW3b/ejKxh
UyjDnj+wfnSJrZ/pWmZKcq35ZNO2aN5o2MBqlmuBNWxmSfye5j9vd5yN6otc
S5Fc2/Rq/ES9vU+wKZKSzZMv7Km09cWupVgO5nb/sX3ouageNsVys3lM7z4H
7f2d7VpmS9WBNR83n7D9YTNbmlZ610su2Hle4lpKJCVzxMNnCs9E9bApkbam
2sSLLbb/s64F1rCBVSnqS6VLXlu0vsW+J9iUSubZjtio3+37K3MtZdJ7342q
Zf1sf9jAao5rgTVs5six1bvLzXBbP9e1wBo2c6WgaXXnznL7fua5FljDBlbl
qC+X+pvbm/Zl2/6wgZXv2oMF9mENB76Me/DeFY097XuFQ1jNd+3Nl9zXhxyo
67LnCfuwhgPYwCGsEq69hKxqi6W31EfzUmA/Id+krn1/eantDwcJKevV83TO
FtsfDhPy3qBp4f6jdv0Vrj1YYL9CDuaXnVyVaucjHFTItm1fjzu82c4LOKyQ
9J2fZeTvsfOn0rUHC+xXyrX09vtearXzCQ5gA4eV8tPWYHlBkV1/lWsPFtiH
NRzABg6rpKNzV9qlfrb/c649WGAf1nAAGziE1QL0hwX2YQ0HsIHDBdKtI/ej
I8re/4WuvYWy81D3HadSbH/YhzUcwAYOYZV07cEC+0nJKNqxsfOinT9wkJSU
nL6tb1bZ9weHSVlvMsZmJu38q3btVcvF5F69vcDOH9ivlu63v2o4+6HtDwew
gcNqOVp8Yt7g/85/kWsPFtiHNRzABg5htRj9YYF9WMMBbOBwsXQOLEhuGGDv
X41rr0bekXXXZYzdf9iHNRzUSLeR5/J7vGX7wyGslrj2YIF9WMMBbOAQVkvR
HxbYhzUcLJVbV/d8/nLM3n84hFWt61TYg2OwwHHYh2thDTfWyugPZsb8BXY9
cCts4ONwCHfBapnrVNiDY7DAcdiHa2ENN8IB3Aob+Dgcwl2wqsP6YQ+OwQLH
YR+uhTXcCAdwK2z+998+g9vJ
     "]]}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesLabel->{
    FormBox[
     StyleBox[
     "\"Iterations\"", FontWeight -> Bold, FontSize -> 28, StripOnInput -> 
      False], TraditionalForm], 
    FormBox[
     StyleBox[
     "\"Estimates\"", FontWeight -> Bold, FontSize -> 28, StripOnInput -> 
      False], TraditionalForm]},
  AxesOrigin->{0, 0},
  AxesStyle->Directive[
    Thickness[Large], FontWeight -> Bold, FontSize -> 18],
  ImageSize->1000,
  PlotRange->{{1., 300.}, {0, 1}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], Automatic}]], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.431454181887677*^9}},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]],

Cell[BoxData["\<\"/shared/uci/research/eric/writing/dd_jrnl_paper/figures/\
paramrez.pdf\"\>"], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.431454182538463*^9}}],

Cell[BoxData["\<\"/shared/uci/research/eric/writing/dd_jrnl_paper/figures/\
paramrez.jpg\"\>"], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.431454183548558*^9}}],

Cell[BoxData["\<\"/shared/uci/research/eric/writing/dd_jrnl_paper/figures/\
paramrez.eps\"\>"], "Output",
 CellChangeTimes->{{3.431454168282246*^9, 3.431454184365402*^9}}]
}, Open  ]]
},
WindowSize->{1366, 1020},
WindowMargins->{{Automatic, -1502}, {-32, Automatic}},
ShowSelection->True,
FrontEndVersion->"6.0 for Linux x86 (32-bit) (April 20, 2007)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[569, 21, 871, 14, 52, "Input"],
Cell[CellGroupData[{
Cell[1465, 39, 755, 19, 209, "Input"],
Cell[2223, 60, 410, 7, 24, "Message"]
}, Open  ]],
Cell[2648, 70, 1040, 21, 77, "Input"],
Cell[3691, 93, 8260, 223, 495, "Input"],
Cell[11954, 318, 188, 5, 36, "Input"],
Cell[CellGroupData[{
Cell[12167, 327, 41022, 1050, 1309, "Input"],
Cell[53192, 1379, 18936, 321, 414, 5247, 95, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[72165, 1705, 5478, 136, 605, "Input"],
Cell[CellGroupData[{
Cell[77668, 1845, 7841, 227, 29, "Print"],
Cell[85512, 2074, 6762, 192, 29, "Print"]
}, Open  ]]
}, Open  ]],
Cell[92301, 2270, 1316, 24, 99, "Input"],
Cell[93620, 2296, 478, 11, 32, "Input"],
Cell[CellGroupData[{
Cell[94123, 2311, 1163, 28, 143, "Input"],
Cell[95289, 2341, 177, 2, 24, "Print"],
Cell[95469, 2345, 179, 2, 31, "Output"],
Cell[95651, 2349, 179, 2, 31, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[95867, 2356, 987, 22, 77, "Input"],
Cell[96857, 2380, 1673, 32, 433, "Output"],
Cell[98533, 2414, 3402, 61, 404, "Output"]
}, Open  ]],
Cell[101950, 2478, 2102, 55, 319, "Input"],
Cell[CellGroupData[{
Cell[104077, 2537, 697, 18, 77, "Input"],
Cell[104777, 2557, 17052, 287, 567, 12373, 208, "CachedBoxData", "BoxData", \
"Output"],
Cell[121832, 2846, 49563, 819, 547, 23514, 390, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[171432, 3670, 304, 7, 55, "Input"],
Cell[171739, 3679, 228, 3, 31, "Output"],
Cell[171970, 3684, 226, 3, 31, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[172233, 3692, 142, 3, 32, "Input"],
Cell[172378, 3697, 313296, 5005, 14672, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[485711, 8707, 5357, 137, 473, "Input"],
Cell[491071, 8846, 15410, 344, 726, "Output"],
Cell[506484, 9192, 428, 8, 24, "Message"],
Cell[506915, 9202, 435, 8, 24, "Message"],
Cell[507353, 9212, 420, 8, 24, "Message"],
Cell[507776, 9222, 440, 8, 24, "Message"],
Cell[508219, 9232, 115082, 2048, 938, 31298, 673, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[623316, 11283, 51343, 1212, 3135, "Input"],
Cell[CellGroupData[{
Cell[674684, 12499, 2243, 58, 253, "Input"],
Cell[676930, 12559, 9388, 160, 232, "Output"],
Cell[686321, 12721, 113, 1, 31, "Output"],
Cell[686437, 12724, 114, 1, 31, "Output"],
Cell[686554, 12727, 113, 1, 31, "Output"],
Cell[686670, 12730, 17275, 297, 581, 3247, 66, "CachedBoxData", "BoxData", \
"Output"],
Cell[703948, 13029, 171, 2, 31, "Output"],
Cell[704122, 13033, 171, 2, 31, "Output"],
Cell[704296, 13037, 171, 2, 31, "Output"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
